//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*- /////////////////////////////////////////////////
//
// $Id: mastercontrol.C,v 1.54 2000/10/19 15:27:37 goetz Exp $
//
// 
// cwz-build-command: make mrbplayer
////////////////////////////////////////////////////////////////

#include "mastercontrol.h"

#include "drawMacros.h"
#include "Config.h"

#undef COUT
#undef ACTTRACE
#undef DRAWTRACE
#undef COMMTTRACE
#undef USETIMETRACE

// trace macro: uncomment the line you want.

//#define COUT(output) cout << getTime() << " MC: " << output << endl
//#define DRAWTRACE(output) cout << getTime() << "% " << output << endl
//#define ACTTRACE(output) cout << getTime() << " MC: " << output << endl
//#define COMMTRACE(output) cout << getTime() << " MC: " << output << endl
//#define USETIMETRACE

#ifndef COUT
#define COUT(output)
#else
#warning: MC debug trace COUT is activated!
#endif

#ifndef DRAWTRACE
#define DRAWTRACE(output)
#else
#warning: MC debug trace DRAWTRACE is activated!
#endif

#ifndef ACTTRACE
#define ACTTRACE(output)
#else
#warning: MC debug trace ACTTRACE is activated!
#endif

#ifndef COMMTRACE
#define COMMTRACE(output)
#else
#warning: MC debug trace COMMTRACE is activated!
#endif

#ifndef USETIMETRACE
#define TIMETRACE(marker)
#define INITTTRACE
#define TOTALTTRACE
#else
#warning: MC debug trace TIMETRACE is activated!
#include <sys/resource.h>
#include <time.h>

static rusage start, total_start, end;

#define INITTTRACE \
getrusage(RUSAGE_SELF, &start); \
getrusage(RUSAGE_SELF, &total_start); \
cout << getTime() << " TT: starting" << endl; 

#define TIMETRACE(marker) \
getrusage(RUSAGE_SELF, &end); \
cout << getTime() << " TT: " marker " " << \
((1.0e-3*end.ru_utime.tv_usec   + 1.0e3*end.ru_utime.tv_sec) - \
 (1.0e-3*start.ru_utime.tv_usec + 1.0e3*start.ru_utime.tv_sec)) << endl; \
start = end;

#define TOTALTTRACE \
getrusage(RUSAGE_SELF, &end); \
cout << getTime() << " TT: total time______________ " << \
((1.0e-3*end.ru_utime.tv_usec         + 1.0e3*end.ru_utime.tv_sec) - \
 (1.0e-3*total_start.ru_utime.tv_usec + 1.0e3*total_start.ru_utime.tv_sec)) << endl;
#endif

MasterControl::MasterControl(const char* team, bool goalie, const char *host,
			     int port, double mapwm, bool log, int recn,
			     const char *configfile) :
  Player(team, goalie, host, port, mapwm, log, recn) ,
  lastTimeMsgSent(-1)
{
  if (!skills)
    exit(-666);

  // try to initialize config file
  if (!Config::readFile(configfile)) {
    cerr << "PANIC: Could not read config file: " << configfile << endl;
    exit(-1);
  }

  goaleval     = new GoalKickEval();
  passeval     = new PassEval(goaleval);

  tactical     = new TacticalModule();
  standard     = new StandardSituationModule(goalie, passeval, goaleval);
  pass         = new PassModule(passeval, goaleval);
  ballhandling = new BallHandlingModule(goalie);
  goalkick     = new GoalKickModule(goalie, goaleval);
  position     = new PositionModule(goalie, passeval, goaleval);
}

MasterControl::~MasterControl() {

  delete tactical;
  delete standard;
  delete pass;
  delete passeval;
  delete ballhandling;
  delete goalkick;
  delete goaleval;
  delete position;
}

void
MasterControl::act() {
  INITTTRACE;

  AudioMsg audiomsg;
  while (getNextMsg(audiomsg)) {
    COMMTRACE("hearing message: " << audiomsg);
    // first give all msgs to worldmodel, then modules
    worldFeedbackAudio(audiomsg);

    // only worldmodel can handle (shorter) coach msgs
    if (audiomsg.isCoachMsg()) {
      COMMTRACE("Heard coach msg.");
      continue;
    }
    
    tactical->feedbackAudio(audiomsg);

    standard->feedbackAudio(audiomsg);
    pass->feedbackAudio(audiomsg);
    ballhandling->feedbackAudio(audiomsg);
    goalkick->feedbackAudio(audiomsg);
    position->feedbackAudio(audiomsg);
  }

  // get snapshot 
  
  Snapshot *snap = getSnapshot();
  
  if (snap == NULL) {
    cerr << "Player (" << getMyTeam() << " No." << getMyNumber()
	 <<"): MC: cannot generate Snapshot of time " << getTime() << endl;
    return; 
  } 
  
  COUT("doing world feedback for tactical module...");
  tactical->feedbackWorld(snap);

  // get tactical snapshot from tactical module

  TacticalSnapshot *tsnap = tactical->getSnapshot();

  COUT("doing world feedback for standardsit module...");
  standard->feedbackWorld(snap, tsnap);
  COUT("doing world feedback for pass module...");
  pass->feedbackWorld(snap, tsnap);
  COUT("doing world feedback for ballhandling module...");
  ballhandling->feedbackWorld(snap, tsnap);
  COUT("doing world feedback for goalkick module...");
  goalkick->feedbackWorld(snap, tsnap);
  COUT("doing world feedback for position module...");
  position->feedbackWorld(snap, tsnap);

  TIMETRACE("after modules feed back ");

  // get all action evaluations from modules
  Grade stndG = standard->evaluate(skills, snap, tsnap);
  TIMETRACE("after standard eval     ");
  COUT("doing evaluate for standardsit module: " << stndG);
  Grade passG = pass->evaluate(skills, snap, tsnap);
  TIMETRACE("after pass     eval     ");
  COUT("doing evaluate for pass module: " << passG);
  Grade handG = ballhandling->evaluate(skills, snap, tsnap);
  TIMETRACE("after ballh    eval     ");
  COUT("doing evaluate for ballhandling module: " << handG);
  Grade goalG = goalkick->evaluate(skills, snap, tsnap);
  TIMETRACE("after goalkick eval     ");
  COUT("doing evaluate for goalkick module:" << goalG);
  Grade posnG = position->evaluate(skills, snap, tsnap);
  COUT("doing evaluate for position module:" << posnG );
  TIMETRACE("after position eval     ");

  enum { DEC_STAND, DEC_PASS, DEC_HAND, DEC_GOAL, DEC_POSN} decision= DEC_POSN;

  // handle absolute values
  
  if (stndG.is_ABSOLUTE()) 
    decision= DEC_STAND;
  else if (goalG.is_ABSOLUTE()) 
    decision= DEC_GOAL;
  else if (handG.is_ABSOLUTE())
    decision= DEC_HAND;
  else if (passG.is_ABSOLUTE())
    decision= DEC_PASS;
  else if (posnG.is_ABSOLUTE())
    decision= DEC_POSN;

  // Magic decision tree
  
  else if (!goalG.is_worse_than( handG ) &&
	   !goalG.is_worse_than( passG ) &&
	   !goalG.is_worse_than( posnG ) ) 
    decision= DEC_GOAL;
  else if (!passG.is_worse_than( handG ) && 
	   !passG.is_worse_than( posnG ) ) 
    decision= DEC_PASS;
  else if (!handG.is_worse_than( posnG ) &&
	   !handG.is_worse_than( passG ) ) 
    decision= DEC_HAND;
  else if (!passG.is_worse_than( posnG )) 
    decision= DEC_PASS;
  else 
    decision= DEC_POSN;
  
  // do actions

  ACTTRACE("std="<< stndG << ", goal=" << goalG << ", pass=" << passG 
	   << ", bhand=" << handG << ", pos=" << posnG );
  COUT("calling one module to act...");
  switch (decision) {
  case DEC_STAND:
    ACTTRACE( "handling standard situation, grade=" << stndG );
    standard->act(skills, snap, tsnap);
    break;
  case DEC_PASS:
    ACTTRACE( "passing, grade=" << passG );
    pass->act(skills, snap, tsnap);
    break;
  case DEC_HAND:
    ACTTRACE( "handling ball, grade=" << handG );
    ballhandling->act(skills, snap, tsnap);
    break;
  case DEC_GOAL:
    ACTTRACE( "kicking into goal, grade=" << goalG );
    goalkick->act(skills, snap, tsnap);
    break;
    // default action: positioning ...
  case DEC_POSN:
  default:
    ACTTRACE( "positioning, grade=" << posnG );
    position->act(skills, snap, tsnap);
  } // switch

  
  TIMETRACE("after module act        ");

  // audio message generation
  COUT( "testing for message to be sent..." );
  switch (snap->time % 5) {
    
  case 0: // empty channel never used, to make sure emergency channel is free
    COUT( "no sending allowed before emergency channel" );
    break;
  
  case 1: { // if emergency channel is used, info channel will be blocked
    if (tactical->useEmergencyChannel) {
      sendAudioMsg(snap);
      tactical->useEmergencyChannel = false;
      COMMTRACE( "sending message on emergency channel" );
    } else {
      COUT( "no sending on emergency channel" );
    }
  } break;
    
  case 2: // if info channel is used, broadcast channel will be blocked
    if ( !( stndG.is_IMPOSSIBLE() ) ) { // Now we have to handle a std sit
      if (standard->sendSpecialSituation()) {
	sendAudioMsg(snap);
	COMMTRACE( "sending message on info channel (standard situation)" );
      } else {
	COUT( "no sending on info channel (standard situation)" );
      }
    } else { // Normal playmode (tactical module decides whether to use info channel)
      if (tactical->useInfoChannel) {
	sendAudioMsg(snap);
	COMMTRACE( "sending message on info channel (tactical module wanted it)" );
      } else {
	COUT( "no sending on info channel (tactical module did not want to)" );
      }
    }
    break;
  case 3: // empty channel never used, to make sure worlmodel is free
    COUT( "no sending allowed before worldmodel broadcast" );
    break;
  case 4: { // broadcast channel always used, but only "blocks" empty channel
    if ((snap->time/5 - snap->my.number) % 2) {
      // using player numbers to avoid heavy load
      // using lastTimeMsgSent to avoid saying two msgs in a 5-cycle-slice
      if (snap->time - lastTimeMsgSent > 4) {
	sendAudioMsg(snap);
	COMMTRACE( "sending worldmodel broadcast");
      } else {
	COUT( "no worldmodel broadcast because of other msg in:" << lastTimeMsgSent );
      }
    }
    else
      COUT( "not my turn sending worldmodel broadcast" );
  } break;

  default:;
  } // switch

  TOTALTTRACE;
} //act


void MasterControl::sendAudioMsg(const Snapshot *snap)
{
  if (snap->time == lastTimeMsgSent) {
    COUT( "Already sent message in this cycle");
  }  else {
    lastTimeMsgSent= snap->time;
    
    AudioMsg sendMsg(0, snap->time);
  
    bool msgState= true;
  
    worldGetSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();
  
    tactical->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    standard->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();
  
    pass->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    ballhandling->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    goalkick->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    position->getSayMsg(sendMsg.clearState());
    msgState &= sendMsg.getState();

    if (msgState) {
      say(sendMsg);
      COUT("Sending AudioMsg: " << sendMsg );
    } else {
      cerr << "Player (" << getMyTeam() << " No." << getMyNumber()
	   <<"): MC: Fatal error: buffer overflow: "<< sendMsg
	   << getTime() << endl;
    }   
  }
}

  
