//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// mrbcoach.C - Online coach for MRB2000
//
// $Id: mrbcoach.C,v 1.14 2000/10/19 15:27:40 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#include "mrbcoach.h"
#include "snapshot.h"
#include "audiomsg.h"
#include "mrbtypes.h"
#include "minmax.h"

#undef COUT
#undef AUDIOTRACE
#undef ANALYZETRACE
#undef ACTTRACE

//#define AUDIOTRACE(output) cout << snap->time << " COACH AUDIO: " << output << endl
//#define ACTTRACE(output) cout << snap->time << " COACH act: " << output << endl
//#define ANALYZETRACE(output) cout << snap->time << " COACH analyze: " << output << endl

//////////////////////////////////////////////////////////////////////
//
// some local constants to determine coach behaviour
//
//////////////////////////////////////////////////////////////////////

// number of messages to be left to allow second coach msgs.
const int NUM_MSGS_SECOND_SAY= 30;
// max dist that is not assumed to be a goalie move (squared!)
const double MAX_GOALIE_DASH= sqr( PLAYER_SPEED_MAX*(1+QUANTIZE_STEP) );

#ifndef AUDIOTRACE
#define AUDIOTRACE(output)
#else
#warning AUDIOTRACE debug trace AUDIOTRACE is activated!
#endif

#ifndef ANALYZETRACE
#define ANALYZETRACE(output)
#else
#warning ANALYZETRACE debug trace ANALYZETRACE is activated!
#endif

#ifndef ACTTRACE
#define ACTTRACE(output)
#else
#warning ACTTRACE debug trace ACTTRACE is activated!
#endif


MrbCoach::MrbCoach(const char *team, const char *host, const int port,
		   const bool logging= false, const double mapmag= -1.0) :
  Onlinecoach(team, host, port, mapmag, logging),
  sayMessageNow(false), lastMode(PLAY_MODE_UNKNOWN), 
  countOfMsgsLeft(SAY_COACH_CNT_MAX), thisGame(0),
  ourGoalieNo(-1), otherGoalieNo(-1),
  ourGoaliePos(0,0), otherGoaliePos(0,0)
{}

// acting in standard situations
void MrbCoach::act(Snapshot *snap) 
{
  ACTTRACE("Coach is acting.");
  
  // give additional messages to coach every 6000 cycles ...
  if (snap->time - thisGame >= HALF_TIME_IN_CYCLES * 2 ) {
    thisGame= snap->time;
    countOfMsgsLeft += SAY_COACH_CNT_MAX;
    ACTTRACE( "Added msg count for additional 6000 cycles. Have now left: " 
	  << countOfMsgsLeft );
  }
  
  if ( sayMessageNow ) {

    ACTTRACE( "Acting in a new standard situation ..." );
    sayMessageNow= false;
    
    if (countOfMsgsLeft == 0) {
      cerr << "Coach: Cannot scream: Lost voice earlier!" << endl;
    }
    
    AudioMsg coachMsg(0, snap->time, true);
    
    // this depends on player's worldfeedbackaudio() method !!
    coachMsg << true << *snap;

    if (!coachMsg.getState()) {
      cerr << "Coach : Fatal error: buffer overflow:" << coachMsg << endl;
    } else {
      say(coachMsg);
      
      AUDIOTRACE( "Sending msg: " << coachMsg );
      
      countOfMsgsLeft--;
      AUDIOTRACE( "Now coach has " << countOfMsgsLeft << " msgs left" );
    }
  }
}

//////////////////////////////////////////////////////////////////////
//
// analyze() - here the playmode is used to determine whether the 
//             coach should say something now or not.
//
//////////////////////////////////////////////////////////////////////

void MrbCoach::analyze(Snapshot *snap)
{ 
  // first get player's audioMsg's to determine goalie numbers
  AudioMsg playerMsg;
  while( getNextMsg(playerMsg)) {
    //////////////////// IMPORTANT: Depends on players msg format!
    ANALYZETRACE("Got player msg:" << playerMsg );
    bool hasSnap;
    playerMsg >> hasSnap;
    if (hasSnap) {
      Snapshot sn;
      playerMsg >> sn;
      if (sn.ourGoalieNo > 0) 
	ourGoalieNo= sn.ourGoalieNo;
      if (sn.otherGoalieNo > 0)
	otherGoalieNo= sn.otherGoalieNo;
      ANALYZETRACE("Got goalie No's: our=" << ourGoalieNo << " other="
	       << otherGoalieNo );
    }
  }

  bool modeToAct(false);

  switch(snap->playMode) {
  case PLAY_MODE_KICK_OFF_US:
  case PLAY_MODE_KICK_OFF_THEM:
  case PLAY_MODE_GOAL_KICK_US:
  case PLAY_MODE_GOAL_KICK_THEM:
  case PLAY_MODE_DROP_BALL:
  case PLAY_MODE_OFFSIDE_US:
  case PLAY_MODE_OFFSIDE_THEM:
  case PLAY_MODE_KICK_IN_US:
  case PLAY_MODE_KICK_IN_THEM:
  case PLAY_MODE_CORNER_KICK_US:
  case PLAY_MODE_CORNER_KICK_THEM:
    modeToAct= true;
    break;
  case PLAY_MODE_FREE_KICK_US:
  case PLAY_MODE_FREE_KICK_THEM:
    if (lastMode != PLAY_MODE_OFFSIDE_THEM && 
	lastMode != PLAY_MODE_OFFSIDE_US ) {
      modeToAct= true;
    }
    break;
  default:
    // "normal" playmodes which cannot be handled by online coach
    ;
  }

  
  // did the playmode change ?
  if (snap->playMode != lastMode) { 
    ANALYZETRACE( "Noted Change of playmode ...");
    lastMode= snap->playMode;
    if (modeToAct) {
      sayMessageNow=true;
      ANALYZETRACE( "Set flag: scream!" );
    } else {
      ANALYZETRACE( "Not right mode to act ...");
    }
  } else { // still same mode. Now handle goalie moves ...
    if (modeToAct) {
      Vector2d ourG(0,0), otherG(0,0);
      if (ourGoalieNo > 0) 
	ourG= snap->ourTeam( ourGoalieNo ).pos;
      if (otherGoalieNo > 0)
	otherG= snap->otherTeam( otherGoalieNo ).pos;
      if ( sqrnorm(ourG - ourGoaliePos) > MAX_GOALIE_DASH ||
	   sqrnorm(otherG - otherGoaliePos) > MAX_GOALIE_DASH ) {
	if (countOfMsgsLeft > NUM_MSGS_SECOND_SAY) {
	  // only scream again if enough msgs left!
	  sayMessageNow= true;
	  ANALYZETRACE("A goalie moved; screaming again!");
	} else {
	  ANALYZETRACE("A goalie moved, but not enough msgs left: "
		       << countOfMsgsLeft );
	}
      } else {
	ANALYZETRACE("Goalies dashed, but did not move");
      }
      ourGoaliePos= ourG;
      otherGoaliePos= otherG;
    }
  }
} 

