//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// cwz-build-command: make all
AdvSkills::DribbleCommand
AdvSkills::dribble(const Snapshot *sn, Angle ddir,
		   Angle hdir = 0, double stamina = STAMINA_INC_MAX)
{
  DribbleCommand com;
  com.state = DOK;

#ifdef DRIBBLEFIGS
  FigDraw drawing("dribble", sn);
#endif

  double maxdash = DASH_POWER_RATE * max(min(stamina, MAXPOWER), 0.0);
  const WorldPlayer &my(sn->my);
  const WorldBall   &ball(sn->ball);
  Vector2d pns = my.pos + my.v, bns = ball.pos + ball.v;
  Angle ddirmor = ddir - my.bodyDir;
  double dashlen = 0;
  bool doturn = false, contBall = false;
  Vector2d res_v; // only set for doturn = contBall = false, e.g. dash

  // modify ball position a little bit
  Vector2d tgtvec = radAng2Vector2d(safeCircle, hdir);
  tgtvec += sn->my.v;
  hdir = tgtvec.getAngle();

  // check direction of ball
  Angle cur = (ball.pos - my.pos).getAngle();
  if (fabs(cur - hdir) > dribbleHoldErr) {
#ifdef DRIBBLEFIGS
    drawing.Text("angle");
#endif
    contBall = true;
  }

  if (!contBall) {
    if (fabs(ddirmor) > dribbleBodyErr) {
      // calc for turn
      doturn = true;
      if (norm(pns - bns) > kickRange) {
#ifdef DRIBBLEFIGS
	drawing.Text("coll");
#endif
	contBall = true;
      }
    }
    else {
      // calc for dash
      double llow, lhigh;
      if (intersectLineCircle(radAng2Vector2d(1, my.bodyDir), bns - pns,
			      kickRange, llow, lhigh) != Valid) {
#ifdef DRIBBLEFIGS
	drawing.Text("dash out of range next step");
#endif
	contBall = true;
      }
      else {
	double lllow, llhigh;
	if (intersectLineCircle(radAng2Vector2d(1, my.bodyDir), -my.v,
				dribbleMaxV, lllow, llhigh) != Valid) {
#ifdef DRIBBLEFIGS
	  drawing.Text("cannot reach a valid vel???");
#endif
	  contBall = true;
	}
	else {
	  // intersect the two possible intervals
	  if (llow < lllow)
	    llow = lllow;
	  if (llow < 0)
	    llow = 0;
	  
	  if (lhigh > llhigh)
	    lhigh = llhigh;
	  if (lhigh > maxdash)
	    lhigh = maxdash;
	  if (llow > lhigh) {
	    // no valid velocities, control ball
#ifdef DRIBBLEFIGS
	    drawing.Text("cannot reach a valid vel???");
#endif
	    contBall = true;
	  }
	  else {
	    dashlen = lhigh;
	    res_v = my.v + radAng2Vector2d(dashlen, my.bodyDir);
	    pns = my.pos + res_v;
	  }
	}
      }
    }
  }

  // check for collision
  if (!contBall) {
    if (norm(my.pos - bns) < innerSafeCircle) {
      // collision by ball movement first => place it away
#ifdef DRIBBLEFIGS
	    drawing.Text("collision if ball moves first");
#endif
	    contBall = true;
      contBall = true;
    }
    else if (norm(pns - bns) < innerSafeCircle) {
      // collision by second movement => place ball away
#ifdef DRIBBLEFIGS
      drawing.Text("collision if player moves first");
#endif
      contBall = true;
    }
  }

  if (contBall) {
    KickCommand kcm = placeBall(sn, tgtvec);
    if (kcm.state == KRANGE) {
      com.state = DGOINGFORBALL;
      // oops, somehow we lost the ball, go get it
      TurtleCommand tcm = controlBall(sn, stamina);
      switch (tcm.command) {
      case TurtleCommand::TURN:
	com.command = DribbleCommand::TURN;
	break;
      case TurtleCommand::DASH:
	com.command = DribbleCommand::DASH;
      default:;
      }
      com.v_p     = tcm.v;
      com.v_b     = sn->ball.v;
      com.bodyDir = tcm.bodyDir;
      return com;
    }

    com.command = DribbleCommand::KICK;
    com.v_p     = sn->my.v;
    com.v_b     = kcm.v;
    com.bodyDir = sn->my.bodyDir;
    return com;
  }

  if (doturn) {
    com.command = DribbleCommand::TURN;
    com.v_p     = sn->my.v;
    com.v_b     = sn->ball.v;
    TurnCommand tmptcm = turn(sn, ddir);
    com.bodyDir = tmptcm.bodyDir;
    return com;
  }

  com.command = DribbleCommand::DASH;
  com.v_p     = sn->my.v + radAng2Vector2d(dashlen, sn->my.bodyDir);
  com.v_b     = sn->ball.v;
  com.bodyDir = sn->my.bodyDir;
  return com;
}

void AdvSkills::exec(const Snapshot *sn, const DribbleCommand &com) {
  switch (com.command) {
  case DribbleCommand::KICK: {
    Vector2d kick_v = com.v_b - sn->ball.v;
    double res_power = norm(kick_v) / sn->my.kickPowerRate(sn->ball);
    Angle res_angle = kick_v.getAngle() - sn->my.bodyDir;
    // renormation necessary for precision reasons
    if (res_power > MAXPOWER)
      res_power = MAXPOWER;
    basicKick(res_power, res_angle);
    break;
  }
  case DribbleCommand::TURN:
    basicTurn(com.bodyDir - sn->my.bodyDir);
    break;
  case DribbleCommand::DASH:
    basicDash(norm(com.v_p - sn->my.v)/DASH_POWER_RATE);
    break;
  default:
    cerr << sn->time << "AdvSkills::exec(DribbleCommand): called without valid command" << endl;
    noAction();
  }  
}

AdvSkills::TurnHeadCommand
AdvSkills::turnHeadForDribble(const Snapshot *sn, const DribbleCommand &dcm, double ddir)
{
  if (dcm.state == DGOINGFORBALL) {
    TurtleCommand tcm;
    if (dcm.command == DribbleCommand::TURN)
      tcm.command = TurtleCommand::TURN;
    else
      tcm.command = TurtleCommand::DASH;
    tcm.bodyDir = dcm.bodyDir;
    tcm.v       = dcm.v_p;
    return turnHeadToBall(sn, tcm);
  }
  switch (dcm.command) {
  case TurtleCommand::TURN:
    return turnHead(sn, ddir + sn->my.bodyDir - dcm.bodyDir);
  default:
    return turnHead(sn, ddir);
  }
}
