//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "Ensure.h"
// cwz-build-command: cd ../advtest; make all

double AdvSkills::goToThresh(double error, double tdist)
{
  // check if turn is needed (distance depended)
  // visual angle of target/error circle
  double gamma = atan2(error, tdist);
  double split = 1/(1 + goToThreshStretch*(tdist - error));
  return (1 - split)*M_PI/2 + split*gamma;
}

AdvSkills::TurtleCommand
AdvSkills::goToPosition(const Snapshot *sn,
			const Vector2d &target,
			double error, double stamina,
			bool minAngle, Angle prefAngle)
{
  TurtleCommand com;
  com.state = TONWAY;
  // time information not valid
  com.time = -1;
  // simple turn/dash algorithm. Will not work for small errors, I guess.
  // But it is not for that at all;

  com.v = target - sn->my.pos; 
  Angle tangle = com.v.getAngle();
  double tdist = norm(com.v);

  if (tdist <= error) {
    com.time  = 0;
    com.state = TFINISHED;
    return com;
  }

  // angle minimization if wanted
  bool reverse = false;
  if (minAngle && (fabs(tangle - prefAngle) > M_PI/2)) {
    reverse = true;
    tangle += PI;
    com.state = TGOINGBACKWARDS;
  }

  // check if turn is needed (distance depended)
  Angle diff = tangle - sn->my.bodyDir;
  if (fabs(diff) > goToThresh(error, tdist)) {
    com.v       = sn->my.v;
    TurnCommand tmptcm = turn(sn, tangle);
    com.bodyDir = tmptcm.bodyDir;
    com.turnMoment = tmptcm.turnMoment;
    com.command = TurtleCommand::TURN;
    return com;
  }

  double minacc, maxacc;
  calcAcc(sn, stamina, minacc, maxacc);

  double l1, l2, tgtDash = 0;
  // overkill since this is very likely more exact than error implies
  switch (intersectLineCircle(radAng2Vector2d(1, sn->my.bodyDir),
			      com.v - sn->my.v, error, l1, l2)) {
  case None: // no panic. We didn't turn since we are far away.
    // dash at full speed (haha)
    if (reverse)
      tgtDash = minacc;
    else
      tgtDash = maxacc;
    
    break;
  case Valid:
    if (l1 > maxacc)
      // dash forward
      tgtDash = maxacc;
    else if (l2 < minacc)
      // dash backwards
      tgtDash = minacc;
    else
      tgtDash = (max(l1, minacc) + min(l2, maxacc))/2;
    break;
  case Instable: // impossible, norm(rad...) = 1
  default: ; // R1larger/R2larger only for Circle/Circle
  }

  com.v       = sn->my.v + radAng2Vector2d(tgtDash, sn->my.bodyDir);
  com.bodyDir = sn->my.bodyDir;
  com.command = TurtleCommand::DASH;
  return com;
}

double AdvSkills::reachRadius(double dlen, int numDashes)
{
  return dlen/(1 - PLAYER_DECAY)*
    (numDashes - PLAYER_DECAY*(1 - power(PLAYER_DECAY, numDashes))/(1 - PLAYER_DECAY));
}

AdvSkills::TurtleCommand
AdvSkills::goToPositionTimed(const Snapshot *sn, const Vector2d &target,
			     double stamina, double error)
{
  TurtleCommand com;
  com.state = TONWAY;

  const WorldPlayer &my(sn->my);
  const Vector2d mybvec = radAng2Vector2d(1, my.bodyDir);

  INTERCEPTDRAW(sn->time, CIRCLE(target, error));

  double tdist = norm(my.pos - target); 
  if (tdist < error) {
    com.state = TFINISHED;
    com.time  = 0;
    return com;
  }

  // calculate the increasing circles of reachable positions
  // simMy gives my position without any dashes, that is the center of circle
  WorldPlayer simMy = my;

  // after this loop, tgt is where we head
  Vector2d tgt = target;
  // power in reduced coordinates
  double dpower = DASH_POWER_RATE*STAMINA_INC_MAX; // stamina to go there, avgd over time to go
  double len; // distance correctable by that stamina
  int time = 0;
  for (;;) {
    time++;
    simMy.simulate();
    // stop simulation if influence of rest velocity is too low
    if ((time > minInterceptTime) &&
	(sqr(1/(1 - PLAYER_DECAY))*sqrnorm(simMy.v) < sqr(0.01*error))) {
      dpower = DASH_POWER_RATE*STAMINA_INC_MAX;
      break;
    }
    // first check if dashing alone would do it
    dpower = DASH_POWER_RATE*min(MAXPOWER, (STAMINA_INC_MAX + stamina/time));
    len = reachRadius(dpower, time);

    INTERCEPTDRAW(sn->time, LINE(simMy.pos, simMy.pos + len*mybvec));

    double l1, l2;
    if ((intersectLineCircle(mybvec, target - simMy.pos, error, l1, l2) == Valid) &&
	(l1 < len) && (l1 > 0)) {
      // I can get there without turn
      // calculate needed dash for one dash
      double dlen = l1*(1- PLAYER_DECAY)/
	(time - PLAYER_DECAY*(1 - power(PLAYER_DECAY, time))/(1 - PLAYER_DECAY));
      // and cut down if more are needed
      if (dlen > dpower)
	dlen = dpower;
      com.time    = time;
      com.v       = sn->my.v + radAng2Vector2d(dlen, sn->my.bodyDir);
      com.bodyDir = sn->my.bodyDir;
      com.command = TurtleCommand::DASH;

      INTERCEPTDRAW(sn->time, CTEXT("dashing straight, dlen= " << dlen << " time " << time));

      return com;
    }

    // now check one turn, if possible
    if (time > 1) {
      dpower = DASH_POWER_RATE*min(MAXPOWER, (STAMINA_INC_MAX + stamina/(time - 1)));
      len = reachRadius(dpower, time - 1);

      INTERCEPTDRAW(sn->time, CIRCLE(simMy.pos, len));

      Vector2d l1, l2;
      bool usetgt = false;
      switch (intersectCircleCircle(len, target - simMy.pos, error, l1, l2)) {
      case Valid: // goto into middle of solutions
	tgt = simMy.pos + 0.5*(l1 + l2);
	usetgt = true;
      case R1larger:
	// go straight into target
	usetgt = true;
      default:;
      }

      if (usetgt) {
	// check for angle not too large
	Angle dashdir = (tgt - simMy.pos).getAngle();
	TurnCommand tcm = turn(sn, dashdir);
	// works, go to tgt
	if (tcm.state != RANGLE)
	  break;
      }
    }

    // check two turns
    if (time > 2) {
      dpower = DASH_POWER_RATE*min(MAXPOWER, (STAMINA_INC_MAX + stamina/(time - 2)));
      len = reachRadius(dpower, time - 2);

      INTERCEPTDRAW(sn->time, CIRCLE(simMy.pos, len));

      Vector2d l1, l2;
      bool usetgt = false;
      switch (intersectCircleCircle(len, target - simMy.pos, error, l1, l2)) {
      case Valid: // goto into middle of solutions
	tgt = simMy.pos + 0.5*(l1 + l2);
	usetgt = true;
      case R1larger:
	// go straight into target
	usetgt = true;
      default:;
      }

      if (usetgt)
	// ok, with two turns we get the ball, go!
	break;
    }
  }
  // if we fall here, tgt is where we should go directly
  Angle dashdir = (tgt - simMy.pos).getAngle();
  // test for turn delay since turn too small
  if (fabs(dashdir - my.bodyDir) < goToThresh(error, tdist)) {
    com.bodyDir = my.bodyDir;
    com.v = my.v + radAng2Vector2d(dpower, my.bodyDir);
    com.time = time;
    com.command = TurtleCommand::DASH;

    INTERCEPTDRAW(sn->time, CTEXT("dashing, turn delayed, time= " << time << " power= " << dpower));

    return com;
  }
  TurnCommand tcm = turn(sn, dashdir);
  com.turnMoment = tcm.turnMoment;
  com.bodyDir = tcm.bodyDir;
  com.v = my.v;
  com.time = time;
  com.command = TurtleCommand::TURN;

  INTERCEPTDRAW(sn->time, LINE(simMy.pos, tgt));
  INTERCEPTDRAW(sn->time, CTEXT("turning, time= " << time  << " dashdir= " << dashdir));

  return com;
}

AdvSkills::TurtleCommand
AdvSkills::controlBall(const Snapshot *sn, double stamina)
{
  WorldBall tmp;
  return controlBall(sn, stamina, kickRange, tmp); 
}

AdvSkills::TurtleCommand
AdvSkills::controlBall(const Snapshot *sn, double stamina, double error, WorldBall &meet)
{
  TurtleCommand com;
  com.state = TONWAY;

  const WorldPlayer &my(sn->my);
  const WorldBall   &ball(sn->ball);
  double bdist = my.distTo(ball);

  meet = ball;

  if (bdist < error) {
    com.time  = 0;
    com.state = TFINISHED;
    return com;
  }

#ifdef INTERCEPTFIGS
  FigDraw drawing("CONTROL", sn);
#endif

  // we need at least one dash
  meet.simulate();

  ///////////////////////////////// single dash ?
  // also dashes backwards
  Vector2d relpos = ball.pos + ball.v - my.pos - my.v;
  double l1, l2;

  // last dash ALWAYS at full power
  double maxacc = DASH_POWER_RATE*MAXPOWER, minacc = DASH_POWER_RATE*MINPOWER;

  // check for dash on safeCircle, or if not possible, into error
  if (intersectLineCircle(radAng2Vector2d(1, my.bodyDir),
			  relpos, innerSafeCircle,
			  l1, l2) == Valid) {
    double tgtDash = 0;
    bool use = false;
    bool l1ok = (l1 >= minacc) && (l1 <= maxacc);
    bool l2ok = (l2 >= minacc) && (l2 <= maxacc);

    if (l1ok && !l2ok) {
      tgtDash = l1;
      use = true;
    }
    else if (l2ok && !l1ok) {
      tgtDash = l2;
      use = true;
    }
    else if (l1ok && l2ok) {
      // use lower dash (dashing backwards costs double stamina)
      double l1stamina = (l1 > 0) ? l1 : 2*l1;
      double l2stamina = (l2 > 0) ? l2 : 2*l2;
      tgtDash = (fabs(l1stamina) < fabs(l2stamina)) ? l1 : l2;
      use = true;
    }
    if (use) {
      Vector2d dashVec = radAng2Vector2d(tgtDash, my.bodyDir);
#ifdef INTERCEPTFIGS
      drawing.Line(my.v, dashVec, 6);
      drawing.Text("dash only safe Circle");
#endif
      com.time    = 1;
      com.v       = sn->my.v + dashVec;
      com.bodyDir = sn->my.bodyDir;
      com.command = TurtleCommand::DASH;
      return com;
    }
  }

  // safeCircle didn't do it, try kickRange
  if (intersectLineCircle(radAng2Vector2d(1, my.bodyDir),
			  relpos, error,
			  l1, l2) == Valid) {
    if ((l1 <= maxacc) && (l2 >= minacc)) {
      // ok, not totally safe, but at least safe kickable
      Vector2d dashVec = radAng2Vector2d((max(l1, minacc) +
					  min(l2, maxacc))/2, my.bodyDir);
      
#ifdef INTERCEPTFIGS
      drawing.Line(my.v, dashVec, 6);
      drawing.Text("dash only kick range");
#endif
      com.time    = 1;
      com.v       = sn->my.v + dashVec;
      com.bodyDir = sn->my.bodyDir;
      com.command = TurtleCommand::DASH;
      return com;
    }
  }

  ///////////////////////////////// turn-dash ?
  // now we need at least two steps
  meet.simulate();

  relpos += BALL_DECAY*ball.v - PLAYER_DECAY*my.v;
  Vector2d vl1, vl2, dashvec(0, 0);
  bool tdgood = false;
  switch(intersectCircleCircle(maxacc, relpos, error, vl1, vl2)) {
  case Valid:
    tdgood = true;
    dashvec = 0.5*(vl1 + vl2);
    break;
  case R1larger:
    tdgood = true;
    dashvec = relpos;
    break;
  case R2larger: // hmm, we already got the ball (???)
    tdgood = true; // do essentially nothing (since we turn fist)
    dashvec = radAng2Vector2d(0.0001, relpos.getAngle());
  case None:
  default:;
  }

  if (tdgood) {
#ifdef INTERCEPTFIGS
    drawing.Line(my.v + PLAYER_DECAY*my.v, dashvec, 7);
    drawing.Text("turn dash");
#endif
    com.time    = 2;
    com.v       = my.v;
    TurnCommand tmptcm = turn(sn, dashvec.getAngle()); 
    com.bodyDir = tmptcm.bodyDir;
    com.turnMoment = tmptcm.turnMoment;
    com.command = TurtleCommand::TURN;
    if (tmptcm.state != RANGLE)
      return com;
    // check if proposed turn still is enough
    double l1, l2;
    if (intersectLineCircle(radAng2Vector2d(1, tmptcm.bodyDir),
			    relpos, error, l1, l2) == Valid)
      return com;
  }

  // ok, now we simulate, more than 3 steps needed
  com.state = TONWAY;

  const Vector2d mybvec = radAng2Vector2d(1, my.bodyDir);

  // calculate the increasing circles of reachable positions
  // simMy gives my position without any dashes, that is the center of circle
  WorldPlayer simMy = my;

  // after this loop, tgt is where we head
  // initialize to infinity position
  Vector2d tgt = ball.pos + (1/(1 - BALL_DECAY))*ball.v;
  double dpower = DASH_POWER_RATE*STAMINA_INC_MAX; // stamina to go there, avgd over time to go
  double len; // distance correctable by that stamina

  int time = 2; // this we know from before
  // hopefully not an endless loop
  // preserve stamina for last turn with full power
  stamina = max(0.0, stamina - 2*MAXPOWER);

  for (;;) {
    time++;
    meet.simulate();
    simMy.simulate();
    // stop simulation if ball and player too slow
    if ((time > minInterceptTime) &&
	(sqr(1/(1 - BALL_DECAY))*sqrnorm(meet.v) < sqr(0.5*error)) &&
	(sqr(1/(1 - PLAYER_DECAY))*sqrnorm(simMy.v) < sqr(0.01*error))) {
      INTERCEPTDRAW(sn->time, TEXT("ball or player too slow, exiting",
				   Vector2d(-FIELD_LENGTH/2, -FIELD_WIDTH/2+23)));
      dpower = DASH_POWER_RATE*STAMINA_INC_MAX;
      // go slowly near target
      break;
    }

    // first check if dashing alone would do it
    dpower = DASH_POWER_RATE*min(MAXPOWER, (STAMINA_INC_MAX + stamina/time));
    len = reachRadius(dpower, time);

    INTERCEPTDRAW(sn->time, LINE(simMy.pos, simMy.pos + len*mybvec));

    double l1, l2;
    if ((intersectLineCircle(mybvec, meet.pos - simMy.pos, error, l1, l2) == Valid) &&
	(l1 < len) && (l1 > 0)) {
      // I can get there without turn
      // calculate needed dash for one dash
      double dlen = l1*(1- PLAYER_DECAY)/
	(time - PLAYER_DECAY*(1 - power(PLAYER_DECAY, time))/(1 - PLAYER_DECAY));
      // and cut down if more are needed
      if (dlen > dpower)
	dlen = dpower;
      com.time    = time;
      com.v       = sn->my.v + radAng2Vector2d(dlen, sn->my.bodyDir);
      com.bodyDir = sn->my.bodyDir;
      com.command = TurtleCommand::DASH;

      INTERCEPTDRAW(sn->time, CIRCLE(meet.pos, BALL_SIZE));
      INTERCEPTDRAW(sn->time, CTEXT("dashing straight " << dlen << " time " << time));

      return com;
    }

    // now check one turn, if possible
    dpower = DASH_POWER_RATE*min(MAXPOWER, (STAMINA_INC_MAX + stamina/(time - 1)));
    len = reachRadius(dpower, time - 1);

    INTERCEPTDRAW(sn->time, CIRCLE(simMy.pos, len));

    Vector2d vl1, vl2;
    bool usetgt = false;
    switch (intersectCircleCircle(len, meet.pos - simMy.pos, error, vl1, vl2)) {
    case Valid: // goto into middle of solutions
      tgt = simMy.pos + 0.5*(vl1 + vl2);
      usetgt = true;
    case R1larger:
      // go straight into target
      usetgt = true;
    default:;
    }

    if (usetgt) {
      // check for angle not too large
      Angle dashdir = (tgt - simMy.pos).getAngle();
      TurnCommand tcm = turn(sn, dashdir);
      // works, go to tgt
      if (tcm.state != RANGLE)
	break;
    }

    // check two turns
    dpower = DASH_POWER_RATE*min(MAXPOWER, (STAMINA_INC_MAX + stamina/(time - 2)));
    len = reachRadius(dpower, time - 2);

    INTERCEPTDRAW(sn->time, CIRCLE(simMy.pos, len));

    usetgt = false;
    switch (intersectCircleCircle(len, meet.pos - simMy.pos, error, vl1, vl2)) {
    case Valid: // goto into middle of solutions
      tgt = simMy.pos + 0.5*(vl1 + vl2);
      usetgt = true;
    case R1larger:
      // go straight into target
      usetgt = true;
    default:;
    }

    if (usetgt)
      // ok, with two turns we get the ball, go!
      break;
  }

  INTERCEPTDRAW(sn->time, CIRCLE(meet.pos, BALL_SIZE));
  
  // if we fall here, tgt is where we should go directly
  Angle dashdir = (tgt - simMy.pos).getAngle();
  // test for turn delay since turn too small
  double tdist = norm(tgt - simMy.pos);
  if (fabs(dashdir - my.bodyDir) < goToThresh(error, tdist)) {
    com.bodyDir = my.bodyDir;
    com.v = my.v + radAng2Vector2d(dpower, my.bodyDir);
    com.time = time;
    com.command = TurtleCommand::DASH;

    INTERCEPTDRAW(sn->time, CTEXT("dashing, turn delayed, time= " << time << " power= " << dpower));

    return com;
  }
  TurnCommand tcm = turn(sn, dashdir);
  com.turnMoment = tcm.turnMoment;
  com.bodyDir = tcm.bodyDir;
  com.v = my.v;
  com.time = time;
  com.command = TurtleCommand::TURN;

  INTERCEPTDRAW(sn->time, LINE(simMy.pos, tgt));
  INTERCEPTDRAW(sn->time, CTEXT("turning, time= " << time  << " dashdir= " << dashdir));

  return com;
}

void AdvSkills::calcAcc(const Snapshot *sn, double stamina,
			double &minacc, double &maxacc)
{
  // ok, dash
  if (stamina < 0)
    stamina = 0;

  maxacc = min(stamina, MAXPOWER)*DASH_POWER_RATE;
  minacc = max(-0.5*stamina, MINPOWER)*DASH_POWER_RATE;

  double l1, l2;
  // max accelerations not exceeding PLAYER_SPEED_MAX
  switch (intersectLineCircle(radAng2Vector2d(1, sn->my.bodyDir),
			      -sn->my.v, PLAYER_SPEED_MAX, l1, l2)) {
  case Valid:
    if ((l1 > minacc) && (l1 <= maxacc))
      minacc = l1;
    if ((l2 < maxacc) && (l2 >= minacc))
      maxacc = l2;
    break;
  case None: // Currently too fast (???)
  case Instable: // 1 == 0 (????)
  default:;
  }
}

void AdvSkills::exec(const Snapshot *sn, const AdvSkills::TurtleCommand &com) {
  if (com.state == TFINISHED) {
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): called when finished" << endl;
    noAction();
    return;
  }
  switch (com.command) {
  case TurtleCommand::TURN:
    basicTurn(com.turnMoment);
    break;
  case TurtleCommand::DASH: {
    // determine sign of direction
    double pow = norm(com.v - sn->my.v)/DASH_POWER_RATE;
    if (com.state == TGOINGBACKWARDS)
      pow = -pow;
    basicDash(pow);
    break;
  }
  default:
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): called without valid command" << endl;
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): time = " << com.time << endl; 
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): com  = " << (int) com.command << endl; 
    cerr << sn->time << "AdvSkills::exec(TurtleCommand): comv = " << com.v << endl; 
    ensure(0);
    noAction();
  }
} 

AdvSkills::TurnHeadCommand
AdvSkills::turnHeadToBall(const Snapshot *sn, const TurtleCommand &tcm)
{
  switch(tcm.command) {
  case TurtleCommand::TURN:
    return turnHead(sn, (sn->ball.v + sn->ball.pos - sn->my.pos - tcm.v).getAngle() +
		    sn->my.bodyDir - tcm.bodyDir);
  default:
    return turnHead(sn, (sn->ball.v + sn->ball.pos - sn->my.pos - tcm.v).getAngle());
  }
}

AdvSkills::TurnHeadCommand
AdvSkills::turnHeadToBall(const Snapshot *sn, const TurnCommand &tcm)
{
  return turnHead(sn, (sn->ball.v + sn->ball.pos - sn->my.pos - sn->my.v).getAngle() +
		  sn->my.bodyDir - tcm.bodyDir);
}
