//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/************************************ kick *********************************/

// cwz-build-command: make all

/*
  KICK policy:
  KICKAWAY: always one kick
  FIXED: placeBall until ball kickable; then always kick.
  SUPERKICK: check if direct kick is ok, else lay back first. Then
             continue in FIXED mode.
 */

// flags for satisfied conditions in iplaceBall
enum { C_MA = 1, C_ST = 2, C_SA = 4, C_C1 = 8, C_C2 = 16 };

bool AdvSkills::kickSafe(const Snapshot *sn)
{
  return (sn->my.distTo(sn->ball) < kickRange);
}

AdvSkills::ExtKickCommand
AdvSkills::kick(const Snapshot *sn, double tgv, const Vector2d &tpos,
		int max_time)
{
  if (tgv < 0)
    tgv = 0;

  ExtKickCommand com;
  com.target   = tpos;
  com.target_v = tgv;
  com.max_time = max_time;

  return rekick(sn, com);
}

AdvSkills::ExtKickCommand
AdvSkills::kick(const Snapshot *sn, const Vector2d &tpos,
		int max_time)
{
  ExtKickCommand com;
  com.max_time = max_time;
  com.target   = tpos;
  double n = norm(com.target);
  if (n > BALL_SPEED_MAX)
    com.target.stretch(BALL_SPEED_MAX/n);
  com.target_v = -1;
  return rekick(sn, com);
}

AdvSkills::ExtKickCommand
AdvSkills::rekick(const Snapshot *sn, const ExtKickCommand &tgtcom)
{
  // for iplace without stopableSpeed check
  const int drop_order[] = { C_SA, C_C2, C_C1, C_MA};
  const int init_mask = C_SA|C_C2|C_C1|C_MA;

#ifdef KICKFIGS
  FigDraw drawing("kick", sn);
  drawing.Value("max_time", tgtcom.max_time);
#endif
  
  ExtKickCommand com;
  
  // setup continuation copies
  com.target   = tgtcom.target;
  com.target_v = tgtcom.target_v;
  
  // for position kicking, calculate velocity vector
  Vector2d target = tgtcom.target;
  if (tgtcom.target_v >= 0) {
    target -= sn->ball.pos;
    double nrm = norm(target);
    if (nrm) {
      double tgv = tgtcom.target_v + nrm*(1-BALL_DECAY);
      if (tgv > BALL_SPEED_MAX)
	tgv = BALL_SPEED_MAX;
      target.stretch(tgv/nrm);
    }
  }

  const WorldBall   &ball(sn->ball);
  const WorldPlayer &my(sn->my);

  if (my.distTo(ball) > KICKRANGE) {
#ifdef KICKFIGS
    drawing.Text("kick: RANGE");
#endif
    com.state = KRANGE;
    return com;
  }

#ifdef KICKFIGS
  drawing.Line(ball.pos - my.pos, target, 1);
#endif

  // kick
  double maxacc = my.kickPowerRate(ball)*MAXPOWER;

  // check one kick
  Vector2d k1;
  Vector2d diff = target - ball.v;
  double diffnrm = norm(diff);
  if (diffnrm < maxacc) {
    k1 = target;
  }
  else {
    // calculate one kick not correct, but optimal
    double l1, l2;
    switch(intersectLineCircle(target, ball.v, maxacc, l1, l2)) {
    case Valid:
      if (l1 > 1)
	k1 = l1*target;
      else
	k1 = l2*target;
      break;
    default:
      k1 = ball.v + (maxacc/diffnrm)*diff;
      break;
    }
  }
  // modify to not collide if necessary
  if (sqrnorm(ball.pos + k1 - my.pos) < sqr(innerSafeCircle)) {
    KickCommand pcm = iplaceBall(sn, ball.pos - my.pos + k1,
				 init_mask, drop_order);
    k1 = pcm.v;
  }
  double k1err = sqrnorm(k1 - target);

  // com keeps the one step less solution
  com.state = KTOOSHORT;
  com.v     = k1;
  com.time  = 1;
  com.max_time = 1;
#ifdef KICKFIGS
  drawing.Line(ball.pos - my.pos, com.v, 2);
#endif

  if ((tgtcom.max_time <= 1) ||
      (sqrnorm(com.v - target) < sqr(directThreshold)))
    return com;

  // two kicks. get ball as near as possible and in front
  Angle frontdir = (target + BALL_DECAY*(ball.pos - my.pos)).getAngle();
  // now ignore stopableSpeed (we want to kick, right?)
  KickCommand pcm2 =
    iplaceBall(sn, radAng2Vector2d(safeCircle, frontdir),
	       init_mask, drop_order);

  // check what kicking there results in
  WorldPlayer smy = my;
  WorldBall sball = ball;
  smy.simulate();
  sball.v = pcm2.v;
  sball.simulate();
  double maxacc2 = smy.kickPowerRate(sball)*MAXPOWER;

  diff    = target - sball.v;
  diffnrm = norm(diff);
  if (diffnrm < maxacc2) {
    // after two kicks, we are finished
    com.time  = 2;
    com.max_time = 2;
    com.state = KPLACING;
    com.v = pcm2.v;
#ifdef KICKFIGS
    drawing.Text("2 kicks opt");
    drawing.Line(ball.pos - my.pos, com.v, 2);
#endif
    return com;
  }

  Vector2d k2 = sball.v + (maxacc2/diffnrm)*diff;
  double k2err = sqrnorm(k2 - target);

  if (k1err > k2err - directThreshold) {
    // 2 kicks better, update com
    com.state = KPLACING;
    com.time  = 2;
    com.max_time = 2;
    com.v     = pcm2.v;
#ifdef KICKFIGS
    drawing.Line(ball.pos - my.pos, com.v, 3);
    drawing.Line(ball.pos - my.pos, k2, 4);
#endif
  }
  
  if (tgtcom.max_time == 2)
    return com;
  
  // here we know that 2 steps are not enough; use three steps
  // place ball in opposite frontdir if too fast for now or not there
  Vector2d laybackpos = radAng2Vector2d(innerSafeCircle, PI + target.getAngle());
  // here be careful about stopableSpeed since somewhat laying back
  KickCommand pcm3 = placeBall(sn, laybackpos);
  // only place if kick is really doing something
  if (sqrnorm(pcm3.v - ball.v) > minSpeed) {
    com.time  = tgtcom.max_time - 1; // take a little time for placing
    com.max_time  = tgtcom.max_time - 1;
    com.state = KPLACING;
    com.v     = pcm3.v;
#ifdef KICKFIGS
    drawing.Text("3 kicks");
    drawing.Line(ball.pos - my.pos, com.v, 5);
#endif
    return com;
  }

  // return old 2 step command
  return com;
}

AdvSkills::KickCommand
AdvSkills::stopBall(const Snapshot *sn)
{
  KickCommand com;
  // stop the ball
  if (sn->my.distTo(sn->ball) > KICKRANGE) {
    com.state = KRANGE;
    return com;
  }

  double ns = norm(sn->ball.v);
  double kickrate = sn->my.kickPowerRate(sn->ball);
  com.v = (ns > MAXPOWER*kickrate) ? (1 - MAXPOWER*kickrate/ns)*sn->ball.v : Vector2d(0, 0);
  com.state = KSTOPPING;
  return com;
}

///////////////////////////////////////////// placeBall stuff starts here

#ifdef KICKFIGS
void printFlag(FigDraw &d, const char *txt, int f1, int f2)
{
  char buf[256];
  strcpy(buf, txt);
  strcat(buf, ": (|");
  if (f1 & C_MA)
    strcat(buf, "MA|");
  if (f1 & C_SA)
    strcat(buf, "SA|");
  if (f1 & C_C1)
    strcat(buf, "C1|");
  if (f1 & C_C2)
    strcat(buf, "C2|");
  if (f1 & C_ST)
    strcat(buf, "ST|");
  strcat(buf, ", ");
  if (f2 & C_MA)
    strcat(buf, "MA|");
  if (f2 & C_SA)
    strcat(buf, "SA|");
  if (f2 & C_C1)
    strcat(buf, "C1|");
  if (f2 & C_C2)
    strcat(buf, "C2|");
  if (f2 & C_ST)
    strcat(buf, "ST|");
  strcat(buf, ")");
  d.Text(buf);
}
#endif

AdvSkills::KickCommand
AdvSkills::placeBall(const Snapshot *sn, const Vector2d &targ)
{
  // this is the order in which conditions are dropped if necessary
  // after dropping the last all points are valid, and tgt is always aviable
  const int drop_order[] = { C_ST, C_SA, C_C1, C_C2, C_MA};
  return iplaceBall(sn, targ, C_ST|C_SA|C_C1|C_C2|C_MA, drop_order);
}
 
AdvSkills::KickCommand
AdvSkills::iplaceBall(const Snapshot *sn, const Vector2d &targ, int cond_mask, const int *drop_order)
{
  KickCommand com;
  if (sn->my.distTo(sn->ball) > KICKRANGE) {
    com.state = KRANGE;
    return com;
  }

  // ok, the one or other way we place the ball
  com.state = KPLACING;

#ifdef KICKFIGS
  FigDraw drawing("place", sn);
#endif

  const WorldPlayer &my(sn->my);
  const WorldBall &ball(sn->ball);
  const Vector2d ptb = ball.pos - my.pos;
  const Vector2d ptbns = ptb + ball.v;
  const Vector2d pnstb = ptb - my.v;
  const Vector2d pnstbns = pnstb + ball.v;
  const double kickrate = my.kickPowerRate(ball);
  const double maxacc = kickrate*MAXPOWER;
  const Vector2d vt = targ - pnstb;

#ifdef KICKFIGS
  drawing.Line(ptb, vt, 2);
  drawing.Circle(ptb, stopableV, 0);
  drawing.Circle(ptb + ball.v, maxacc, 0);
  drawing.Circle(my.v, safeCircle, 0);
  drawing.Circle(Vector2d(0,0), innerSafeCircle, 0);
#endif

  // ok, now we have to find the best approximation.
  // Simply calculate all critical points and test them until one good is found
  // often needed test constants for comparison
  double sinns = sqr(innerSafeCircle) - precision; // compared down
  double smaxa = sqr(maxacc) + precision;
  double ssafe = sqr(safeCircle) + precision;
  double sstop = sqr(stopableV) + precision;

  // intersecting pts of circles ma(xacc), st(opable), sa(fe) and c1/c2(collrange)
  // no c2sa since collRange < safeCircle (better is this)
  const int mast(0), mac1(2), mac2(4), masa(6), stsa(8), stc1(10), stc2(12), c1sa(14),
    c1c2(16);
  // and the nonconstrained critical points (including free)
  const int ma(18), st(19), c1(20), c2(21), sa(22), tgt(23), numPoints(24);
  Vector2d critpoint[numPoints];
  // cond_mask gives all conditions that have to be satisfied
  int flag[numPoints];
  // later some may be dropped

  for (int i = 0; i < numPoints; i++)
    flag[i] = 0;

  // first set tgt to vt and set flags (for evtl. dropping conditions)
  // just to avoid errors from copy and paste...
  int curPnt = tgt;
  critpoint[curPnt] = vt;
  if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
    flag[curPnt] |= C_MA;
  if (sqrnorm(critpoint[curPnt]) <= sstop)
    flag[curPnt] |= C_ST;
  if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
    flag[curPnt] |= C_SA;
  if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
    flag[curPnt] |= C_C1;
  if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
    flag[curPnt] |= C_C2;
#ifdef KICKFIGS
  printFlag(drawing, "tgt", flag[curPnt], 0);
#endif

  if (flag[curPnt] == cond_mask) {
    // short cut, hopefully often used
    com.v = critpoint[curPnt];
    return com;
  }

  // calc ma/st intersecting points
  curPnt = mast;
  if (intersectCircleCircle(stopableV, ball.v, maxacc,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    flag[curPnt] = flag[curPnt + 1] = C_MA|C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb - critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "mast", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc ma/c1 intersecting points
  curPnt = mac1;
  if (intersectCircleCircle(innerSafeCircle, ptbns, maxacc,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_MA|C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "mac1", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc ma/c2 intersecting points
  curPnt = mac2;
  if (intersectCircleCircle(innerSafeCircle, pnstbns, maxacc,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= pnstb;
    critpoint[curPnt + 1] -= pnstb;

    flag[curPnt] = flag[curPnt + 1] = C_MA|C_C2;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "mac2", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc ma/sa intersecting points
  curPnt = masa;
  if (intersectCircleCircle(safeCircle, pnstbns, maxacc,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= pnstb;
    critpoint[curPnt + 1] -= pnstb;
    
    flag[curPnt] = flag[curPnt + 1] = C_MA|C_SA;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;
    
    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "flags masa", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc st/sa intersecting points
  curPnt = stsa;
  if (intersectCircleCircle(safeCircle, pnstb, stopableV,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= pnstb;
    critpoint[curPnt + 1] -= pnstb;

    flag[curPnt] = flag[curPnt + 1] = C_ST|C_SA;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;

    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "flags stsa", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc st/c1 intersecting points
  curPnt = stc1;
  if (intersectCircleCircle(innerSafeCircle, ptb, stopableV,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_ST|C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;
 
    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "flags stc1", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc st/c2 intersecting points
  curPnt = stc2;
  if (intersectCircleCircle(innerSafeCircle, pnstb, stopableV,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= pnstb;
    critpoint[curPnt + 1] -= pnstb;

    flag[curPnt] = flag[curPnt + 1] = C_ST|C_C2;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;
 
    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "flags stc2", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc c1/sa intersecting points
  curPnt = c1sa;
  if (intersectCircleCircle(innerSafeCircle, my.v, safeCircle,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_C1|C_SA;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "flags stc1", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc c1/sa intersecting points
  curPnt = c1sa;
  if (intersectCircleCircle(innerSafeCircle, my.v, safeCircle,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_C1|C_SA;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "flags stc1", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc c1/c2 intersecting points
  curPnt = c1c2;
  if (intersectCircleCircle(innerSafeCircle, my.v, innerSafeCircle,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_C1|C_C2;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt    ], 3);
    drawing.Line(ptb, critpoint[curPnt + 1], 3);
    printFlag(drawing, "flags c1c2", flag[curPnt], flag[curPnt + 1]);
#endif
  }

  // calc projection onto ma-circle
  curPnt = ma;
  Vector2d proj = vt - ball.v;
  double nrmprj = norm(proj);
  if (nrmprj > maxacc) {
    flag[curPnt] = C_MA;
    critpoint[curPnt] = ball.v + (maxacc/nrmprj)*proj;

    if (sqrnorm(critpoint[curPnt]) <= sstop)
      flag[curPnt] |= C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
      flag[curPnt] |= C_SA;

    if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C2;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt], 4);
    printFlag(drawing, "flag ma", flag[curPnt], 0);
#endif
  }

  // calc projection onto co-circle
  curPnt = c1;
  proj = ptb + vt;
  nrmprj = norm(proj);
  if ((nrmprj < innerSafeCircle) && (nrmprj > precision)) {
    flag[curPnt] = C_C1;
    critpoint[curPnt] = (innerSafeCircle/nrmprj)*proj - ptb;

    if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
      flag[curPnt] |= C_MA;

    if (sqrnorm(critpoint[curPnt]) <= sstop)
      flag[curPnt] |= C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
      flag[curPnt] |= C_SA;

    if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C2;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt], 4);
    printFlag(drawing, "flag c1", flag[curPnt], 0);
#endif
  }

  // calc projection onto co-circle
  curPnt = c2;
  proj = pnstb + vt;
  nrmprj = norm(proj);
  if ((nrmprj < innerSafeCircle) && (nrmprj > precision)) {
    flag[curPnt] = C_C2;
    critpoint[curPnt] = (innerSafeCircle/nrmprj)*proj - pnstb;

    if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
      flag[curPnt] |= C_MA;

    if (sqrnorm(critpoint[curPnt]) <= sstop)
      flag[curPnt] |= C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
      flag[curPnt] |= C_SA;

    if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C1;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt], 4);
    printFlag(drawing, "flag c2", flag[curPnt], 0);
#endif
  }

  // calc projection onto sa-circle
  curPnt = sa;
  proj = pnstb + vt;
  nrmprj = norm(proj);
  if ((nrmprj < safeCircle - precision) ||
      (nrmprj > safeCircle + precision)) {
    flag[curPnt] = C_SA;
    critpoint[curPnt] = (safeCircle/nrmprj)*proj - pnstb;

    if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
      flag[curPnt] |= C_MA;

    if (sqrnorm(critpoint[curPnt]) <= sstop)
      flag[curPnt] |= C_ST;

    if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C2;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt], 4);
    printFlag(drawing, "flag sa", flag[curPnt], 0);
#endif
  }

  // calc projection onto st-circle
  curPnt = st;
  nrmprj = norm(vt);
  if (nrmprj > stopableV) {
    flag[curPnt] = C_ST;
    critpoint[curPnt] = (stopableV/nrmprj)*vt;

    if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
      flag[curPnt] |= C_MA;

    if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C2;

    if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
      flag[curPnt] |= C_SA;

#ifdef KICKFIGS
    drawing.Line(ptb, critpoint[curPnt], 4);
    printFlag(drawing, "flag st", flag[curPnt], 0);
#endif
  }

  // and now search for a valid solution
  int drop = 0;
  while (cond_mask != 0) {
#ifdef KICKFIGS
    drawing.Value("mask is", cond_mask);
#endif
    // optimize over all points matching condition
    double min = 1e30;
    int    mindex = -1;
    for (int i = 0; i < numPoints; i++) {
      if ((flag[i] & cond_mask) == cond_mask) {
	double dist = sqrnorm(critpoint[i] - vt); 
	if (dist < min) {
	  min = dist;
	  mindex = i;
	}
      }
    }
    if (mindex >= 0) {
      // got a point
#ifdef KICKFIGS
      drawing.Line(ptb, critpoint[mindex], 11);
      char buf[256];
      sprintf(buf, "final is %d", mindex);
      printFlag(drawing, buf, flag[mindex], 0);
#endif
      com.v = critpoint[mindex];
      return com;
    }
    // drop next condition
    cond_mask &= ~drop_order[drop++];
  }
  // never reached
  cerr << "AdvSkills::placeBall: didn't find a point to place!" << endl;
  // at least a valid kick command...
  com.v = flag[ma] ? critpoint[ma] : vt;
  return com;
}

void AdvSkills::exec(const Snapshot *sn, const AdvSkills::KickCommand &com) {
  if (com.state == KRANGE) {
    cerr << sn->time << " AdvSkills::exec(KickCommand): ball too far away to kick!" << endl;
    noAction();
    return;
  }
  Vector2d kick_v = com.v - sn->ball.v;
  double res_power = norm(kick_v) / sn->my.kickPowerRate(sn->ball);
  Angle res_angle = kick_v.getAngle() - sn->my.bodyDir;
  // renormation necessary for precision reasons
  if (res_power > MAXPOWER)
    res_power = MAXPOWER;
  basicKick(res_power, res_angle);
}
