//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "Polygon.h"
#include <stdio.h>

#define EPS 1e-8  // Numbers between -EPS and EPS are treated as 0

// create a polygon out of a list of points
Polygon::Polygon( const Vector2d v[], const int len = 0 ){
  if ((len > 2) && &v){
    p = new Vector2d[n = len];
    for (int i = 0; i < n; i++) 
      p[i] = v[i];
  }
  else {
   if (p){
     delete p;
   }
   p = NULL;
   n = 0;
   cerr << "Polygon::Polygon " << len << " are to less points for a polygon";
  }
}

// copy a polygon
Polygon::Polygon( const Polygon &P ){
  if ((n = P.nop()) > 0) {
    p = P.points();
  }
  else {
    n = 0;
    p = NULL;
  }
} 

// create an empty polygon
Polygon::Polygon(){
  p = NULL;
  n = 0;
}

// clean up the polygon
Polygon::~Polygon(){
  if (p)
    delete p;
}

// return the number of points the polygon has
int Polygon::nop() const{
  return n;
}

// adds a point to the polygon at the end of p
// if it is not equal to the last or the first
int Polygon::addPoint( const Vector2d c ) {
  if (!n || (sqrnorm(p[n-1]-c) > EPS*EPS &&
      sqrnorm(p[0]-c) > EPS*EPS)) {
    Vector2d *help = new Vector2d[n+1];
    for (int i = 0; i < n; i++)
      help[i] = p[i];
    help[n] = c;
    if (p)
      delete p;
    p = help;
    n++;
    return 1;
  }
  return 0;
}

// deletes point no out of the points array p
int Polygon::delPoint( const int no ) {
  if (0 <= no && no < n) {
    Vector2d *help = new Vector2d[n-1];
    for (int i = 0; i < no; i++) 
      help[i] = p[i];
    for (int k = no; k < n-1; k++) 
      help[k] = p[k+1];
    if (p)
      delete p;
    p = help;
    n--;
    return 1;
  }
  return 0;
}

// calculates a point inside the polygon
Vector2d Polygon::getInside() const {
  Vector2d m(0,0);
  for (int i = 0 ; i < n; i++)
    m += p[i];
  if (n)
    m *= 1.0/n;
  return m;
}

// insert a point to the polygon
// the point will be inserted according to its angle
// (the points are sorted against the clock
//  which is possible as the polygon is konvex)
int Polygon::insertPoint( const Vector2d c ){
  // the new points array
  Vector2d *help = new Vector2d[n+1];
  // a point inside
  Vector2d m = getInside();

  int i = 0, offset = 0;
  Angle angle = M_PI;  // biggest angle according to libMRB/coordinates.h

  // the points may start with the wrong angle
  // so shift 'em as they are sorted
  Vector2d *shifted = new Vector2d[n];

  // find smallest angle
  for (int j = 0; j < n; j++) {
    if (p[j].getAngle() < angle) {
      angle = p[j].getAngle();
      offset = j;
    }
  }

  // create shifted array
  for (int j = 0; j < n; j++)
    shifted[j] = p[(j+offset)%n];

  if (p)
    delete p;
  p = shifted;
  
  // looking for a place to insert the new point
  while (i < n && Vector2d(p[i]-m).getAngle() < Vector2d(c-m).getAngle())
    help[i] = p[i++];

  if (!n){ // if there is no point yet, add the new one
    help[i++] = c;
    if (p)
      delete p;
    p = help;
    n++;
    return 1;
  }
  else {  // add the new one if it is a different point
     if (!(sqrnorm(p[i] - c) < EPS*EPS)){ 
       help[i++] = c;
       for (int k = i; k <= n; k++)
	 help[k] = p[k-1];
       if (p)
	 delete p;
       p = help;
       n++;
       return 2;
     }
     else {
       delete help;
       return 0;
     }
   }
}

// delete all polygon points
void Polygon::clear() {
  if (p)
    delete p;
  p = NULL;
  n = 0;
}

// return a pointer to an array containig a copy of the polygons points
Vector2d *Polygon::points() const{
  if (n) {
    Vector2d *c = new Vector2d[n];
    for (int i = 0; i < n; i++)
      c[i] = p[i];
    return c;
  }
  return NULL;
}

// checks if v exists
bool Polygon::exists( const Vector2d v ) {
  for (int i = 0; i < n; i++) {
    if (sqrnorm(p[i]-v) <= EPS*EPS) {
      return true;
    }
  }
  return false;
}

Polygon Polygon::operator=( const Polygon &P ) {
  if (p)
    delete p;
  p = P.points();
  n = P.nop();
  return *this;
}

// calculate the intersection of two KONVEX polygons
// ^= is much faster!
Polygon Polygon::operator^( const Polygon &P ) const {
  Polygon interP = Polygon(p, n);
  Vector2d  *p1 = points(),
    *p2 = P.points();
  int pcounter = 0,  // points counter
    scounter = 0;  // sides counter
  char inside[1024] = {0};

  if (n >= 1024) 
    cerr << "Polygon::operator^(): To many points use opeartor^=() instead";

  for (int j = 0; j < P.nop(); j++) {
    for (int i = 0; i < interP.nop(); i++) { // counting points inside the half-plaine;
      if (((p1[i]-p2[j])*(p2[(j+1)%P.nop()]-p2[j]).perp()) >= -EPS) {
	pcounter++;
	inside[i] = 1;     // marking the points inside
      }
    }

    if (!pcounter) {  // if no points are inside, the intersection is empty
      interP.clear();
      return interP;
    }

    int n2 = interP.nop();
    if (pcounter) {  // if not all points inside
      interP.clear();
      scounter++;
      pcounter = 0;
      for (int i = 0; i < n2; i++) { // get points inside half-plaine
	if (inside[i]) {
	  interP.addPoint(p1[i]);
	}
	if ((!inside[i] && inside[(i+1)%n2]) || (inside[i] && !inside[(i+1)%n2])) {
	  // calculate intersection of two lines given by 
	  // a + mu*(b-a)            
	  // c + lambda*(d-c)   
	  // <=>  (b-a, c-d)*(mu, lambda)^T=(c-a)
	  // where a,b,c,d in |R^2
	  double detA, mu;
	  detA = (p1[(i+1)%n2].getX()-p1[i].getX())*
	    (p2[j].getY()-p2[(j+1)%P.nop()].getY())-
	    (p1[(i+1)%n2].getY()-p1[i].getY())*
	    (p2[j].getX()-p2[(j+1)%P.nop()].getX());
	  if (detA) {  // when those lines are not parallel
	    mu = 1/detA*((p2[j].getY()-p2[(j+1)%P.nop()].getY())*
			 (p2[j].getX()-p1[i].getX())+
			 (p2[(j+1)%P.nop()].getX()-p2[j].getX())*
			 (p2[j].getY()-p1[i].getY()));
	    // the instersection belongs to the intersectionpolygon...
	    if (0<= mu && mu <= 1) {
	      interP.addPoint(p1[i]+mu*(p1[(i+1)%n2]-p1[i]));
	    }
	  }
	}
      }
    }
    if (p1)
      delete p1;
    p1 = interP.points();
    for (int i = 0; i < n2; i++) {
      inside[i] = 0;
    }
  }
  if (!scounter)  // true if all points where inside all half-plaines
    interP = Polygon(p,n);
  if (p1)
    delete p1;
  if (p2)
    delete p2;
  return interP;
}

// calculate the intersection of two KONVEX polygons
Polygon Polygon::operator^=( const Polygon &P ) {
  Vector2d *p2 = P.points();

  for (int j = 0; j < P.nop(); j++) {
    for (int i = 0; i < n; i++) { // keeping the points inside the half-plaine;
      
      Vector2d plainNormal = (p2[(j+1)%P.nop()]-p2[j]).perp();
      // if this point (p[i]) is inside and the next point is not
      if ((p[i]-p2[j])*plainNormal >= -EPS && 
	  (p[(i+1)%n]-p2[j])*plainNormal < -EPS) {
	// calculate intersection of two lines given by 
	// a + mu*(b-a)            
	// c + lambda*(d-c)   
	// <=>  (b-a, c-d)*(mu, lambda)^T=(c-a)
	// where a,b,c,d in |R^2
	double detA, mu;
	detA = (p[(i+1)%n].getX()-p[i].getX())*
	  (p2[j].getY()-p2[(j+1)%P.nop()].getY())-
	  (p[(i+1)%n].getY()-p[i].getY())*
	  (p2[j].getX()-p2[(j+1)%P.nop()].getX());
	if (detA) {
	  mu = 1/detA*((p2[j].getY()-p2[(j+1)%P.nop()].getY())*
		       (p2[j].getX()-p[i].getX())+
		       (p2[(j+1)%P.nop()].getX()-p2[j].getX())*
		       (p2[j].getY()-p[i].getY()));
	  Vector2d cut1 = p[i]+mu*(p[(i+1)%n]-p[i]);
	// if the point after the next point is outside too
	  if ((p[(i+2)%n]-p2[j])*plainNormal < -EPS) {
	    if (sqrnorm(p[i]-cut1) > EPS*EPS) {// if the intersection does not exist
	      p[(i+1)%n] = cut1;  // replace the next one by the intersection
	      i++; // and continue with the point after the next
	      continue;
	    }
	    else {  // if the intersection exists forget it
	      delPoint((i+1)%n);
	      continue;
	    }
	  }
	  else {
	    // if the point after the next point (p[i+2]) is inside 	  
	    detA = (p[(i+2)%n].getX()-p[(i+1)%n].getX())*
	      (p2[j].getY()-p2[(j+1)%P.nop()].getY())-
	      (p[(i+2)%n].getY()-p[(i+1)%n].getY())*
	      (p2[j].getX()-p2[(j+1)%P.nop()].getX());
	    if (detA) {
	      mu  = 1/detA*((p2[j].getY()-p2[(j+1)%P.nop()].getY())*
			    (p2[j].getX()-p[(i+1)%n].getX())+
			    (p2[(j+1)%P.nop()].getX()-p2[j].getX())*
			    (p2[j].getY()-p[(i+1)%n].getY()));
	      Vector2d cut2 = p[(i+1)%n]+mu*(p[(i+2)%n]-p[(i+1)%n]);
	      if (sqrnorm(p[i]-cut1) > EPS*EPS) {
		p[(i+1)%n] = cut1;  // replace the next one by the intersection
		// if the new one (between p[i+1] and p[i+2]) does not exist
		if (sqrnorm(p[i]-cut2) >EPS*EPS &&
		    sqrnorm(p[(i+2)%n]-cut2) >EPS*EPS) {
		  Vector2d *help = new Vector2d[n+1];
		  int k;
		  for (k = 0; k <= (i+1)%n; k++)
		    help[k] = p[k];
		  help[k] = cut2;  // add it after the new next one
		  for (int l = ++k; l <= n; l++)
		    help[l] = p[l-1];
		  n++;
		  i++;  // and continue after the newest
		  if (p)
		    delete p;
		  p = help;
		}
	      }
	      else {
		// if the new one does not exist
		if (sqrnorm(p[i]-cut2) >EPS*EPS &&
		    sqrnorm(p[(i+2)%n]-cut2) >EPS*EPS) {
		  p[(i+1)%n] = cut2;  // replace the next one
		  i++;
		}
		else {
		  delPoint((i+1)%n);
		}
	      }
	    }
	  }
	  continue;
	}	
      }

      // if this point is outside and the next is inside
      if ((p[i]-p2[j])*plainNormal < -EPS &&
	  (p[(i+1)%n]-p2[j])*plainNormal >= -EPS) {
	double detA, mu;
	detA = (p[(i+1)%n].getX()-p[i].getX())*
	  (p2[j].getY()-p2[(j+1)%P.nop()].getY())-
	  (p[(i+1)%n].getY()-p[i].getY())*
	  (p2[j].getX()-p2[(j+1)%P.nop()].getX());
	if (detA) {
	  mu = 1/detA*((p2[j].getY()-p2[(j+1)%P.nop()].getY())*
		       (p2[j].getX()-p[i].getX())+
		       (p2[(j+1)%P.nop()].getX()-p2[j].getX())*
		       (p2[j].getY()-p[i].getY()));
	  // if the point before is inside too at the beginning
	  if ((p[n-1]-p2[j])*plainNormal >= -EPS && !i) {
	    detA = (p[0].getX()-p[n-1].getX())*
	      (p2[j].getY()-p2[(j+1)%P.nop()].getY())-
	      (p[0].getY()-p[n-1].getY())*
	      (p2[j].getX()-p2[(j+1)%P.nop()].getX());
	    if (detA) {
	      double lambda = 1/detA*((p2[j].getY()-p2[(j+1)%P.nop()].getY())*
				      (p2[j].getX()-p[n-1].getX())+
				      (p2[(j+1)%P.nop()].getX()-p2[j].getX())*
				      (p2[j].getY()-p[n-1].getY()));
	      addPoint(p[n-1]+lambda*(p[0]-p[n-1]));  // add the intersection
	    }
	  }
	  Vector2d cut = p[i]+mu*(p[(i+1)%n]-p[i]);
	  // if the intersection does not exist
 	  if (sqrnorm(p[(i+1)%n]-cut) > EPS*EPS) {
 	    p[i] = cut; // replace this point by the intersection
 	  }
 	  else {
	    delPoint(i); // else forget the point outside
	  }
	}
	continue;
      }

      // if this point is outside and the point before too at the beginning
      // or the point is outside
      if (((p[0]-p2[j])*plainNormal < -EPS && 
	  (p[n-1]-p2[j])*plainNormal < -EPS && !i) ||
	  ((p[i]-p2[j])*plainNormal < -EPS && i)) {
	delPoint(i);
      }
    }
  }

  delete p2;  // this one is no longer needed
  return *this;  // that's what we wanted to get
}
