//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// $Id: audiomsg.C,v 1.16 2000/10/19 15:27:43 goetz Exp $
//
//////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdio.h>
#include <string.h>

#include "constants.h"
#include "coordinates.h"

#include "Ensure.h"
#include "audiomsg.h"

/* characters allowed by server: table for encoding */
const int  VALID_MSG_CHARS   =64;
const char msgEncodeChars[VALID_MSG_CHARS+1] =  \
               "abcdefghijklmnopqrstuvwxyz" \
               "ABCDEFGHIJKLMNOPQRSTUVWXYZ" \
               "0123456789" \
               "+-";

/* characters allowed by server: table for decoding, has to be initialized
 by initializeTable() first */
static char msgDecodeChars[256];
static bool tableInitialized= false;


/** initializes the table for converting server characters */
void initializeTable() 
{    
  /* fill it with inverse of above map (ie table) */
  for (int i= 0; i < 256; i++)
    msgDecodeChars[i]= 0;
  for (int i= 0; i < VALID_MSG_CHARS; i++)
    msgDecodeChars[msgEncodeChars[i]]= i;
  tableInitialized= true;
}

//////////////////////////////////////////////////////////////////////
//
// class Int3B
//
//////////////////////////////////////////////////////////////////////

Int3B::Int3B () : value(0) 
{
  // create 0 value
  encoding[0]= encoding[1]= encoding[2]= msgEncodeChars[0];
}

Int3B::Int3B (const char* encoded)
{
 // copy it into our local element
  for (int i=0; i< 3; i++)
    encoding[i]= encoded[i];
  decode();
}

Int3B::Int3B (const unsigned int &val) : value(val)
{
  encode();
}

void Int3B::encode()
{
  encoding[0]= msgEncodeChars[value & 63];
  encoding[1]= msgEncodeChars[(value >> 6) & 63];
  encoding[2]= msgEncodeChars[(value >> 12) & 63];
} // Int3B::encode()

void Int3B::decode()
{
  unsigned int c0, c1, c2;

  // if not already done, initialize table!
  if (!tableInitialized)
    initializeTable(); 
  
  c0= msgDecodeChars[encoding[0]];
  c1= msgDecodeChars[encoding[1]];
  c2= msgDecodeChars[encoding[2]];

  value= c0 | (c1 << 6) | (c2 << 12);
} // Int3B::decode()

ostream &operator << (class ostream &os, const Int3B &i)
{
  return os << "(Int3B " << i.getValue() << ")";
}

Int3B::operator int()
{
  return value;
}

//////////////////////////////////////////////////////////////////////
//
// class Val5B
//
//////////////////////////////////////////////////////////////////////

Val5B::Val5B () : x(0), y(0), age(0), state(UNSAFE)
{
  for (int i=0; i < 5; i++)
    encoding[i]= 0;
}

Val5B::Val5B (const char* encoded)
{
  for (int i=0; i < 5; i++)
    encoding[i]= encoded[i];
  decode();
}

Val5B::Val5B (const double x_, const double y_, const unsigned int age_,
	      const WorldObjectState state_) : 
  x(x_), y(y_), age(age_), state(state_)
{
  encode();
}

Val5B::Val5B(const Vector2d c, const unsigned int age_,
	     const WorldObjectState state_) :
  x(c.getX()), y(c.getY()), age(age_), state(state_)
{
  encode();
}

void Val5B::encode()
{
  unsigned char c0, c1, c2, c3, c4;
  int xr= int(floor((x*10.0) + 1024)), yr=int(floor((y*10.0) + 1024));

  // are values valid (otherwise set valid to false
  if ( (xr > 2047) || (yr > 2047) || (age > 63) || (xr < 0) || (yr < 0) ||
       (state < UNSAFE) || (state > AUDIAL)) 
    age= INVALIDAGE;

  // now remove the sign of the rounded values
  unsigned int x_(xr), y_(yr), st1, st2;
  
  switch (state) {
  case FUZZY:  st1= 1 << 5; st2= 0;      break;
  case VISUAL: st1= 0;      st2= 1 << 5; break;
  case AUDIAL: st1= 1 << 5; st2= 1 << 5; break;
  default:     st1= 0;      st2= 0;
  }
    

  c0= x_ & 63;
  c1= ((x_ >> 6) & 31) | st1;
  c2= y_ & 63;
  c3= ((y_ >> 6) & 31) | st2;
  c4= age & 63;

  encoding[0]= msgEncodeChars[c0];
  encoding[1]= msgEncodeChars[c1];
  encoding[2]= msgEncodeChars[c2];
  encoding[3]= msgEncodeChars[c3];
  encoding[4]= msgEncodeChars[c4];
}

void Val5B::decode()
{
  unsigned int c0, c1, c2, c3, c4;
  // if not already done, initialize table!
  if (!tableInitialized)
    initializeTable(); 
  
  c0= msgDecodeChars[encoding[0]];
  c1= msgDecodeChars[encoding[1]];
  c2= msgDecodeChars[encoding[2]];
  c3= msgDecodeChars[encoding[3]];
  c4= msgDecodeChars[encoding[4]];

  x= ( (c0 | ((c1 & 31) << 6)) - 1024.0) / 10.0;
  y= ( (c2 | ((c3 & 31) << 6)) - 1024.0) / 10.0;
  age= c4;

  // decode the state
  if ( c1 >> 5 != 0)
    if ( c3 >> 5 != 0)
      state= AUDIAL;
    else 
      state= FUZZY;
  else if ( c3 >> 5 != 0)
    state= VISUAL;
  else
    state= UNSAFE;
}

ostream & operator << (class ostream &os, const Val5B &v)
{
  if (v.isValid()) {
    os << "(Val5B (" << v.getX() << "," << v.getY() << ")(age " << 
      v.getAge() << ")(state ";
    switch (v.getState()) {
    case AUDIAL: os << "AUDIAL"; break;
    case VISUAL: os << "VISUAL"; break;
    case FUZZY: os << "FUZZY"; break;
    default: os << "UNSAFE"; break;
    }
    return os << "))";
  } else
    return os << "(Val5B invalid)";
}

Val5B::operator Vector2d()
{
  return Vector2d(x, y);
}

//////////////////////////////////////////////////////////////////////
//
// NodaChar: Holds one 6-bit char
//
//////////////////////////////////////////////////////////////////////

NodaChar::NodaChar(const unsigned char val_) 
{
  val= val_;
  if (val > 63) val=0;
}

NodaChar::operator char() const
{
  return char(val);
}

ostream &operator << (class ostream &os, const NodaChar &c) 
{
  return os << char(c);
}


//////////////////////////////////////////////////////////////////////
//
// internal format of AudioMsg
//
//////////////////////////////////////////////////////////////////////

const unsigned int AUMSG_HEADER=0;
/* The header looks like:
 * 0 1 2 3 4 5 6 7 8 9 10 11 12 13 
 * N N N S D C C T T T R  R  W  W 
 *
 * N - Name (Id of team)
 * S - src Player (coach=0) and Side
 * D - dest Player and iscoachmsg (shorter length)
 * C - Checksums
 * T - Server-Time for msg
 * R - reader - pointer of stream
 * W - writer - pointer of stream
 */

const unsigned int AUMSG_STREAMDATA=14;


//////////////////////////////////////////////////////////////////////
//
// class AudioMsg
//
//////////////////////////////////////////////////////////////////////

AudioMsg::AudioMsg () : state( true ), MSG_LENGTH(SAY_MSG_SIZE),
  readMark( 0 ), writeMark( 0 )
{
  for (unsigned int i=0; i<AUMSG_STREAMDATA; i++)
    encoded[i]= msgEncodeChars[0];
  encoded[AUMSG_STREAMDATA]= 0;

  if (!tableInitialized)
    initializeTable();
  
  createChkSums();
}

AudioMsg::AudioMsg(const unsigned int destno,
		   const unsigned int stime,
		   const bool iscoachmsg) 
  : state( true ), 
    MSG_LENGTH( iscoachmsg ? SAY_COACH_MSG_SIZE : SAY_MSG_SIZE), 
    readMark( 0 ), writeMark( 0 )
{
  for (unsigned int i=0; i<AUMSG_STREAMDATA; i++)
    encoded[i]= msgEncodeChars[0];
  encoded[AUMSG_STREAMDATA]= 0;

  if (!tableInitialized)
    initializeTable();

  // set destination and msg type directly
  encoded[4]= msgEncodeChars[destno | ((iscoachmsg)? 16 : 0) ];
  
  setServerTime(stime);

  createChkSums();
}

AudioMsg::AudioMsg( const char *encoding)
   : state( true ), readMark( 0 ), writeMark( 0 )
{
  if (!tableInitialized)
    initializeTable();
  
  strcpy(encoded, encoding);

  if (isCoachMsg()) {
    MSG_LENGTH= SAY_COACH_MSG_SIZE;
  } else {
    MSG_LENGTH= SAY_MSG_SIZE;
  }
}
 
bool AudioMsg::isCoachMsg() const
{
  return (msgDecodeChars[encoded[4]] & 16) != 0;
}
  
AudioMsg& AudioMsg::setPrefix( const char prefix[4])
{
  encoded[0]= prefix[0];
  encoded[1]= prefix[1];
  encoded[2]= prefix[2];
  
  return *this;
}
  
AudioMsg& AudioMsg::setSourceAndSideIsLeft( const unsigned int playerno, 
					    const bool isLeft)
{
  ensure( playerno <= 11);
  encoded[3]= msgEncodeChars[playerno | ((isLeft)? 16 : 0) ];
  
  return *this;
}

AudioMsg& AudioMsg::setDestination( const unsigned int playerno)
{
  unsigned int isc= isCoachMsg() ? 1 : 0;
  ensure( playerno <= 11);
  encoded[4]= msgEncodeChars[ (playerno & 15) | (isc << 4) ];
  
  return *this;
}

AudioMsg& AudioMsg::setServerTime( const unsigned int servertime)
{
  Int3B stime( servertime );
  for (unsigned int i=0; i < 3; i++)
    encoded[AUMSG_HEADER+7+i]= stime.getEncoding(i);
  
  return *this;
}

unsigned int AudioMsg::getSource() const
{
  return msgDecodeChars[encoded[3]] & 15;
}

unsigned int AudioMsg::getDestination() const
{
  return msgDecodeChars[encoded[4]] & 15;
}

unsigned int AudioMsg::getServerTime() const
{
  return Int3B(&encoded[AUMSG_HEADER+7]).getValue();
}

bool AudioMsg::verifyPrefix(const char prefix[4]) const
{
  return (encoded[0] == prefix[0]) &&
    (encoded[1] == prefix[1]) &&
    (encoded[2] == prefix[2]);
}

bool AudioMsg::getSideIsLeft() const
{
  return ((msgDecodeChars[encoded[3]] & 16) > 0);
}

AudioMsg& AudioMsg::createChkSums ()
{
  unsigned int chk= chksum();
  encoded[5]= msgEncodeChars[chk & 63];
  encoded[6]= msgEncodeChars[ (chk >> 6) & 63 ];
  
  return *this;
}

bool AudioMsg::isChkSumsValid () const
{
  unsigned int chk= chksum();
  return ( (encoded[5] == msgEncodeChars[chk & 63]) &&
	   (encoded[6] == msgEncodeChars[ (chk >> 6) & 63 ]) );
}

unsigned int AudioMsg::chksum() const
{
  unsigned int chk= 0;
  for (int i=0; (i < MAX_SAY_MSG_SIZE) && (encoded[i] != 0) ; i++)
    if ((i !=  5) && (i != 6)) 
      chk+= msgDecodeChars[encoded[i]];
  return (chk & 4095); // only two 6 bit chars!!!
}

//////////////////////////////////////////////////////////////////////
//
// Stream operations and methods
//
//////////////////////////////////////////////////////////////////////

AudioMsg& AudioMsg::rewindWrite()
{
  setWritePos(0);
  setReadPos(0);
  writeMark=0;
  readMark=0;
  encoded[AUMSG_STREAMDATA]= 0; // mark end of string
  return *this;
}

AudioMsg& AudioMsg::rewindRead()
{
  readMark=0;
  setReadPos(0);
  return *this;
}

AudioMsg& AudioMsg::markWrite()
{
  writeMark= getWritePos();
  return *this;
}
  
AudioMsg& AudioMsg::rollbackWrite()
{
  if (writeMark > getWritePos())
    writeMark= 0;
  
  if (writeMark >= getReadPos())
    setWritePos(writeMark);
  else
    state= false;
  return *this;
}

AudioMsg& AudioMsg::markRead()
{
  readMark= getReadPos();
  return *this;
}

AudioMsg& AudioMsg::rollbackRead()
{
  if (readMark <= getReadPos())
    setReadPos(readMark);
  else
    state= false;
  return *this;
}


bool AudioMsg::getState() const
{
  return state;
}

unsigned int AudioMsg::getWritePos() const
{
  unsigned int i1= msgDecodeChars[encoded[AUMSG_HEADER+12]],
    i2= msgDecodeChars[encoded[AUMSG_HEADER+12+1]];
  return i1 | (i2 << 6);
}

bool AudioMsg::setWritePos(unsigned int pos)
{
  if (pos + AUMSG_STREAMDATA >= MSG_LENGTH)
    return false;
  unsigned int c1= pos & 63,
    c2= (pos >> 6) & 63;
  encoded[AUMSG_HEADER+12]= msgEncodeChars[c1];
  encoded[AUMSG_HEADER+12+1]= msgEncodeChars[c2];
  return true;
}

unsigned int AudioMsg::getReadPos() const
{
  unsigned int i1= msgDecodeChars[encoded[AUMSG_HEADER+10]],
    i2= msgDecodeChars[encoded[AUMSG_HEADER+10+1]];
  return i1 | (i2 << 6);
}

bool AudioMsg::setReadPos(unsigned int pos)
{
  if ( (pos + AUMSG_STREAMDATA >= MSG_LENGTH) ||
       (pos > getWritePos()) ) 
    return false;
  unsigned int c1= pos & 63,
    c2= (pos >> 6) & 63;
  encoded[AUMSG_HEADER+10]= msgEncodeChars[c1];
  encoded[AUMSG_HEADER+10+1]= msgEncodeChars[c2];
  return true;
}

AudioMsg& AudioMsg::clearState()
{
  state= true;
  return *this;
}

unsigned int AudioMsg::getSize() const
{
  return getWritePos() + AUMSG_STREAMDATA;
}

unsigned int AudioMsg::getReadLeft() const
{
  return getWritePos() - getReadPos();
}

unsigned int AudioMsg::getWriteLeft() const
{
  return MSG_LENGTH - getWritePos();
}


AudioMsg& AudioMsg::operator << (const Int3B &i)
{
  int writepos= getWritePos();
  if (!setWritePos(writepos + 3))
    state= false;
  else {
    for (unsigned int j=0; j < 3; j++)
      encoded[writepos+AUMSG_STREAMDATA + j]= i.getEncoding(j);
    encoded[getWritePos() + AUMSG_STREAMDATA]= 0; // mark end of string
  }
  return *this;
}

AudioMsg& AudioMsg::operator >> (Int3B &i)
{
  int readpos= getReadPos();
  if (!setReadPos(readpos + 3))
    state= false;
  else {
    i= Int3B(&encoded[readpos+AUMSG_STREAMDATA]);
  }
  return *this;
}

AudioMsg& AudioMsg::operator << (const Val5B &v)
{
  int writepos= getWritePos();
  if (!setWritePos(writepos + 5))
    state= false;
  else {
    for (unsigned int i=0; i < 5; i++)
      encoded[writepos+AUMSG_STREAMDATA + i]= v.getEncoding(i);
    encoded[getWritePos() + AUMSG_STREAMDATA]= 0; // mark end of string
  }
  return *this;
}
AudioMsg& AudioMsg::operator >> (Val5B &v)
{
  int readpos= getReadPos();
  if (!setReadPos(readpos + 5))
    state= false;
  else {
    v= Val5B(&encoded[readpos+AUMSG_STREAMDATA]);
  }
  return *this;
}

AudioMsg& AudioMsg::operator << (const bool &b)
{
  int writepos= getWritePos();
  if (!setWritePos(writepos + 1))
    state= false;
  else {
    encoded[writepos + AUMSG_STREAMDATA]= 
      b ? msgEncodeChars[1] : msgEncodeChars[0]; 
    encoded[getWritePos() + AUMSG_STREAMDATA]= 0; // mark end of string
  }
  return *this;
}
AudioMsg& AudioMsg::operator >> (bool &b)
{
  int readpos= getReadPos();
  if (!setReadPos(readpos + 1))
    state= false;
  else {
    b= msgDecodeChars[encoded[readpos + AUMSG_STREAMDATA]] != 0;
  }
  return *this;
}

AudioMsg& AudioMsg::operator << (const char &c)
{
  int writepos= getWritePos();
  if (!setWritePos(writepos + 2))
    state= false;
  else {
    encoded[writepos + AUMSG_STREAMDATA]= msgEncodeChars[c & 63];
    encoded[writepos + AUMSG_STREAMDATA + 1]= msgEncodeChars[(c >> 6) & 3];
    encoded[getWritePos() + AUMSG_STREAMDATA]= 0; // mark end of string
  }
  return *this;
}
AudioMsg& AudioMsg::operator >> (char &c)
{
  int readpos= getReadPos();
  if (!setReadPos(readpos + 2))
    state= false;
  else {
    c= msgDecodeChars[encoded[readpos + AUMSG_STREAMDATA]];
    c|= (msgDecodeChars[encoded[readpos + AUMSG_STREAMDATA + 1]] & 3) << 6;
  }
  return *this;
}

AudioMsg& AudioMsg::operator << (const Angle &a)
{
 int writepos= getWritePos();
  if (!setWritePos(writepos + 2))
    state= false;
  else {
    unsigned int val= int(floor(double(a+PI)*4096/(2 * M_PI)));
    encoded[writepos + AUMSG_STREAMDATA]= 
      msgEncodeChars[ val & 63 ];
    encoded[writepos + AUMSG_STREAMDATA + 1]= 
      msgEncodeChars[ (val >> 6) & 63 ];
    encoded[getWritePos() + AUMSG_STREAMDATA]= 0; // mark end of string
  }
  return *this;
}
AudioMsg& AudioMsg::operator >> (Angle &a)
{
  int readpos= getReadPos();
  if (!setReadPos(readpos + 2))
    state= false;
  else {
    unsigned int i1= msgDecodeChars[encoded[readpos+ AUMSG_STREAMDATA]],
      i2= msgDecodeChars[encoded[readpos+ AUMSG_STREAMDATA + 1]];
    a= ((i1 + i2 << 6) * 2 * M_PI / 4096) - PI;
  }
  return *this;
}

AudioMsg& AudioMsg::operator << (const double &d)
{
  int dr= int(floor((d*10.0) + 2048));
  if ( (dr > 4095) || (dr < 0))
    state= false;
  else {    
    int writepos= getWritePos();
    if (!setWritePos(writepos + 2))
      state= false;
    else {
      // now remove the sign of the rounded values
      unsigned int d_(dr);
      encoded[writepos + AUMSG_STREAMDATA]= 
	msgEncodeChars[ d_ & 63 ];
      encoded[writepos + AUMSG_STREAMDATA + 1]= 
	msgEncodeChars[ (d_ >> 6) & 63 ];           
      encoded[getWritePos() + AUMSG_STREAMDATA]= 0; // mark end of string
    }
  }
  return *this;
}
AudioMsg& AudioMsg::operator >> (double &d)
{
  int readpos= getReadPos();
  if (!setReadPos(readpos + 2))
    state= false;
  else {
    unsigned int c0= msgDecodeChars[encoded[readpos + AUMSG_STREAMDATA]],
      c1= msgDecodeChars[encoded[readpos + AUMSG_STREAMDATA + 1]];
    d= ( (c0 | (c1 << 6)) - 2048.0) / 10.0;
  }
  return *this;
}

AudioMsg& AudioMsg::operator << (const NodaChar &c)
{
  int writepos= getWritePos();
  if (!setWritePos(writepos + 1))
    state= false;
  else {
    encoded[writepos + AUMSG_STREAMDATA]=msgEncodeChars[c];
    encoded[getWritePos() + AUMSG_STREAMDATA]= 0; // mark end of string
  }
  return *this;
}
AudioMsg& AudioMsg::operator >> (NodaChar &c)
{
  int readpos= getReadPos();
  if (!setReadPos(readpos + 1))
    state= false;
  else {
    c= NodaChar(msgDecodeChars[encoded[readpos + AUMSG_STREAMDATA]]);
  }
  return *this;
}


//////////////////////////////////////////////////////////////////////
//
// Output for debugging
//
//////////////////////////////////////////////////////////////////////

class ostream & operator << (class ostream &os, const AudioMsg &a)
{
  char side= a.getSideIsLeft() ? 'l' : 'r';
  if (a.isChkSumsValid()) 
    os << "(AudioMsg "; 
  else 
    os << "(AudioMsg-invalid ";
  os << " (" << a.getSource() << side << " -> " <<
      a.getDestination() << ")(t= " << a.getServerTime() << ")";
    if (a.isCoachMsg())
      os << "(coachmsg)";
    os << "(Size= " << a.getSize() << " unread= " 
       << a.getReadLeft();
    if (a.getState()) 
      os << " OK ))";
    else 
      os << " FAIL ))";
    return os;
}

