//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// File : coordinates.h

//////////////////////////////////////////////////////////////////////
//
// $Id: coordinates.h,v 1.15 2000/10/19 15:27:45 goetz Exp $
//
//////////////////////////////////////////////////////////////////////



// Headerfile for angle- and 2d-vectorclass
// All methods are inline; there is no coordinates.C

#ifndef COORDINATES_H
#define COORDINATES_H

#include <iostream.h>
#include <math.h>
#include "minmax.h"


// Class for angles in the interval [-pi, pi].
// This is the kind of angles the soccerserver uses.

class Angle
{
public:

  Angle() : v( 0.0 ) {}

  Angle( double _v ) : v( mod2pi( _v ) ) {}

  // conversion to double; provides comparison operators <,>,<=,>=,==,!=
  operator double() const { return v; }

  // arithmetic (+,-,*,/)

  Angle operator - ( ) const
  { return Angle( -v ); }
 
  Angle &operator += ( const Angle &ang )
  { v = mod2pi( v + ang.v ); return *this; }

  Angle &operator -= ( const Angle &ang )
  { v = mod2pi( v - ang.v ); return *this; }

  Angle &operator /= ( double s )
  { v = mod2pi( v / s ); return *this; }

  Angle &operator *= ( double s )
  { v = mod2pi( v * s ); return *this; }

  Angle operator + ( const Angle &ang ) const
  { return Angle( v + ang.v ); }

  Angle operator - ( const Angle &ang ) const
  { return Angle( v - ang.v ); }

  Angle operator / ( double s ) const
  { return Angle( v / s ); }

  Angle operator * ( double s ) const
  { return Angle( v * s ); }

  // just for writing a + 2 * b instead of a + b * 2
  friend Angle operator * ( double s, const Angle &b )
  { return Angle( s * b.v ); }

private:

  double v;

  // shift an angle into the intervall [-pi, pi]
  inline static double mod2pi( double angle )
  {
    while( angle > M_PI )
      angle -= 2.0 * M_PI;
    while( angle <= -M_PI )
      angle += 2.0 * M_PI;
    return angle;
  }
};

const Angle PI = M_PI;
const Angle PI_2 = M_PI_2;
const Angle PI_3 = Angle( M_PI / 3.0 );
const Angle PI_4 = M_PI_4;
const Angle PI_6 = Angle( M_PI / 6.0 );
const Angle PI_8 = Angle( M_PI / 8.0 );

// Class for 2-dimensional coordinates (x,y).
// Zero angle is in direction of positive x-axis.

class Vector2d
{
public:

  Vector2d() :
    x( 0.0 ), y( 0.0 ) {}
  
  Vector2d( double _x, double _y ) :
    x( _x ), y( _y ) {}

  double getX() const
  { return x; }
  
  double getY() const
  { return y; }

  double getLength() const
  { return sqrt( x * x + y * y ); }

  Angle getAngle() const
  { return Angle( atan2( y, x ) ); }

  int operator == ( const Vector2d &c ) const
  { return x == c.x && y == c.y; }
  
  void setX( double _x )
  { x = _x; }
  
  void setY( double _y )
  { y = _y; }
  
  void setXY( double _x, double _y )
  { x = _x; y = _y; }
  
  // init vector with polar coordinates
  void setRadAng( double r, Angle angle )
  {
    x = r * cos( angle );
    y = r * sin( angle );
  }

  // arithmetic (+,-)

  Vector2d operator - ( ) const
  { return Vector2d( -x, -y ); }
 
  Vector2d &operator += ( const Vector2d &c )
  { x += c.x; y += c.y; return *this; }
  
  Vector2d &operator -= ( const Vector2d &c )
  { x -= c.x; y -= c.y; return *this; }
  
  Vector2d operator + ( const Vector2d &c ) const
  { return Vector2d( x + c.x, y + c.y ); }
  
  Vector2d operator - ( const Vector2d &c ) const
  { return Vector2d( x - c.x, y - c.y ); }

  // stretch vector by double value s (same as method 'stretch')
  Vector2d &operator *= ( double s )
  { x *= s; y *= s; return *this; }

  Vector2d &stretch( double s )
  { x *= s; y *= s; return *this; }
  
  Vector2d operator * ( double s ) const
  { return Vector2d( s * x, s * y ); }

  // just for writing v1 + 2 * v2 instead of v1 + v2 * 2
  friend Vector2d operator * ( double s, const Vector2d &c )
  { return c * s; }

  // dot product
  double operator * ( const Vector2d &b ) const
  { return x * b.x + y * b.y; }
  
  // turn by angle alpha (uses sin and cos)
  Vector2d &turn( Angle alpha )
  {
    double s = sin( alpha );
    double c = cos( alpha );
    
    double x_old = x;
    x = c * x - s * y;
    y = s * x_old + c * y;
    return *this;
  }
  
  // (faster) turn by +90 DEG
  Vector2d &perp()
  {
    double x_old = x;
    x = -y;
    y = x_old;
    return *this;
  }

private:
  
  double x, y;
};


// conversion RAD -> DEG

inline double
RAD_TO_DEG( double ang )
{
  return double( ang ) * 180.0 / M_PI;
}


// conversion DEG -> RAD

inline Angle
DEG_TO_RAD( double ang )
{
  return Angle( ang * M_PI / 180.0 );
}


// Square of Euclidean norm

inline double
sqrnorm( const Vector2d &c )
{
  return c.getX() * c.getX() + c.getY() * c.getY();
}


// Euclidean norm

inline double
norm( const Vector2d &c )
{
  return sqrt( c.getX() * c.getX() + c.getY() * c.getY() );
}


// Distance between two points (vectors)

inline double
dist( const Vector2d &a, const Vector2d &b )
{
  return norm( a - b );
}


// Angle between the vectors a and b

inline Angle
angleBetween( const Vector2d &a, const Vector2d &b )
{
  return b.getAngle() - a.getAngle();
}


// Bisecting line/angle between two angles

inline Angle
bisect( Angle angle1, Angle angle2 )
{
  return angle1 + 0.5 * ( angle2 - angle1 );
}


// Intersect circle with line or circle or other line

enum Intersect { None, Valid, Instable, R1larger, R2larger };

Intersect intersectLineCircle( const Vector2d &linedirection,
			       const Vector2d &circlecenter,
			       double radius, double &l1, double &l2 );

Intersect intersectCircleCircle( double r1, const Vector2d &center2,
				 double r2, Vector2d &l1, Vector2d &l2 );

Intersect intersectLineLine( const Vector2d &line1dir,
			     const Vector2d &line2offset,
			     const Vector2d &line2dir,
			     double &l1, double &l2);

// Generate Vector2d with polar coordinates

inline Vector2d
radAng2Vector2d( double r, Angle angle )
{
  return Vector2d( r * cos( angle ), r * sin( angle ) );
}


// Vector output to stream 

inline
class ostream & operator << ( class ostream &os, const Vector2d &c )
{
  return os << "( " << c.getX() << ", " << c.getY() << " )";
}

#endif
