//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*-/////////////////////////////////////////////////////////
//
// $Id: player.h,v 1.11 2000/10/19 15:27:47 goetz Exp $
//
// player.h - interface for player client
//
///////////////////////////////////////////////////////////////////////


#ifndef __PLAYER_H__
#define __PLAYER_H__

#include <unistd.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>

#include "constants.h"
#include "comm.h"
#include "worldmodel.h"
#include "AdvSkills.h"
#include "worldmap_r.h"
#include "audiomgr.h"


// filename of snapshot log files
#define SNAP_LOG_FILENAME "snaps_"
#define SNAP_LOG_EXTENSION ".log"


const int MAX_MAP_AGE = 20;


class Player {

public:


  // Constructor / Destructor
  ///////////////////////////////////////////////////////////////////////////

  // Generates and logs in a player client at the given soccerserver.
  // After the constructor has finished all properties of the agent are set.
  // 
  // Call this constructor first in your derived classes constructors before
  // you implement any further initialization of the agent.
  //
  // Parameter description:
  // - numSnapshots
  // Specifies the memory usage of the Worldmodel. Larger numbers will keep
  // more snapshots of the past accessible, but consume quite some memory.
  Player (const char* team_name,             // name of own team
	  bool        isGoalie,              // init as goalie ?
	  const char* host         = DEFAULT_HOST,
	  int         port         = PORT,   
	  double      MapWindowMagnify = 0,  // size of mapwindow (only if >0)
	  bool        logging_snap = false,  // logs all snapshots
	  int         reconnectNumber = 0,   // reconnecting (only if number>0)
	  int         numSnapshots = DEFAULT_SNAPSHOT_DEPTH); // see below


  // Starts execution of the Player client. 
  // After calling this method, the agent will repeatedly call act
  // every time it needs to send an action while the server is running.
  //
  // You cannot override this method.
  void play();

  // returns the snapshot for the current timestep
  // this method should be called after WorldModel::feedbackAudio() !!
  Snapshot* getSnapshot() { return world->getSnapshot(getTime()); }
  
  // Called once every time player needs to send an action to the server.
  //
  // Override this method virtual in your derived classes to implement 
  // player behaviour.
  virtual void act() = 0;

  
  // Cleans up and logs out
  //
  // You can put own code in your derived destructors. As the final action call
  // this destructor so the agent will leave the simulation properly 
  // and release its resources (files, sockets, windows, etc).
  virtual ~Player();  


  
  // general information about the client himself and the server state
  /////////////////////////////////////////////////////////////////////

  // obvious
  int          getMyNumber()  {return comm->getMyNumber();}
  bool         amIGoalie()    {return comm->amIGoalie();}
  const char*  getMyTeam()    {return comm->getMyTeam();}
  playing_side getMySide()    {return comm->getMySide();}
  play_mode    getPlayMode()  {return comm->getPlayMode();}

  // returns the timestep the server is in, our current time.
  // (There is only ONE current time: This is it.)
  int          getTime()      {return comm->getServerTime();} 

  // if needed, returns a snapshot for a given time from the worldmodel
  Snapshot*    getSnapshot(int time) {return world->getSnapshot(time);}
  
  // sends an audiomessage to other clients immediately
  void         say(AudioMsg &msg)         {comm->say(msg);}

  // returns true, if a new audiomessage is available in msg, false otherwise
  bool         getNextMsg(AudioMsg &msg) {return comm->getNextMsg(msg);}

  // feed audio information into worldmodel
  bool           worldFeedbackAudio( AudioMsg &audioMsg ) {
    return world->feedbackAudio(audioMsg);
  }

  // write worldmodel information for timestep maxTime into AudioMsg stream
  bool           worldGetSayMsg( AudioMsg &audioMsg ) {
    return world->getSayMsg(audioMsg);
  }
  
  // the clients modules
  //////////////////////////////////////////////////////////////////////////

  AdvSkills *skills;

private:

  bool writeSnaps;

  Comm *comm;

  Worldmodel *world;

  WorldMap_r *mapwindow;

  ofstream out;

};


#endif
