//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module ballHandling
//
// $Id: ballhandlingmodule.C,v 1.67 2000/10/19 16:30:11 eri Exp $
//
//////////////////////////////////////////////////////////////////////


#include "ballhandlingmodule.h"


// definition of trace macros
////////////////////////////////////////////////////////////////

// general tracing, used to observe functioning of the module
#undef CERR          // for all error messages (leave this on always)
#undef COUT          // just regular, reduced output of all eval and actions
#undef DRAWTRACE     // draw commands for replayer

// detailed and specific debugging traces (may produce a lot of output!)
#undef FINDTRACE     // for details on findBall
#undef ICPTTRACE     // for details on intercepting
#undef DRIBBLETRACE  // for details on fastdribble, dribble and placing
#undef ATTACKTRACE   // for details on attack
#undef PLACETRACE    // for details on place ball
#undef SIMTRACE      // for details on simulations made
#undef TURNHEADTRACE // for details on turn head
#undef TACTICALTRACE // for details on tactical states


#define CERR(output) cerr << time << " BALLH: " << output << endl
//#define COUT(output) cout << time << " BALLH: " << output << endl
//#define DRAWTRACE(output) cout << time << "% " << output << endl

//#define TURNHEADTRACE(output) cout << time << " TurnHead  -> " << output << endl
//#define FINDTRACE(output) cout << time << " FindBall  -> " << output << endl
//#define ICPTTRACE(output) cout << time << " Intercept -> " << output << endl
//#define DRIBBLETRACE(output) cout << time << " Dribble   -> " << output << endl
//#define ATTACKTRACE(output) cout << time << " Attack    -> " << output << endl
//#define PLACETRACE(output) cout << time << " Place     -> " << output << endl
//#define SIMTRACE(output) cout << time << " Sim       -> " << output << endl
//#define TACTICALTRACE(output) cout << time << " Tactic -> " << output << endl

#ifndef COUT
#define COUT(output)
#else
#warning: BALLH debug trace COUT is activated!
#endif

#ifndef CERR
#define CERR(output)
#warning: BALLH error trace is deactivated !
#endif

#ifndef TURNHEADTRACE
#define TURNHEADTRACE(output)
#else
#warning: BALLH debug trace TURNHEADTRACE is activated!
#endif

#ifndef SIMTRACE
#define SIMTRACE(output)
#else
#warning: BALLH debug trace SIMTRACE is activated!
#endif

#ifndef DRAWTRACE
#define DRAWTRACE(output)
#else
#warning: BALLH debug trace DRAWTRACE is activated!
#endif


#ifndef FINDTRACE
#define FINDTRACE(output)
#else
#warning: BALLH debug trace FINDTRACE is activated!
#endif

#ifndef ICPTTRACE
#define ICPTTRACE(output)
#else
#warning: BALLH debug trace ICPTTRACE is activated!
#endif

#ifndef DRIBBLETRACE
#define DRIBBLETRACE(output)
#else
#warning: BALLH debug trace DRIBBLETRACE is activated!
#endif

#ifndef ATTACKTRACE
#define ATTACKTRACE(output)
#else
#warning: BALLH debug trace ATTACKTRACE is activated!
#endif

#ifndef PLACETRACE
#define PLACETRACE(output)
#else
#warning: BALLH debug trace PLACETRACE is activated!
#endif

#ifndef TACTICALTRACE
#define TACTICALTRACE(output)
#else
#warning: BALLH debug trace TACTICALTRACE is activated!
#endif


BallHandlingModule::BallHandlingModule(bool _goalie) {


  // default ballhandling parameters
  //////////////////////////////////////
  
  // findBall
  maxBallAge       = 10;
  maxViewBallAge   = 5;

  // intercept
  maxPlayerAge     = 20;

  safeKickRange = PLAYER_SIZE + BALL_SIZE + 0.7 * KICKABLE_MARGIN;

  // dribble
  maxDribbleStamina   = 2.0*STAMINA_INC_MAX; // < MAXPOWER, so we can intercept


  // initialization
  ////////////////////////
  goalie           = _goalie;

  // findBall
  lookingForBall   = false;
  lastBallAge      = -1;

  // data storage
  initDataStorage();

  // communication
  sendInfo = new BallHandlingModuleInfo();
  sendSomething = false;

  lastDistToBall = FIELD_LENGTH;
  
}

BallHandlingModule::~BallHandlingModule() {
  delete (sendInfo);
} 



/////////////////////////////////////////////////////////////////////////
//
// FEEDBACK
//
/////////////////////////////////////////////////////////////////////////



bool 
BallHandlingModule::feedbackWorld(const Snapshot* snap, 
				  TacticalSnapshot* tsnap) {

  if (! updateDataStorage(snap, tsnap) ) {
    
    return false;
    
  } else {

    // reset findBall-flag when ball was newly detected
    if ( lookingForBall ) {
      if ( ball.age < lastBallAge+1 ) {
	lookingForBall = false; 
	FINDTRACE("found the ball");
      } else {
	FINDTRACE("did not find the ball yet");
      }
    }

    return true;

  }
}

bool 
BallHandlingModule::feedbackAudio(AudioMsg &audioMsg) {
  
  bool hasInfo;

  // read info from stream
  audioMsg >> hasInfo;
  
  if (hasInfo) {
  
    BallHandlingModuleInfo receivedInfo;
    audioMsg >> receivedInfo;

    // evaluate received information into module member variables here
    // (not used yet)
    
  }

  return true;
}



/////////////////////////////////////////////////////////////////////////
//
// EVALUATE
//
/////////////////////////////////////////////////////////////////////////
#include "ballhandlingEvaluate.C"


/////////////////////////////////////////////////////////////////////////
//
// ACT
//
/////////////////////////////////////////////////////////////////////////
#include "ballhandlingAct.C"


//////////////////////////////////////////////////////////////////////////
//
// private auxiliary methods
//
//////////////////////////////////////////////////////////////////////////

void 
BallHandlingModule::initDataStorage() {
  
  playMode = PLAY_MODE_UNKNOWN;
  time = -1;
  stamina = 0;
  availableStamina = 0;
  distToBall = FIELD_LENGTH;
  faceAngleToBall = 0;
  bodyAngleToBall = 0;

  snapTime  = time;
  tsnapTime = time;
  
}

 
bool 
BallHandlingModule::updateDataStorage( const Snapshot*   snap, 
				       TacticalSnapshot* tsnap ) {
  // returns true if given data (which is to be used for evaluate) is valid
  // and could be stored temporarily,
  // false otherwise
  
  if (snap == NULL) {

    if (time > -1) CERR ("data update -> failed: worldSnap == NULL" );
    return false;
    
  } else if (tsnap == NULL) {

    if (time > -1) CERR("data update -> failed: tacticalSnap == NULL" ); 
    return false;
    
  } else {
    
    snapTime = snap->time;
    tsnapTime = tsnap->time;
    
    playMode = snap->playMode;
    time     = snap->time;
    my       = snap->my;
    stamina  = snap->senseBody.stamina;
    ball     = snap->ball;
    ourGoal  = snap->ourGoal;

    lastDistToBall = distToBall;
    distToBall       = snap->my.distTo(snap->ball);

    // Noda checks for <= stamina_limits, so rather stay slightly above:
    availableStamina = max (0.0, stamina - EFFORT_DEC_THR * STAMINA_MAX - 1 );

    faceAngleToBall  = snap->my.faceAngleTo( ball );
    bodyAngleToBall  = snap->my.bodyAngleTo( ball );
    
    action   = WAIT;
   
    TACTICALTRACE("state = " << tsnap->ourAction.state <<
		  ", performer = " << tsnap->ourAction.performer <<
		  ", receiver = " << tsnap->ourAction.receiver);

    return true;
    
  }

} // update


bool 
BallHandlingModule::checkDataStorage( const Snapshot*   snap, 
				      TacticalSnapshot* tsnap) {
  // returns true if data in store (which was used for evaluate) is same 
  // as given data (which is to be used for act), 
  // false otherwise
  
  if (snap == NULL) {

    if (time > -1) CERR ("data check -> failed: worldSnap == NULL" );
    return false;

  } else if (tsnap == NULL) {

    if (time > -1) CERR("data check -> failed: tacticalSnap == NULL" ); 
    return false;
    
  } else if (snapTime != snap->time) {

    CERR ("data check -> failed: different timestamps in worldSnap !" );
    return false;

  } else if (tsnapTime != tsnap->time) {

    CERR (" data check -> failed: different timestamps in tacticalSnap !" );
    return false;
    
  } else return true;

} //check    


///////////////////////////////////////////////////////////////////////////
//
// SAY MSG and Info Class
//
///////////////////////////////////////////////////////////////////////////



bool 
BallHandlingModule::getSayMsg(AudioMsg &audioMsg) {
  
  if (sendSomething) {
    // will currently never be true, communication between ballhandlingmodules
    // is still unused.

    audioMsg.markWrite();
    
    audioMsg << true;

    audioMsg << sendInfo;
    
    // check if error occurred
    if( !audioMsg.getState() )
      {
	cerr << "BALLH: Stream overflow: " << audioMsg << endl;
	audioMsg.rollbackWrite().clearState();

	audioMsg << false;

      }
  } else {

    audioMsg << false;

  }
    
  return true;
  
}


BallHandlingModuleInfo::BallHandlingModuleInfo() {

};

BallHandlingModuleInfo::~BallHandlingModuleInfo() {

};

class AudioMsg &operator<< (AudioMsg &aumsg, 
			    const BallHandlingModuleInfo &) {
  // write all data from Info class here, and make sure there are 
  // AudioMsg-operators defined for the specific type of data you write !
  
  // for example:
  // aumsg << info.value1;
  // aumsg << info.value2;

  return aumsg;
}


class AudioMsg &operator>> (AudioMsg &aumsg, 
			    BallHandlingModuleInfo &) {
  // read data for Info class in same order as writing it

  // aumsg >> info.value1;
  // aumsg >> info.value2;
  
  return aumsg;
}
