//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

///////////////////////////
//
// class PassEval
//
///////////////////////////

#include "passeval.h"

// utility function - is a given coordinate inside the field? (with a security interval)
static inline bool
inField( Vector2d pos )
{
  const double SECURITY_DIST = KICKABLE_MARGIN + BALL_SIZE + PLAYER_SIZE;   // = 1m

  if( fabs( pos.getX() ) > FIELD_LENGTH/2.0 - SECURITY_DIST || 
      fabs( pos.getY() ) > FIELD_WIDTH/2.0 - SECURITY_DIST )
    return false;
  return true;
}

////////////////////////////
//
// Constructor
//
///////////////////////////

PassEval::PassEval(const GoalKickEval * _goalkickeval):
  goalkickeval(_goalkickeval),
  DANGERZONE(3),
  SAMELEVEL_LENGTH(4*PLAYER_SIZE),  // consider this!
  SUPERKICK_DURATION(5),            // estimated values!?
  FIXEDKICK_DURATION(2),
  KICKAWAY_DURATION(1),
  SUPERKICK_LENGTH(35.0),           // some estimated values, communicated by CM
  FIXEDKICK_LENGTH(26.0),
  KICKAWAY_LENGTH(20.0),
  RPASS_LATTICELENGTH(3.0),         // length of lattice to determine rangepasses
  SHORTPASS_LENGTH(10.0)
{}

////////////////////////////
//
// Destructor
//
///////////////////////////

PassEval::~PassEval()
{}  

////////////////////////////////////////////////////////////////////////////////////
//
// public methods
//
////////////////////////////////////////////////////////////////////////////////////

////////////////////////////
//
// Grade PassEval::ourEval - with rangepasses
//
///////////////////////////

Grade PassEval::ourEval(const Snapshot* worldSnap, int plnumber, int panumber, Vector2d &posOfPass) const
{
  PropOf pprop;
  get_pPass(true, worldSnap, plnumber, worldSnap->ourTeam(panumber).pos, panumber, pprop);
  
  PropOf rprop;
  get_rPass(true, worldSnap, plnumber, panumber, rprop);

  // we have only a p-pass
  if ( (pprop.perhaps) && (!rprop.perhaps) ) {
    posOfPass = pprop.pos;
    return pprop.grade;
  }
  // we have only a r-pass
  if ( (!pprop.perhaps) && (rprop.perhaps) ) {
    posOfPass = rprop.pos;
    return rprop.grade;
  }
  // we have both possibilities
  if ( (pprop.perhaps) && (rprop.perhaps) ) 
    if ( betterPass(worldSnap, rprop, pprop) ) {
      posOfPass = rprop.pos;
      return rprop.grade;
    }
    else {
      posOfPass = pprop.pos;
      return pprop.grade;
    }
  
  // there are no passes
  pprop.grade.set_BAD();
  return pprop.grade;
}

////////////////////////////
//
// Grade PassEval::ourEvalProp
//
///////////////////////////

PropOf PassEval::ourEvalProp(const Snapshot* worldSnap, int plnumber, int panumber) const
{
  PropOf pprop;
  get_pPass(true, worldSnap, plnumber, worldSnap->ourTeam(panumber).pos, panumber, pprop);
  
  PropOf rprop;
  get_rPass(true, worldSnap, plnumber, panumber, rprop);

  // we have only a p-pass
  if ( (pprop.perhaps) && (!rprop.perhaps) )
    return pprop;

  // we have only a r-pass
  if ( (!pprop.perhaps) && (rprop.perhaps) )
    return rprop;

  // we have both possibilities
  if ( (pprop.perhaps) && (rprop.perhaps) ) 
    if ( betterPass(worldSnap, rprop, pprop) )
      return rprop;
    else
      return pprop;
  
  // there are no passes
  pprop.perhaps = false;
  pprop.grade.set_BAD();
  return pprop;
}

////////////////////////////
//
// Grade PassEval::otherEval
//
///////////////////////////

Grade PassEval::otherEval(const Snapshot* worldSnap, int plnumber, int panumber) const
{
  PropOf prop;
  get_pPass(false, worldSnap, plnumber, worldSnap->otherTeam(panumber).pos, panumber, prop);
  
  PAEVTRACE(cout << worldSnap->time << " PASS: ourEval - got some props for other team:" << endl;
	    coutPropOf(worldSnap, prop););
 
  return prop.grade;
}

////////////////////////////////////////////////////////////////////////////////////
//
// private methods
//
////////////////////////////////////////////////////////////////////////////////////

////////////////////////////
//
// PassEval::check_for_safekick
//
///////////////////////////

int PassEval::check_for_safekick(const Snapshot *wS, int player) const
{ 
  // in these playmodes we have got enough time
  if ( (wS->playMode == PLAY_MODE_GOAL_KICK_US) ||
       (wS->playMode == PLAY_MODE_FREE_KICK_US) ||
       (wS->playMode == PLAY_MODE_CORNER_KICK_US) )
    return ( inField(wS->ourTeam(player).pos) ? SUPERKICK_DURATION : FIXEDKICK_DURATION );

  // closest opponent to player
  int opp = wS->otherTeamClosestTo(wS->ourTeam(player).pos);
  
  if (opp == UNKNOWN)          // player does not see any opponent so everything seems to be allright
    return ( inField(wS->ourTeam(player).pos) ? SUPERKICK_DURATION : FIXEDKICK_DURATION );
  
  // distance between player and closest opponent
  double pod = dist(wS->ourTeam(player).pos, wS->otherTeam(opp).pos); 
  
  // superkick
  if (pod > PLAYER_SPEED_MAX * SUPERKICK_DURATION) 
    return ( inField(wS->ourTeam(player).pos) ? SUPERKICK_DURATION : FIXEDKICK_DURATION );
  
  // fixed kick
  if (pod > PLAYER_SPEED_MAX * FIXEDKICK_DURATION)
    return FIXEDKICK_DURATION;
  
  // kickaway
  return KICKAWAY_DURATION;
}

////////////////////////////
//
// PassEval::check_passlength
//
///////////////////////////

double PassEval::check_passlength(const Snapshot* wS, int player) const
{
  // time the player has to perform a kick
  int time_f_kick = check_for_safekick(wS, player);
  
  if (time_f_kick >= SUPERKICK_DURATION)
    return SUPERKICK_LENGTH;

  if (time_f_kick >= FIXEDKICK_DURATION)
    return FIXEDKICK_LENGTH;

  return KICKAWAY_LENGTH;
}

///////////////////////////
//
// bool PassEval::get_pPass
//
////////////////////////////

bool PassEval::get_pPass(bool we, const Snapshot *wS, int plnumber,
			 const Vector2d target, int panumber, PropOf &prop) const
{
  double offsideLine = we ? wS->ourOffsideLine(MAXAGE) : wS->otherOffsideLine(MAXAGE);

  Vector2d player    = we ? wS->ourTeam(plnumber).pos : wS->otherTeam(plnumber).pos;
  Vector2d partner   = we ? wS->ourTeam(panumber).pos : wS->otherTeam(panumber).pos;

  // check_passlength does not work for the opponent team
  double passLength  = we ? check_passlength(wS, plnumber) : 40;

  int time_for_kick  = we ? check_for_safekick(wS, plnumber) : SUPERKICK_DURATION;

  // set the PropOf
  prop.grade.set_BAD();
  prop.perhaps = true;
  prop.number  = panumber;
  prop.pos     = target;
  prop.dist    = dist(player, target);

  // partner close and no dangerous situation
  if (time_for_kick > KICKAWAY_DURATION && prop.dist < SHORTPASS_LENGTH) {
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  // partner too far away
  if (passLength > 0 && passLength < prop.dist) {
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  prop.offside = (offsideLine <= partner.getX());

  // if player is not offside, test teammates in a 10-m-corridor
  // if one of them is offside, do not pass
  if (!prop.offside) {
    for (int i=1; i <= 11; i++)
      if (i != plnumber && i != panumber)
	if (we) {
	  if ( (wS->ourTeam(i).inCorridor(player, target, 10, PI_3)) && 
	       (wS->ourTeam(i).pos.getX() >= offsideLine) )
	    prop.offside = true;
	}
	else {
	  if ( (wS->otherTeam(i).inCorridor(player, target, 10, PI_3)) && 
	       (wS->otherTeam(i).pos.getX() >= offsideLine) )
	    prop.offside = true;
	}
  }
  if (prop.offside) {
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  // estimated time player will need to reach the target position (good estimation for n large)
  int time_to_reach = int( max( 0.0, dist(partner, target) - PLAYER_SIZE - BALL_SIZE - KICKABLE_MARGIN )
			   * ( 1 - PLAYER_DECAY ) / STAMINA_INC_MAX / DASH_POWER_RATE );

  // calculate maximal velocity the ball may have at the beginning so that partner can reach it
  prop.vel = min( BALL_SPEED_MAX, prop.dist * (1 - BALL_DECAY) / (1 - power(BALL_DECAY, time_to_reach + 1)) );

  if (prop.vel < MAXPOWER/2 * KICK_POWER_RATE) {      // partner will not reach the pass
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  // set velocity to target velocity
  prop.vel *= power( BALL_DECAY, time_to_reach );

  // calculate grade and fill rest of prop
  prop.grade = calculateGrade(we, wS, player, target, prop);

  // partner not standing free or no pass corridor found
  if (prop.free < DANGERZONE || prop.corridor <= 0) {
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  return true;
} // get_pPass

///////////////////////////
//
// bool PassEval::get_rPass
//
////////////////////////////

bool PassEval::get_rPass(bool we, const Snapshot *wS, int plnumber, int panumber,
			 PropOf &prop) const
{
  // let be an array with:
  //
  //
  //  *  *  *  +       |       p = partner 
  //
  //  *  *  *  +       |       * = possible points to do a rangepass
  // 
  //  p  *  *  *       |       + = unwanted points
  //
  //  *  *  *  +      \ /      distance beetween points: 3 meters
  //
  //  *  *  *  +       j
  //
  //
  //  -  -  -  > i  

  PropOf rprops[4][5];   // propOf for the different rangepasses

  Vector2d player  = we ? wS->ourTeam(plnumber).pos : wS->otherTeam(plnumber).pos;
  Vector2d partner = we ? wS->ourTeam(panumber).pos : wS->otherTeam(panumber).pos;

  for (int i = 0; i < 4; i++)      // do something for all possible points
    for (int j = 0; j < 5; j++) {  // do it with rangepasses[i][j]
      // calculate the positions of the points
      rprops[i][j].pos.setX(partner.getX() +     i * RPASS_LATTICELENGTH);
      rprops[i][j].pos.setY(partner.getY() + (j-2) * RPASS_LATTICELENGTH);

      // all rangepasses are a priori possible ...
      rprops[i][j].perhaps = true;
      // ...but not the partner himself and not the ones marked with "+"
      if ( ((i == 0) && (j == 2)) || ((i == 3) && (j != 2)) ||

	   // there are other points we do not want - but too complicated to draw in ascii-code:
	   // draw the line from player to partner and draw the perpendicular to that line;
	   // no point behind that is wanted

	   // dist(player, rprops[i][j].pos) 
	   // * cos(fabs(angleBetween(partner - player, rprops[i][j].pos - player)))
	   // > dist(player, partner) ||
	   
	   // we want no points outside the field
	   !inField(rprops[i][j].pos) ) {
	rprops[i][j].perhaps = false;
	rprops[i][j].grade.set_BAD();
      }
      else {
	// if this rangepass is possible - we did by now nothing else than considering this question - 
	// calculate the rest of its propOf
	get_pPass(we, wS, plnumber, rprops[i][j].pos, panumber, rprops[i][j]);

	if (rprops[i][j].perhaps)
	  // is there an opponent that is closer to the target than the partner?
	  if (dist(partner, rprops[i][j].pos) > rprops[i][j].free)
	    {
	      rprops[i][j].perhaps = false;
	      rprops[i][j].grade.set_BAD();
	    }
      }
    }

  TESTPAEVTRACE(
		cout << wS->time << " checked r-passes for player" << panumber << endl;
		for (int jj = 0; jj < 5; jj++) {
		  for (int ii = 0; ii < 4; ii++)
		    cout << (rprops[ii][jj].perhaps ? "1" : "0" );
		  cout << endl;
		} );
  
  bool found_pass = false;

  // set prop to default values
  prop.grade.set_BAD();
  prop.perhaps = false;
  prop.number = -1;
  prop.offside = false;
  prop.corridor = 0;
  prop.free = 0;
  prop.dist = 0;
  prop.pos = player;
  prop.vel = 0;

  // now search the "best" rangepass
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 5; j++)
      if (rprops[i][j].perhaps)
	if (!found_pass) {
	  prop = rprops[i][j];
	  found_pass = true;
	}
	else if (betterPass(wS, rprops[i][j], prop))
	  prop = rprops[i][j];

  PAEVTRACE(
	    if (wS->time > 0) {
	      cout << wS->time << " check r-passes for player "
		   << panumber << ", result: " << prop.grade << endl;
	    } );
  
  return found_pass;
} // get_rPass

////////////////////////////
//
// PassEval::calculateGrade
//
///////////////////////////

Grade PassEval::calculateGrade(bool we, 
			       const Snapshot *wS, const Vector2d &player, const Vector2d &partner,
			       PropOf &prop) const
{
  double playerx  = player.getX();
  double playery  = player.getY(); 
  double partnerx = partner.getX();
  double partnery = partner.getY();

  WorldObject ourGoal = ( we ? wS->ourGoal : wS->otherGoal );
  WorldObject otherGoal = ( we ? wS->otherGoal : wS->ourGoal );

  // calculate maximal corridor width and maximal free circle around partner
  prop.corridor = 20;
  prop.free = 20;

  if (we) {
    int closestOpponent = wS->otherTeamClosestTo(partner);
    if (closestOpponent != UNKNOWN)
      prop.free = int(dist(partner,wS->otherTeam(closestOpponent).pos));
    prop.corridor = min(20, prop.free/4*4);
    while (wS->otherTeamInCorridor(player, partner, prop.corridor, PI_6) && prop.corridor > 0) {
      prop.corridor -= 4;
    }
  }
  else {
    int closestOpponent = wS->ourTeamClosestTo(partner);
    if (closestOpponent != UNKNOWN)
      prop.free = int(dist(partner,wS->ourTeam(closestOpponent).pos));
    prop.corridor = min(20, prop.free/4*4);
    while (wS->ourTeamInCorridor(player, partner, prop.corridor, PI_6) && prop.corridor > 0) {
      prop.corridor -= 4;
    }
  }

  switch (prop.corridor) {
  case 20 :
    prop.grade.set_PERFECT(); break;
  case 16 :
    prop.grade.set_PERFECT(); break;
  case 12 :
    prop.grade.set_PERFECT(); break;
  case  8 :
    prop.grade.set_PERFECT();
    if (prop.dist > 20)         // better replace that by a constant
      prop.grade.downgrade();
    break;
  case  4 :
    prop.grade.set_PERFECT();
    if (prop.dist > 20)
      prop.grade.downgrade();
    if (prop.dist > 16)
      prop.grade.downgrade();
    if (prop.dist > 12)
      prop.grade.downgrade();
    break;
  case  0 :
    prop.grade.set_BAD(); return prop.grade;
  default :
    prop.grade.set_BAD(); return prop.grade;   // should not happen
  }

  // if this method is called for the other team, swap sign of the x-coordinates
  // so that the following algorithm will work
  if (!we) {
    playerx  = -playerx;
    partnerx = -partnerx;
  }

  // partner in own third (-> defensive situation)
  if (partnerx < -FIELD_LENGTH/6) {
    // pass should take the ball away from goal
    if (sqrnorm(partner - ourGoal.pos) < sqrnorm(player - ourGoal.pos)) {
      prop.grade.downgrade();
    }
    // pass should not cross the line y=0
    if (fabs(playery) > 5 && playery * partnery < 0) {
      prop.grade.downgrade();
    }
    // no passing back
    if (partnerx - playerx < SAMELEVEL_LENGTH) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -8) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -15) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -20) {
      prop.grade.downgrade();
    }
  }  
  // partner in middle third (-> midfield situation)
  else if (partnerx < FIELD_LENGTH/4) {
    // pass should not cross the line y=0 if player is too far from it
    if (fabs(playery) > 10 && playery * partnery < 0) {
      prop.grade.downgrade();
    }
    // pass towards the middle of the field
    if (fabs(partnery) < fabs(playery) - 10) {
      prop.grade.downgrade();
    }
    // partner behind
    if (partnerx - playerx < SAMELEVEL_LENGTH) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -10) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -15) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -20) {
      prop.grade.downgrade();
    }
  }
  // partner in opponent's quarter (->offensive situation)  
  else {
    // partner too near to the border and too far in front
    if ( (1.5*(FIELD_LENGTH/2.0 - partnerx) +
	  (FIELD_WIDTH/2.0 - partnery) < (FIELD_WIDTH - GOAL_WIDTH)/2.0) ||
	 (1.5*(FIELD_LENGTH/2.0 - partnerx) +
	  (FIELD_WIDTH/2.0 + partnery) < (FIELD_WIDTH - GOAL_WIDTH)/2.0) ) {
      prop.grade.downgrade();
    }
    // partner too far behind
    if (partnerx - playerx < -20) {
      prop.grade.downgrade();
    }    
    // pass should take the ball nearer to the goal
    if (sqrnorm(partner - otherGoal.pos) > sqrnorm(player - otherGoal.pos)) {
      prop.grade.downgrade();
    }
  }

  return prop.grade;
}

////////////////////////////
//
// PassEval::betterPass
//
/////////////////////////// 

bool PassEval::betterPass(const Snapshot *wS, PropOf good, PropOf bad) const
{
  if (good.grade.is_better_than(bad.grade))
    return true;
  if (bad.grade.is_better_than(good.grade))
    return false;

  // ok, so the grades are the same
  if (good.grade.is_BAD())
    return false;

  int goodsum = 0;
  int badsum = 0;
  
  // player freestanding is better : 1 point
  if (good.free > bad.free)
    goodsum++;
  else if (bad.free > good.free)
    badsum++;

  bool defensive = ( min(good.pos.getX(), bad.pos.getX()) < 0.0 );

  // broader corridor : 4/2 points
  if (good.corridor > bad.corridor)
    goodsum += ( defensive ? 4 : 2 );
  else if(bad.corridor > good.corridor)
    badsum += ( defensive ? 4 : 2 );

  // larger x-value of pass target : 2/4 points
  if( good.pos.getX() > bad.pos.getX() )
    goodsum += ( defensive ? 2 : 4 );
  else
    badsum += ( defensive ? 2 : 4 );
  
  // better goalkicker : 8 points
  if (good.number != bad.number) {
    Grade goodGoalGrade = (goalkickeval->ourEval(wS, good.pos));
    Grade badGoalGrade  = (goalkickeval->ourEval(wS, bad.pos));
    
    if (goodGoalGrade.is_better_than(badGoalGrade))
      goodsum += 8;
    else if (badGoalGrade.is_better_than(goodGoalGrade))
      badsum += 8;
  }

  if (goodsum > badsum)
    return true;
  return false;
}

///////////////
//
// PassEval::coutPropOf
//
///////////////

void PassEval::coutPropOf(const Snapshot *wS, PropOf prop) const
{
  int t = wS->time;
  if (t > 0) 
    cout << t << " pass to number " << prop.number << " at pos " << prop.pos << " with dist " << prop.dist << endl
	 << t << "          free: " << prop.free << " corridor: " << prop.corridor << " offside: " << prop.offside << endl
	 << t << "          grade: " << prop.grade << endl 
	 << t << "      result is perhaps: " << prop.perhaps << endl;
	 
}
