//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "passmodule.h"
#include "passeval.h"
#include "goalkickeval.h"

/////////////////////////////
//
// class PassModule
//
///////////////////////////

////////////////////////////
//
// Constructor
//
///////////////////////////

PassModule::PassModule(PassEval* _passeval, const GoalKickEval* _goalkickeval):
  callMeAgain(false),             // why should this module be recalled?
  passeval(_passeval),
  goalkickeval(_goalkickeval),
  lastSnapshotTime(-1),           // module has never seen a snaphshot
  time_for_kick(0),
  target_selected(false),
  look_act(false),
  look_for_ball(false), 
  look_for_ball_time(-3),   // hope that's enough
  OPP_POWER_FACTOR(0.7)  // try it
{}

////////////////////////////
//
// Destructor
//
///////////////////////////

PassModule::~PassModule()
{}  

////////////////////////////////////////////////////////////////////////////////////
//
// public methods
//
////////////////////////////////////////////////////////////////////////////////////

////////////////////////////
//
// bool PassModule::feedbackWorld
//
///////////////////////////

bool PassModule::feedbackWorld(const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap)
{
  // process new snapshot 
  updateDataStorage(worldSnap);
  
  // just to suppress the unsused-parameter-warnings
  if (tacticalSnap);
  
  // if pass is over check if it failed
  if (look_for_ball)
    {
      // if ball velocity does not point to target something must be wrong
      if ( fabs( angleBetween( worldSnap->ball.v, target_pos - worldSnap->my.pos ) ) > PI_4 ||
	   worldSnap->ball.v.getLength() < 0.1 )   // collision?
	{
	  // tactical action can not stay TA_PASS
	  if( tacticalSnap->ourAction.state == TA_PASS &&
	      tacticalSnap->ourAction.performer == worldSnap->my.number )
	    {
	      PASSTRACE( cout << worldSnap->time <<
			 " PASS: Pass seems to have gone bananas -> change tacticalAction" << endl; );

	      tacticalSnap->ourAction.state = TA_UNKNOWN;
	      tacticalSnap->ourAction.quality = TQ_GUESSED;
	      tacticalSnap->ourAction.startingTime = worldSnap->time;
	      tacticalSnap->ourAction.performer = UNKNOWN;
	      tacticalSnap->ourAction.receiver = UNKNOWN;
	    }
	}
    }

  return true;
}

///////////////////////////
//
// bool PassModule::feedbackAudio
//
///////////////////////////

bool PassModule::feedbackAudio(AudioMsg &audioMsg)

{
  bool hasInfo;

  audioMsg >> hasInfo;

  if (hasInfo) {
    PASSTRACE(cerr << "PASS: received message but never sent one!" << endl;);
  }

  return true;
}

////////////////////////////
//
// Grade PassModule::evaluate
//
///////////////////////////

Grade PassModule::evaluate(AdvSkills *skills, const Snapshot *worldSnap,
                           TacticalSnapshot *tacticalSnap)
{
  // deactivate the last target
  ///////////////////////////////
  target_selected = false;

  // just to suppress the unsused-parameter-warnings
  if (tacticalSnap);
 
  // if we have a newer snapshot, process it
  if (worldSnap->time > lastSnapshotTime) { 
    updateDataStorage(worldSnap);
    PASSTRACE(cerr << lastSnapshotTime << " "
                   << "PASS: eval -> world snapshot changed from feedback to evaluate" 
	           << endl;);
  }  

  Grade grade;  // storage for the evaluated grade
  
  // do we need a rekick?
  ///////////////////////////////
  if (callMeAgain && skills->kickSafe(worldSnap)) {
    // action: rekick 
    target_selected = true; // just for safety!
    PASSTRACE(cout << lastSnapshotTime << " "
                   << "PASS: eval -> rekick needed => ABSOLUTE" 
	           << endl;);    
    grade.set_ABSOLUTE();
    return grade;
  }
   
  // thought a rekick is needed, but the ball is not kickable
  // kick finished surprisingly or ball is lost!?
  if (callMeAgain && !skills->kickSafe(worldSnap)) {
    callMeAgain = false;
    look_for_ball = false;
    PASSTRACE(cout << lastSnapshotTime << " "
	           << "PASS: eval -> don't know if kick is ready or ball is lost"
	           << endl
	           << lastSnapshotTime << " "
	           << "              don't look for ball"              
	           << endl;);
    look_act = false;
    target_selected = false;
  }

  // time for looking after ball is over
  if ( (look_for_ball) && (look_for_ball_time + 2 < lastSnapshotTime) ) {
    look_for_ball = false;
    PASSTRACE(cout << lastSnapshotTime << " "
		       << "PASS: eval -> don't want to look for the ball anymore"
		       << endl;);
  }

  look_act = false;
  // we have to look for the ball
  if ( (look_for_ball) && (look_for_ball_time + 2 >= lastSnapshotTime) ) {
    PASSTRACE(cout << lastSnapshotTime << " "
	           << "PASS: eval -> want to look for ball => ABSOLUTE"
	           << endl;);
    look_act = true;
    grade.set_ABSOLUTE();
    return grade;
  }
    
  // we don't have an emergency-kick-away (anymore, by now...)
  em_case = false;
 
  // technical impossibility
  ///////////////////////////////
  if ( !skills->kickSafe(worldSnap) ) { 
    PASSTRACE(cout << lastSnapshotTime << " "
                   << "PASS: eval -> can't kick the ball => IMPOSSIBLE" << endl);  
    callMeAgain = false;     // no recall needed
    grade.set_IMPOSSIBLE(); 
    return grade;
  }

  // just here we're able to KICK THE BALL
  ///////////////////////////////////////////
 
  // information gathering
  /////////////////////////////

  // time available for kick
  time_for_kick = passeval->check_for_safekick(worldSnap, worldSnap->my.number);
 
  // as we probably want to pass in what follows, we  
  // gather some information first:
  
  // fill the propertyof-array and the rangepassof-array
  getPasses(worldSnap);
  
  PropOf prop;
  prop.grade.set_BAD();

  // try to get best p-pass and best r-pass as partner
  for (int l = 1; l <= 11; l++) {
    if (pprops[l].perhaps)
      if (passeval->betterPass(worldSnap, pprops[l], prop)) {
	prop = pprops[l];
	target_breed = true;
      }
    if (rprops[l].perhaps)
      if (passeval->betterPass(worldSnap, rprops[l], prop)) {
	prop = rprops[l];
	target_breed = false;
      }
  }
  
  // no p- and no r-pass found, so it's an emergency-kick-away-do-something
  if (prop.grade.is_BAD()) {
    em_case = true;
    em_vec = get_ePass( worldSnap, tacticalSnap );

    PASSTRACE(cout << lastSnapshotTime;);
    PASSTRACE(cout << " PASS: eval -> no pass found, emergency-kick-away considered => BAD" << endl;);
    grade.set_BAD();
    return grade;
  }
  
  lockPass(prop);  

  // change intern grade to extern grade
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!! ATTENTION !!!!!!!!!!!!! ATTENTION !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  // remove this if not compatible with other modules
  // if (prop.grade.is_PERFECT())
  //   prop.grade = get_externGrade(worldSnap, worldSnap->my.pos, target_pos, prop.grade, time_for_kick);

  // downgrade pass if player is the goalie and should catch instead
  if (prop.grade.is_PERFECT() && worldSnap->my.number == worldSnap->ourGoalieNo) {
    if (worldSnap->playMode == PLAY_MODE_PLAY_ON)
      prop.grade.downgrade();
  }

  if (prop.grade.is_GOOD() &&
      goalkickeval->ourEval(worldSnap, target_pos).is_PERFECT())
    prop.grade.upgrade();

  PASSTRACE(
	    if (worldSnap->time > 0) {
	      cout << lastSnapshotTime << " PASS: eval -> found a partner:" << endl;
	      passeval->coutPropOf(worldSnap, prop);
	      cout << lastSnapshotTime << " time for kick : " << time_for_kick;
	      if (!target_breed)
		cout << " ; it is a rangepass!";
	      cout << endl;
	    } );

  grade = prop.grade;  
  return grade;
}  // PassModule::eval

////////////////////////////
//
// PassModule::act
//
///////////////////////////

bool PassModule::act(AdvSkills *skills, const Snapshot *worldSnap,
                     TacticalSnapshot *tacticalSnap)
{
  // if we have a newer snapshot, process it
  if (worldSnap->time > lastSnapshotTime) { 
    updateDataStorage(worldSnap);
    PASSTRACE(cerr << lastSnapshotTime << " " 
	           << "PASS: act -> worldsnapshot changed from evaluate to act" 
                   << endl;);
  }  

  // we have to do an emergency-kick-away
  if (em_case) {
    PASSTRACE(cout << lastSnapshotTime << " "
	      << "PASS: act -> performed emergency-kick-away; direction = "
	      << RAD_TO_DEG( em_vec.getAngle() ) << endl);
    kickparams = skills->kick(worldSnap, em_vec, AdvSkills::KICKAWAY);
    skills->exec(worldSnap, kickparams);
    return true;
  }

  if (look_act) {  // look for the ball; don't do anything else
    AdvSkills::TurnHeadCommand turnheadparams = skills->turnHead(worldSnap, worldSnap->my.bodyDir);
    skills->exec(worldSnap, turnheadparams);
    PASSTRACE(cout << lastSnapshotTime << " "
	           << "PASS: act -> turned head to body-direction"
	           << endl;);
    
    AdvSkills::TurnCommand turnparams = skills->turn(worldSnap,
						     (worldSnap->ball.pos - worldSnap->my.pos).getAngle());
    skills->exec(worldSnap, turnparams);
    PASSTRACE(cout << lastSnapshotTime << " "
	           << "PASS: act -> turned body to ball to look at it"
	           << endl;);
    DRAWTRACE(LINE(target_source,target_pos));
    DRAWTRACE(VLINE(worldSnap->ourOffsideLine(MAXAGE)));
    DRAWTRACE(CORRIDOR(target_source,target_pos,target_corridor,PI_6));

    return true;
  }

  if(!(skills->kickSafe(worldSnap))){
    callMeAgain = false;
    skills->noAction();
    PASSTRACE(cout << lastSnapshotTime << " "
	           << "PASS: act -> did no action because ball not kickable"
	           << endl;);
    return false;
  }

  // the kick is not finished, so perform a rekick
  if (callMeAgain) {
    kickparams = skills->rekick(worldSnap, kickparams);
    skills->exec(worldSnap, kickparams);
    if (kickparams.state == AdvSkills::KFINISHING) {  // the kick is finished
      callMeAgain = false;
      look_for_ball = true;                  // look after the ball
      look_for_ball_time = lastSnapshotTime; // from now on (for 2 cycles)
    }
    DRAWTRACE(LINE(target_source,target_pos));
    DRAWTRACE(VLINE(worldSnap->ourOffsideLine(MAXAGE)));
    DRAWTRACE(CORRIDOR(target_source,target_pos,target_corridor,PI_6));
    PASSTRACE(cout << lastSnapshotTime << " "
	           << "PASS: act -> did a rekick" 
	           << endl;
	      if (kickparams.state == AdvSkills::KFINISHING)
	        cout << lastSnapshotTime << "       finished" << endl;
              else 
                cout << lastSnapshotTime << "       not finished" << endl;
	      );
    
    return true;             
  }

  if (target_selected) {
    target_source = worldSnap->my.pos;

    double end_vel;   // ball velocity when it reaches the partner

    // get the kickparams
    if (target_breed)
      // playerpass => ball should have vel. == 0.8 when reaching partner
      end_vel = ( MAXPOWER * KICK_POWER_RATE /
		  ( dist( target_source, target_pos ) < passeval->SHORTPASS_LENGTH ? 4 : 2 ) );
    else
      // rangepass => ball should leave with calculated velocity
      end_vel = min( target_vel, MAXPOWER * KICK_POWER_RATE /
		     ( dist( target_source, target_pos ) < passeval->SHORTPASS_LENGTH ? 4 : 2 ) );

    kickparams = skills->kick(worldSnap, end_vel, target_pos, time_for_kick);
    
    // perform the kick
    skills->exec(worldSnap, kickparams);

    // report the kick
    tacticalSnap->reportOurTacticalAction( TA_PASS, worldSnap->my.number, target_number );

    // if we want to look for the ball later, here's the pos
    look_for_ball_pos = target_pos;

    if (kickparams.state != AdvSkills::KFINISHING) {           // pass is not finished
      callMeAgain = true;                                      // prepare for rekick
      PASSTRACE(cout << lastSnapshotTime << " "
	             << "PASS:        not finished" << endl;);
    }
    else {                                                     // kick is ready by now 
      look_for_ball = true;                                    // prepare for looking after ball
      look_for_ball_time = lastSnapshotTime;
      PASSTRACE(cout << lastSnapshotTime << " "
	             << "PASS:        finished" << endl;);
    }
    
    DRAWTRACE(LINE(target_source,target_pos));
    //DRAWTRACE(CIRCLE(target_pos, passeval->DANGERZONE));
    DRAWTRACE(VLINE(worldSnap->ourOffsideLine(MAXAGE)));
    DRAWTRACE(CORRIDOR(target_source,target_pos,target_corridor,PI_6));
    
    PASSTRACE(cout << lastSnapshotTime << " "
	      << " PASS: act -> trying to pass to " << target_number << " at pos" << target_pos 
	      << " with " << time_for_kick << " cycles time to kick";);
    PASSTRACE(if (!target_breed)
	      cout << endl 
                   << lastSnapshotTime << "       this is a rangepass" << endl;);
    PASSTRACE(cout << endl;);

    return true;
  }

  PASSTRACE(cout << lastSnapshotTime << " "
	    << "PASS: act -> shit, no action sent" << endl;); 
  skills->noAction(); // to be changed
  return false;
} // PassModule::act

////////////////////////////
//
// PassModule::getSayMsg
//
///////////////////////////

bool PassModule::getSayMsg(AudioMsg &audioMsg)
{
  bool sendSomething = false;

  if (sendSomething) {

    audioMsg.markWrite();
    
    audioMsg << true;

    // audioMsg << sendInfo;
    
    // check if error occurred
    if( !audioMsg.getState() )
      {
        cerr << "PASS: Stream overflow: " << audioMsg << endl;
        audioMsg.rollbackWrite().clearState();
	
        audioMsg << false;
	
      }
  } 
  else {
    
    audioMsg << false;

  }

  return true;;
}

////////////////////////////////////////////////////////////////////////////////////
//
// private methods
//
////////////////////////////////////////////////////////////////////////////////////

////////////////////////////
//
// PassModule::updateDataStorage
//
///////////////////////////

bool PassModule::updateDataStorage(const Snapshot *wS)
{
  // get new snapshot-time
  lastSnapshotTime = wS->time;

  return true;
}

////////////////////////////
//
// bool PassModule::getPasses
//
///////////////////////////

bool PassModule::getPasses(const Snapshot *wS)
{
  TESTPASSTRACE( int t = wS->time; );

  for (int i = 1; i <= 11; i++) {
    TESTPASSTRACE(cout << t << " These are the passes for player " << i << endl );
    pprops[i].perhaps = false;
    pprops[i].grade.set_BAD();
    rprops[i].perhaps = false;
    rprops[i].grade.set_BAD();
    
    // we don't want to pass to ourselves
    if( i == wS->my.number )
      TESTPASSTRACE( cout << t << " This is my own number!" << endl );
    else
      // we don't want to pass to players that are older than MAXAGE
      if( wS->ourTeam(i).age > MAXAGE )
	TESTPASSTRACE( cout << t << " This player is too old!" << endl );
      else
	// we don't want to pass to our goalie
	if( i == wS->ourGoalieNo )
	  TESTPASSTRACE( cout << t << " This player is the goalie!" << endl );
	else
	  // if we do not know our goalie yet we don't want to pass to players that are close to our goal
	  if( wS->ourGoalieNo == UNKNOWN && wS->ourTeam(i).pos.getX() < - FIELD_LENGTH/6 )
	    TESTPASSTRACE( cout << t << " I don't know our goalie number"
			   << "and this player is too close to our goal!" << endl );
	  else
	    {
	      // in all other cases: get player passes and range passes
	      passeval->get_pPass(true, wS, wS->my.number, wS->ourTeam(i).pos, i, pprops[i]);
	      passeval->get_rPass(true, wS, wS->my.number, i, rprops[i]);
	      TESTPASSTRACE( cout << t << " Evaluation of playerpass:" << endl; );
	      TESTPASSTRACE( passeval->coutPropOf(wS, pprops[i]); );
 	      TESTPASSTRACE( cout << t << " Best rangepass:" << endl; );
	      TESTPASSTRACE( passeval->coutPropOf(wS, rprops[i]); );
	    }
  }

  return true;
} // getPasses

////////////////////////////
//
// PassModule::get_ePass
//
///////////////////////////

Vector2d PassModule::get_ePass(const Snapshot *wS, TacticalSnapshot *tS )
{
  // 9 possible directions for emergency kick
  Angle kickangle[9] = { -PI_2, -PI_4 - PI_8, -PI_4, -PI_8, 0.0, PI_8, PI_4, PI_4 + PI_8, PI_2 };

  // evaluations of the 9 possible directions
  Grade dir[9];
  for( int i = 0; i < 9; i++ )
    dir[i].set_PERFECT();

  Vector2d my_pos = wS->my.pos;

  const int AREA_LEFT = 1;
  const int AREA_MIDDLE = 2;
  const int AREA_RIGHT = 3;

  int my_area = ( ( my_pos.getY() < -FIELD_WIDTH / 3.0 ) ? AREA_LEFT :
		  ( my_pos.getY() > FIELD_WIDTH / 3.0 ) ? AREA_RIGHT : AREA_MIDDLE );
  
  SituationState my_situation = ( ( tS->ourSituation != SI_UNKNOWN ) ?  tS->ourSituation : 
				  ( my_pos.getX() < -FIELD_LENGTH / 3.0 ) ? SI_DEFENSIVE :
				  ( my_pos.getX() > FIELD_LENGTH / 3.0 ) ? SI_OFFENSIVE : SI_MIDFIELD );
  
  // suppress certain directions due to situation and my position
  switch( my_situation )
    {
    case SI_DEFENSIVE :
      switch( my_area )
	{
	case AREA_LEFT   :
	  dir[5].set_BAD(); dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	  break;
	case AREA_MIDDLE :
	  break;
	case AREA_RIGHT  :
	  dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD(); dir[3].set_BAD();
	  break;
	}
      break;
      
    case SI_MIDFIELD :
      switch( my_area )
	{
	case AREA_LEFT   :
	  dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	  dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	  break;
	case AREA_MIDDLE :
	  dir[0].set_BAD(); dir[1].set_BAD();
	  dir[7].set_BAD(); dir[8].set_BAD();
	  break;
	case AREA_RIGHT  :
	  dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	  dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	  break;
	}
      break;
      
    case SI_OFFENSIVE :
      switch( my_area )
	{
	case AREA_LEFT   :
	  dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD(); dir[3].set_BAD();
	  break;
	case AREA_MIDDLE :
	  dir[0].set_BAD(); dir[1].set_BAD();
	  dir[7].set_BAD(); dir[8].set_BAD();
	  break;
	case AREA_RIGHT  :
	  dir[5].set_BAD(); dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	  break;
	}
      break;
      
    default :   // should not happen
      break;
    }
  
  // first check if a short and narrow corridor is free
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_PERFECT() )
      if( wS->otherTeamInCorridor( my_pos, my_pos + radAng2Vector2d( 15.0, kickangle[i] ), 5.0, PI_2 ) )
	dir[i].set_POOR();
  
  // ...then try it with a longer and broader one
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_PERFECT() )
      if( wS->otherTeamInCorridor( my_pos, my_pos + radAng2Vector2d( 17.0, kickangle[i] ), 8.0, PI_2 ) )
	dir[i].set_AVERAGE();
  
  // ...finally test the longest and broadest corridor
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_PERFECT() )
      if( wS->otherTeamInCorridor( my_pos, my_pos + radAng2Vector2d( 20.0, kickangle[i] ), 10.0, PI_2 ) )
	dir[i].set_GOOD();
  
  // search for best grade
  int bestDir = ( my_pos.getY() < 0 ? 0 : 8 );
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_better_than( dir[bestDir] ) )
      bestDir = i;
  
  // determine kick vector
  return radAng2Vector2d( BALL_SPEED_MAX, kickangle[bestDir] );
}

////////////////////////////
//
// PassModule::lockPass
//
///////////////////////////

Grade PassModule::lockPass(PropOf prop)
{
  target_selected = true;
  target_number = prop.number;
  target_pos = prop.pos;
  target_vel = prop.vel;
  target_corridor = prop.corridor;

  return prop.grade;
}

////////////////////////////
//
// PassModule::get_externGrade
//
///////////////////////////

Grade PassModule::get_externGrade(const Snapshot *wS, Vector2d player, Vector2d partner,
				  Grade internGrade, int playerTimeToKick) const
{
  Vector2d w( -(partner.getY()-player.getY()), (partner.getX() - player.getX()) );
  
  Grade maxGrade;
  maxGrade.set_PERFECT();
  bool downgrade = false;
  
  for (int i = 1; i <= 11; i++) 
    if (wS->otherTeam(i).age < MAXAGE) {
      double a = (w.getX() * (player.getY() - wS->otherTeam(i).pos.getY()) + 
		  w.getY() * (wS->otherTeam(i).pos.getX() - player.getX()))    /
	(w.getX()*w.getX() + w.getY()*w.getY());
      double b;
      if (w.getX() != 0)
	b = (wS->otherTeam(i).pos.getX() - a*w.getY() - player.getX())/(w.getX());
      else
	b = (wS->otherTeam(i).pos.getY() + a*w.getX() - player.getY())/(w.getY());
      
      // test, if the meetingpoint of ball and opponent lies on the passline
      if ( (0<b) && (b<1) ) {
	Vector2d m( (player.getX()+b*w.getX()), (player.getY()+b*w.getY()) );
	
	double dist_o = dist(wS->otherTeam(i).pos, m);
	int time_o = 1 - playerTimeToKick;
	bool stopit = false;
	
	// time_o that opponent needs to reach m
	while ( (dist_o >= OPP_POWER_FACTOR*MAXPOWER/(1-PLAYER_DECAY) *
		 (time_o - 
		  PLAYER_DECAY * ( (1-pow(PLAYER_DECAY,time_o))/(1-PLAYER_DECAY) ) )) &&
		(!stopit) ) {
	  time_o++;
	  if (time_o > 50)
	    stopit = true;
	}
	
	// ...
	time_o = max(1, time_o);
	
	// velocity ball needs to be "faster" than ball
	double ball_v = dist(player, m)*(1 - BALL_DECAY)/(1 - pow(BALL_DECAY,time_o));
	
	if ( (ball_v > BALL_SPEED_MAX) && (internGrade.is_PERFECT()) )
	  downgrade = true;
      } // if b in (0,1)
    } // if MAXAGE
  
  if (downgrade) {
    TESTPASSTRACE(cout << wS->time << " changed intern grade " << internGrade);
    internGrade.downgrade();
    TESTPASSTRACE(cout << " to " << internGrade << endl);
  }
  else 
    TESTPASSTRACE(cout << wS->time << " kept intern grade " << internGrade << " as extern grade" << endl);
  
  return internGrade;
} // get_externGrade

