//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                              //
//                                                                                              //
//                             POSITIONMODULE by F. Schulz                                      //
//                                                                                              //
//                                                                                              //
//////////////////////////////////////////////////////////////////////////////////////////////////
#include "PositionModule.h"
#include "constants.h"
#include "minmax.h"

PositionModule::PositionModule(bool _isGoalie, 
			       const PassEval *_passEval,  
			       const GoalKickEval *_goalKickEval) : 
  posBallControl(POS_BC_WE),
  posGoal(0,0),
  oldPosGoal(0,0),
  middleOfGoal(-FIELD_LENGTH/2, 0), 
  priority(3), 
  nearestFriendNumber(UNKNOWN), 
  internalTime(0), 
  passEval(_passEval), 
  goalKickEval(_goalKickEval),
  myMaxX(FIELD_LENGTH/2-1), 
  myMaxY(FIELD_WIDTH/2-1),
  myMinX(-FIELD_LENGTH/2+1),
  myMinY(-FIELD_WIDTH/2+1),
  nearestOther(UNKNOWN),
  oldX(0),
  oldY(0)
{
  isGoalie = _isGoalie;
  posStamina[0][0] = 100;
  posStamina[0][1] =  80;
  posStamina[0][2] =  50;
  posStamina[0][3] =  30;
  posStamina[0][4] =  10;
  posStamina[1][0] = 100;
  posStamina[1][1] =  70;
  posStamina[1][2] =  30;
  posStamina[1][3] =  20;
  posStamina[1][4] =  10;
  posStamina[2][0] =  80;
  posStamina[2][1] =  50;
  posStamina[2][2] =  25;
  posStamina[2][3] =  15;
  posStamina[2][4] =   0;

  xPoints = Config::getValueAsInt("pos_xPoints");
  yPoints = Config::getValueAsInt("pos_yPoints");
  // differences in x resp. y axis
  double xDiff;
  double yDiff;
  xDiff = Config::getValueAsDouble("pos_xDiff");
  yDiff = Config::getValueAsDouble("pos_yDiff");
  
  double minX = -(xPoints-1)*xDiff/2;
  double minY = -(yPoints-1)*yDiff/2;
  
  evalGrid = new (Vector2d*)[xPoints];
  for(int i = 0; i < xPoints; i++) 
    evalGrid[i] = new Vector2d[yPoints];
  // calculate the positions
  for(int x = 0; x < xPoints; x++)
    for(int y = 0; y < yPoints; y++){
      evalGrid[x][y].setX(minX + x*xDiff);
      evalGrid[x][y].setY(minY + y*yDiff);
    }
    
}

PositionModule::~PositionModule()
{
  delete [] evalGrid;
}

bool PositionModule::feedbackWorld(const Snapshot* wS, TacticalSnapshot* tS)
{
  int i;
  
  // just to undo the unsused-parameter-warnings
  if (wS==NULL){
    cerr << wS->time << "POS: feedback -> no wS available" << endl;
    return false;
  }

  // is there a tacticalSnap?
  if(tS == NULL){
    cerr << wS->time << "POS: feedback -> no tacticalSnap available" << endl;
    return false;
  }
  
  // Now everything is checked and fine.
  switch(tS->ourBallControl){
  case BC_SAFE: POSTRACE("BC_SAVE"); break;
  case BC_DANGER: POSTRACE("BC_DANGER"); break;
  case BC_LOST: POSTRACE("BC_LOST"); break;
  case BC_UNKNOWN: POSTRACE("BC_UNKNOWN"); break;
  }
  // We have ball control iff:
  // BC_SAFE and responsible is known
  //         and resposible is not goalie
  //             or goalie is free
  // or in the diffrent free kick us modes
  // or BC_DANGER and we have had the ball
  //              and responsible is not goalie
  // or BC_UNKNOWN and we have had the ball
  //               and responsible s not goalie
  if(
     ((tS->ourBallControl == BC_SAFE)
      &&(tS->firstResponsiblePlayer != UNKNOWN)
      &&((tS->firstResponsiblePlayer != wS->ourGoalieNo)
	 ||(ourPlayerFree(wS, wS->ourGoalieNo)))
      )
     ||(wS->playMode == PLAY_MODE_KICK_OFF_US)
     ||(wS->playMode == PLAY_MODE_KICK_IN_US)
     ||(wS->playMode == PLAY_MODE_FREE_KICK_US)
     ||(wS->playMode == PLAY_MODE_CORNER_KICK_US)
     ||(wS->playMode == PLAY_MODE_GOAL_KICK_US)
     ||((tS->ourBallControl == BC_DANGER)
	&&(posBallControl == POS_BC_WE)
	&&(tS->firstResponsiblePlayer != wS->ourGoalieNo))
     ||((tS->ourBallControl == BC_UNKNOWN)
	&&(posBallControl == POS_BC_WE)
	&&(tS->firstResponsiblePlayer != wS->ourGoalieNo))
     )
    { // our ball     
      posBallControl = POS_BC_WE;
    }
  else
    { //their ball
      posBallControl = POS_BC_THEY;
    }

  oldPosGoal = posGoal;
  
  // SET OFFSIDE VARIABLES

  ourOffsideLine = wS->ourOffsideLine(MAXAGE);
  otherOffsideLine = wS->otherOffsideLine(MAXAGE);

  otherTeamMinX = 0;
  for(i = 1; i <= 11; i++) 
    if(wS->otherTeam(i).age < MAXAGE)
      // player i exists ?
      if(wS->otherTeam(i).pos.getX() < otherTeamMinX){ 
	otherTeamMinX = wS->otherTeam(i).pos.getX();
      }

  otherTeamMaxX = 0;
  for(i = 1; i <= 11; i++) 
    if((wS->otherTeam(i).age < MAXAGE)&&(i != wS->otherGoalieNo)){
      // player i exists
      if(wS->otherTeam(i).pos.getX() > otherTeamMaxX){ 
	otherTeamMaxX = wS->otherTeam(i).pos.getX();
      }
    }
  
  POSTRACE("ourOffsideLine " << ourOffsideLine << " otherTeamMaxX " << otherTeamMaxX);

  // shorten the code
  me = wS->my;

  nearestFriendNumber = wS->ourTeamClosestTo(me.pos,20,2);

  internalTime++; 

  myTactNo = tS->ourTeamTacticalNo(me.number);
  if(myTactNo == UNKNOWN) myTactNo = me.number;
  
  nearestOther = wS->otherTeamClosestTo(me.pos);

  calculateMaximalPositions(wS, tS);
  
  return true;
} 

bool PositionModule::feedbackAudio(AudioMsg &audioMsg)
{
  PositionModuleInfo receivedInfo(false);
  audioMsg >> receivedInfo;
  if (receivedInfo.valid)
    {
      POSTRACESHORT("got valid info");
      // evaluate
    }
  else
    {
      POSTRACESHORT("got invalid info");
    }
  
  return true;
}

Grade PositionModule::evaluate(AdvSkills* skills,
			       const Snapshot* wS,
			       TacticalSnapshot* tS)
{
  Grade g;
  g.set_BAD();
  goToPrecision = 1.0;
   
  // standard position is my position
  posGoal = me.pos;
    
  // do we have a valid snapshot?
  if(wS == NULL)
    {
      cerr << "POS: eval -> no wS available" << endl;
      g.set_IMPOSSIBLE();
      return g;
    }
  
  // are the skills available?
  if(skills == NULL)
    {
      cerr << wS->time << "POS: eval -> no skills available" << endl;
      g.set_IMPOSSIBLE();
      return g;
    }
  
  // is there a tacticalSnap?
  if(tS == NULL){
    cerr << wS->time << "POS: eval -> no tacticalSnap available" << endl;
    g.set_IMPOSSIBLE();
    return g;
  }

  // Now everything is checked and fine, so we can start to think.
  
  // my ball ?
  if (skills->kickSafe(wS)){ 
    POSTRACE("I am responsible for ball");
    posGoal = wS->ball.pos;
    priority = 2;
    goToPrecision = KICKABLE_MARGIN/2;
    g.set_BAD();
    return g;
  }
    
  if(isGoalie){ // treat goalie diffrent
    double goalDistance = 5.0; // how far is the ball or the nearest opponent from goal?
    int nearestOpp = wS->otherTeamClosestTo(middleOfGoal);

    if(nearestOpp != UNKNOWN){
      goalDistance = min(dist(wS->ball.pos, middleOfGoal),
			 dist(wS->otherTeam(nearestOther).pos, middleOfGoal));
    }
    else
      goalDistance = wS->ball.pos.getX() + FIELD_LENGTH/2;

    // calculation of position (between ball and middle of goal)
    // normaly half way ball <-> goal
    // but not out of penalty area
    posGoal = middleOfGoal
      + (min(goalDistance/2, PENALTY_AREA_LENGTH)
	 /norm(wS->ball.pos - middleOfGoal))
      *(wS->ball.pos - middleOfGoal);
    POSTRACE("my calculated pos: " << posGoal);

    if(me.pos.getX() < -FIELD_LENGTH/2 + PENALTY_AREA_LENGTH)
      goToPrecision = max(2*PLAYER_SIZE, goalDistance/10);
    else
      goToPrecision = 2*PLAYER_SIZE;
    priority = 1;

    if(dist(posGoal, me.pos) > 3.0){
      g.set_AVERAGE();
      return g;
    }
    g.set_AVERAGE();
    return g;
  } // if(isGoalie)

  if(posBallControl == POS_BC_WE){
    // OUR BALL
    
    // should not happen and is very stupid
    if (tS->firstResponsiblePlayer == UNKNOWN){ 
      POSTRACE("Resp. player unknown");
      posGoal = wS->ball.pos;
      priority = 4;
      goToPrecision = KICKABLE_MARGIN/2;
      g.set_BAD();
      return g;
    }
    
    // pass receiver => not my job
    if((tS->ourAction.state == TA_PASS)
       &&(tS->ourAction.receiver == me.number)
       ){
      POSTRACE("my ball");
      posGoal = wS->ball.pos;
      priority = 0; 
      goToPrecision = 0.5;
      g.set_BAD();
      return g;
    }
    
    // NOT MY BALL
    Grade gradeOfPlayer[12];
    Vector2d bestPosOfPlayer[12];
    
    // evaluate all players
    for(int num = 1; num <= 11; num++){
      gradeOfPlayer[num].set_IMPOSSIBLE();
      if((wS->ourTeam(num).age < MAXAGE)
	 &&(wS->ourGoalieNo != num)
	 &&(num != tS->firstResponsiblePlayer)
	 ){
	gradeOfPlayer[num] = passEval->ourEval(wS, tS->firstResponsiblePlayer,
					       num, bestPosOfPlayer[num]);
      }
    }
    
    // how many good pass receivers ?
    int countGoodPassPartners = 0;
    for(int num = 1; num <= 11; num++){
      if(gradeOfPlayer[num].is_PERFECT()||gradeOfPlayer[num].is_GOOD()){
	countGoodPassPartners++;
	POSTRACE("PASSPARTNER: " << num);
      }
    }

    bool iAmPassPartner = ((gradeOfPlayer[me.number].is_PERFECT())
			   ||(gradeOfPlayer[me.number].is_GOOD()));
    
    // true, if more then two pass receivers
    bool enoughPassPartners = (((iAmPassPartner)&&(countGoodPassPartners >= 1))
			       ||((!iAmPassPartner)&&(countGoodPassPartners >= 2))) ? true : false;
    
    if(iAmPassPartner){
      posGoal = bestPosOfPlayer[me.number];
      POSTRACE("I am pass partner");
      POSDRAWTRACE(CIRCLE(posGoal, 2*PLAYER_SIZE));
      priority = 0;
      goToPrecision = 1.5;
      g.set_PERFECT();
      return g;
    }
    
    // not enough pass receivers
    // detect up to three players, which will try to become pass receivers 
    if(!enoughPassPartners){
      POSTRACE("Nicht gengend Passpartner");
      // next three players to ball (except responsible);
      int first  = UNKNOWN;
      int second = UNKNOWN;
      int third = UNKNOWN;

      int j = 1;
      int dummy = wS->ourTeamClosestTo(wS->ball.pos, MAXAGE, j);
      while((j<11)
	    &&((dummy == UNKNOWN)
	       ||(dummy == tS->firstResponsiblePlayer) // not pass player
	       ||(wS->ourTeam(dummy).pos.getX() < wS->ball.pos.getX()) // not behind ball
	       ||(dist(wS->ourTeam(dummy).pos, wS->ball.pos) > 30) // not far away
	       ||(wS->ourGoalieNo == dummy)) // not goalie
	    ){
	j++;
	dummy =  wS->ourTeamClosestTo(wS->ball.pos, MAXAGE, j);
      }
      // same procedure another two times:
      if(j <= 11) first = dummy;
      j++;
      if(j <= 11){
	dummy = wS->ourTeamClosestTo(wS->ball.pos, MAXAGE, j);
	while((j<11)
	      &&((dummy == UNKNOWN)
		 ||(dummy == tS->firstResponsiblePlayer)
		 ||(wS->ourTeam(dummy).pos.getX() < wS->ball.pos.getX())
		 ||(dist(wS->ourTeam(dummy).pos, wS->ball.pos) > 30)
		 ||(wS->ourGoalieNo == dummy))
	      ){
	  j++;
	  dummy =  wS->ourTeamClosestTo(wS->ball.pos, MAXAGE, j);
	}
      }
      if(j <= 11) second = dummy;      
      j++;      
      if(j <= 11){
	dummy = wS->ourTeamClosestTo(wS->ball.pos, MAXAGE, j);
	while((j<11)
	      &&((dummy == UNKNOWN)
		 ||(dummy == tS->firstResponsiblePlayer)
		 ||(wS->ourTeam(dummy).pos.getX() < wS->ball.pos.getX())
		 ||(dist(wS->ourTeam(dummy).pos, wS->ball.pos) > 30)
		 ||(wS->ourGoalieNo == dummy))
	      ){
	  j++;
	  dummy = wS->ourTeamClosestTo(wS->ball.pos, MAXAGE, j);
	}
      }
      if(j <= 11) third = dummy;

      POSTRACE("1: " << first << " 2: " << second << " 3: " << third);
      // me.number in {first, second, third}?  => responsible
      if((me.number == first)||(me.number == second)||(me.number == third)
	 ){
	double bX = wS->ball.pos.getX();
	double bY = wS->ball.pos.getY();
	if(me.pos.getY() < wS->ball.pos.getY()) // left of ball
	  posGoal =  calculateFreePosition(wS,
					   max(max(-FIELD_LENGTH/2+5,myMinX), bX-2),
					   min(min(FIELD_LENGTH/2-5, myMaxX), bX+FIELD_LENGTH/5),
					   max(max(-FIELD_WIDTH/2+5, myMinY), bY-FIELD_WIDTH/2),
					   min(min(FIELD_WIDTH/2-5, myMaxY), bY)
					   );
	else // right
	  posGoal =  calculateFreePosition(wS,
					   max(max(-FIELD_LENGTH/2+5, myMinX), bX-2),
					   min(min(FIELD_LENGTH/2-5, myMaxX), bX+FIELD_LENGTH/5),
					   max(max(-FIELD_WIDTH/2+5, myMinY), bY),
					   min(min(FIELD_WIDTH/2-5, myMaxY), bY+FIELD_WIDTH/2)
					   );
	  
	priority = 0;
	goToPrecision = 1.5;
	g.set_PERFECT();
	return g;
      }
    } // if(!enoughPassPartners)
    
    //nothing to do with ball
    
    if(me.pos.getX() > ourOffsideLine){ //offside => go home (with fixed y position)
      posGoal.setX(ourOffsideLine-0.8);
      posGoal.setY(me.pos.getY());
      goToPrecision = 0.5;
      if(tS->ourTeamDefender(myTactNo)) priority = 4;
      else if (tS->ourTeamMidfield(myTactNo)) priority = 3;
      else if (tS->ourTeamOffender(myTactNo)) priority = 0;
      g.set_GOOD();
      POSTRACE("Leaving Offside");
      return g;
    }
    //NO offside

    if(me.pos.getX() < otherTeamMinX){ // enemy not offside => generate offside
      posGoal.setX(otherTeamMinX+0.8);
      posGoal.setY(me.pos.getY());
      goToPrecision = 0.5;
      if(tS->ourTeamDefender(myTactNo)) priority = 0;
      else if (tS->ourTeamMidfield(myTactNo)) priority = 3;
      else if (tS->ourTeamOffender(myTactNo)) priority = 4;
      g.set_GOOD();
      POSTRACE("Generating Offside");
      return g;
    }
    
    // IDLE

    // REALLY NOTHING IMPORTANT TO DO (BE OFFENSIVE)
    if(tS->ourTeamOffender(myTactNo))
      posGoal = calculateFreePosition(wS,
				      max(myMinX, me.pos.getX()), // go ahead (anyway)
				      ourOffsideLine,
				      myMinY,// my side of play
				      myMaxY);
    else if(tS->ourTeamMidfield(myTactNo))
      posGoal = calculateFreePosition(wS,
				      max(myMinX, me.pos.getX()), //  go ahead (anyway)
				      max(myMaxX, me.pos.getX()+1),
				      myMinY,// my side of play
				      myMaxY);
    else if(tS->ourTeamDefender(myTactNo))
      posGoal = calculateFreePosition(wS,
				      max(myMinX, me.pos.getX()), // go ahead (anyway)
				      min(min(myMaxX, me.pos.getX()+5.0), // only up to maxX 
					  max(myMinX, me.pos.getX())),
				      myMinY,// my side of play
				      myMaxY);

    
    POSTRACE("IDLE");
    POSDRAWTRACE(CIRCLE(posGoal, 2*PLAYER_SIZE));

    if(nearestOther != UNKNOWN){
      double distToOther = dist(me.pos, wS->otherTeam(nearestOther).pos);
      // near to opponent OR offensive player => GO
      if(
	 (distToOther < COVER_DISTANCE)
	 ||(tS->ourTeamOffender(myTactNo))
	 ||(tS->ourTeamMidfield(myTactNo))
	 )
	priority = 0;
      else if(distToOther < 2*COVER_DISTANCE) // still far away from opponent
	priority = 1;
      else priority = 4;
    }
    else priority = 4;
    
    goToPrecision = 2.0;
    g.set_GOOD();
    return g;    
  }
  else{ //THEIR BALL
 
    // should not happen and is very stupid
    if (tS->firstResponsiblePlayer == UNKNOWN){ 
      posGoal = wS->ball.pos;
      priority = 4;
      goToPrecision = KICKABLE_MARGIN/2;
      g.set_BAD();
      return g;
    }
    
    // ball possessing opponent is called tS.firstDangerousOpponent (<- INT )
    // is he attacked ?
    if(tS->firstDangerousOpponent != UNKNOWN){ //Opps? => NO Opps (return BAD)
      // I am responsible for ball
      // only in PLAY_ON (not in standard situations)
      if(
	 (wS->playMode == PLAY_MODE_PLAY_ON)
	 &&
	 ((tS->firstResponsiblePlayer == me.number)
	  ||(tS->secondResponsiblePlayer == me.number))
	 ){
	POSTRACE("I am responsible for ball.");
	posGoal = wS->ball.pos; // responsible => go to ball
	// first far from ball ?
	if(dist(wS->ourTeam(tS->firstResponsiblePlayer).pos,
		wS->ball.pos)> 2*KICKABLE_MARGIN){
	  goToPrecision = KICKABLE_MARGIN/2;
	  priority = 1;
	  g.set_AVERAGE();
	  return g;
	}
	else{ // first near ball
	  goToPrecision = KICKABLE_MARGIN/2;
	  priority = 2;
	  g.set_POOR();
	  return g;
	}
      }
      
      //nothing to do with ball
      
      // COVER FREE OTHERS
      
      // detect nearest free player behind me (IMA = in my area, IMS in my sector)
      int freeNearestOtherPlayerBehindMeIMA = UNKNOWN;
      int freeNearestOtherPlayerBehindMeIMS = UNKNOWN;
      int freeNearestOtherPlayerBehindMe    = UNKNOWN;
      bool coveredOtherPlayers[12];
      
      for(int num = 1; num <= 11; num++){
	coveredOtherPlayers[num] = false;
	if(wS->otherTeam(num).age < MAXAGE){
	  POSTRACE("Player " << num << " pos: " << wS->otherTeam(num).pos);
	  // player free
	  if(otherPlayerFree(wS, num)){
	    POSTRACE("Free");
	    if(wS->otherTeam(num).pos.getX() < me.pos.getX()){ //behind me?
	      if(freeNearestOtherPlayerBehindMe != UNKNOWN){
		if(dist(wS->otherTeam(num).pos, me.pos)
		   <  dist(wS->otherTeam(freeNearestOtherPlayerBehindMe).pos,me.pos)){
		  freeNearestOtherPlayerBehindMe = num;
		}
	      }
	      else
		freeNearestOtherPlayerBehindMe = num;
	    }
	    //player in my area
	    if(playerInMyArea(wS->otherTeam(num))){
	      POSTRACE("my area");
	      // player behind me
	      if(wS->otherTeam(num).pos.getX() < me.pos.getX()+3){
		POSTRACE("my area");
		// new player closer to me
		if(freeNearestOtherPlayerBehindMeIMA != UNKNOWN){
		  if(dist(wS->otherTeam(num).pos, me.pos)
		     <  dist(wS->otherTeam(freeNearestOtherPlayerBehindMeIMA).pos,me.pos)){
		    freeNearestOtherPlayerBehindMeIMA = num;
		  }
		}
		else
		  freeNearestOtherPlayerBehindMeIMA = num;
	      }
	    }
	    else{ // not in my area, but in my sector ?
	      if(playerInMySector(wS->otherTeam(num))){
		POSTRACE("my sector");
		// new player closer to me
		if(freeNearestOtherPlayerBehindMeIMS != UNKNOWN){
		  if(dist(wS->otherTeam(num).pos, me.pos)
		     <  dist(wS->otherTeam(freeNearestOtherPlayerBehindMeIMS).pos,me.pos)){
		    freeNearestOtherPlayerBehindMeIMS = num;
		  }
		}
		else
		  freeNearestOtherPlayerBehindMeIMS = num;
	      }
	    }
	  }
	  else // player NOT free
	    coveredOtherPlayers[num] = true;
	}
	else POSTRACE("NO NUM" <<num);
      }// for num
      POSTRACE("free behind num IMA" << freeNearestOtherPlayerBehindMeIMA);      
      POSTRACE("free behind num IMS" << freeNearestOtherPlayerBehindMeIMS);
      // are there free players behind me?
      // in this case => go back
      // free player in my area?
      if(freeNearestOtherPlayerBehindMeIMA != UNKNOWN){ 
	// attack!
	posGoal = calculateCoverPosFromPlayerNo(wS, freeNearestOtherPlayerBehindMeIMA); 
	goToPrecision = COVER_DISTANCE/4;
	priority = 0;
	g.set_GOOD();
	return g;      
      }
      
      // free player in my sector
      if(freeNearestOtherPlayerBehindMeIMS != UNKNOWN){ 
	// attack!
	posGoal = calculateCoverPosFromPlayerNo(wS, freeNearestOtherPlayerBehindMeIMS);
	goToPrecision = COVER_DISTANCE/4;
	priority = 0;
	g.set_GOOD();
	return g;      
      }

      // free player behind me ?
      // if I do not cover another player => go back, direction "free player" 
      // BUT first goal is nearest opponent in this direction (not free player)
      if(freeNearestOtherPlayerBehindMe != UNKNOWN){
	// cover opponent ? => go on
	if(nearestOther != UNKNOWN){
	  int ourFirst = wS->ourTeamClosestTo(wS->otherTeam(nearestOther).pos);
	  int ourSecond = wS->ourTeamClosestTo(wS->otherTeam(nearestOther).pos, MAXAGE ,2);
	  if((ourFirst == me.number)||(ourSecond == me.number)){// it's me?
	    if(ourSecond != UNKNOWN){ // remember: ourFirst != UNKNOWN
	      // ATTENTION
	      // hack! if I am resonsible I am always: ourFirst 
	      if(ourFirst != me.number){
		int help = ourFirst;
		ourFirst = me.number;
		ourSecond = help;
	      }
	      // does second cover this player too?
	      if(wS->otherTeamClosestTo(wS->ourTeam(ourSecond).pos) == nearestOther){
		// we are two !
		int ourSecondTac = tS->ourTeamTacticalNo(ourSecond);
		if(ourSecondTac == UNKNOWN) ourSecondTac = ourSecond;
		// if tactical number UNKNOWN use normal number
		// if I am midfield and second is defender => cover
		// OR
		// if I am attacker and second is midfield => cover
		// OR
		// if I am attacker and second is defender => cover
		if(
		   ((tS->ourTeamMidfield(myTactNo))&&(tS->ourTeamDefender(ourSecondTac)))
		   ||((tS->ourTeamOffender(myTactNo))&&(tS->ourTeamMidfield(ourSecondTac)))
		   ||((tS->ourTeamOffender(myTactNo))&&(tS->ourTeamDefender(ourSecondTac)))
		   ){
		  posGoal = calculateCoverPosFromPlayerNo(wS, nearestOther); 
		  goToPrecision = COVER_DISTANCE/4;
		  if(dist(me.pos, wS->otherTeam(nearestOther).pos) < 5.0)
		    priority = 1;
		  else
		    priority = 3;	      
		  g.set_AVERAGE();
		  return g;
		}
		// we have both the same tactical job
		// decision based on tactical position
		if(
		   ((tS->ourTeamDefender(myTactNo))&&(tS->ourTeamDefender(ourSecondTac)))
		   ||((tS->ourTeamMidfield(myTactNo))&&(tS->ourTeamMidfield(ourSecondTac)))
		   ||((tS->ourTeamOffender(myTactNo))&&(tS->ourTeamOffender(ourSecondTac)))
		   ){
		  // free player on the left => left tact. player is responsible
		  // player right => vice versa
		  if(
		     (wS->otherTeam(freeNearestOtherPlayerBehindMe).pos.getY() < me.pos.getY())
		     &&(ourSecondTac < myTactNo)
		     ||
		     (wS->otherTeam(freeNearestOtherPlayerBehindMe).pos.getY() > me.pos.getY())
		     &&(ourSecondTac > myTactNo)
		     ){
		    posGoal = calculateCoverPosFromPlayerNo(wS, nearestOther);
		    goToPrecision = COVER_DISTANCE/4;
		    if(dist(me.pos, wS->otherTeam(nearestOther).pos) < 5.0)
		      priority = 1;
		    else
		      priority = 3;	      
		    g.set_AVERAGE();
		    return g;
		  }
		}
	      }
	      else{ // second covers different opponent => I will cover this one
		posGoal = calculateCoverPosFromPlayerNo(wS, nearestOther);
		goToPrecision = COVER_DISTANCE/4;
		if(dist(me.pos, wS->otherTeam(nearestOther).pos) < 5.0)
		  priority = 1;
		else
		  priority = 3;	      
		g.set_AVERAGE();
		return g;	      
	      }
	    }
	    else{ // ourSecond UNKNOWN
	      if(ourFirst == me.number){ // it's me!
		posGoal = calculateCoverPosFromPlayerNo(wS, nearestOther); 
		goToPrecision = COVER_DISTANCE/4;
		if(dist(me.pos, wS->otherTeam(nearestOther).pos) < 5.0)
		  priority = 1;
		else
		  priority = 3;	      
		g.set_AVERAGE();
		return g;	      
	      }
	    }
	  }
	}

	// now we have to detect the nearest opponent in our direction:
	int targetNo = UNKNOWN;
	for(int num = 1; num <= 11; num++){
	  if((coveredOtherPlayers[num])&&(wS->otherTeam(num).pos.getX() < me.pos.getX()+10.0)){
	    if(wS->otherTeam(num).inCorridor(me.pos, 
					     wS->otherTeam(freeNearestOtherPlayerBehindMe).pos,
					     10.0)){
	      if(targetNo != UNKNOWN){
		if(sqrnorm(me.pos - wS->otherTeam(num).pos) 
		   < sqrnorm(me.pos - wS->otherTeam(targetNo).pos))
		  targetNo = num;
	      }
	      else
		targetNo = num;
	    }
	  }
	} // for num
	
	if(targetNo != UNKNOWN){
	  POSTRACE("FREE BEHIND. Cover: " << targetNo);
	  posGoal = wS->otherTeam(targetNo).pos;
	  goToPrecision = COVER_DISTANCE/4;
	  priority = 2;
	  g.set_AVERAGE();
	  return g;	  
	}
	else{ // NO targetNo found
	  // shift in direction of free opponent
	  posGoal = me.pos;
	  if(wS->otherTeam(freeNearestOtherPlayerBehindMe).pos.getY() < me.pos.getY())
	    posGoal.setY(max(posGoal.getY()-1, myMinY));
	  else
	    posGoal.setY(min(posGoal.getY()+1, myMaxY));
	  priority = 3;
	  goToPrecision = 0.5;
	  g.set_POOR();
	  return g;
	}
      }

      // do I cover opponent => go on
      // in this case: no free opponent behind me!
      int nearestOther = wS->otherTeamClosestTo(me.pos);
      if(nearestOther != UNKNOWN){
	int ourFirst = wS->ourTeamClosestTo(wS->otherTeam(nearestOther).pos);
	int ourSecond = wS->ourTeamClosestTo(wS->otherTeam(nearestOther).pos, MAXAGE ,2);
	if((ourFirst == me.number)||(ourSecond == me.number)){// it's me?
	  if(ourSecond != UNKNOWN){ // remember: ourFirst != UNKNOWN
	    // ATTENTION
	    // hack! if I am resonsible I am always: ourFirst 
	    if(ourFirst != me.number){
	      int help = ourFirst;
	      ourFirst = me.number;
	      ourSecond = help;
	    }
	      // does second cover this player too?
	    if(wS->otherTeamClosestTo(wS->ourTeam(ourSecond).pos) == nearestOther){
	      // we are two !
	      int ourSecondTac = tS->ourTeamTacticalNo(ourSecond);
	      if(ourSecondTac == UNKNOWN) ourSecondTac = ourSecond;
	      // if tactical number UNKNOWN use normal number
	      // if I am midfield and second is attacker => cover
	      // OR
	      // if I am defender and second is midfield => cover
	      // OR
	      // if I am defender and second is attacker => cover
	      // notice the changed tactical setting! (because: no free opponent behind!)
	      if(
		 ((tS->ourTeamMidfield(myTactNo))&&(tS->ourTeamOffender(ourSecondTac)))
		 ||((tS->ourTeamDefender(myTactNo))&&(tS->ourTeamMidfield(ourSecondTac)))
		 ||((tS->ourTeamDefender(myTactNo))&&(tS->ourTeamOffender(ourSecondTac)))
		 ){
		posGoal = calculateCoverPosFromPlayerNo(wS, nearestOther);
		goToPrecision = COVER_DISTANCE/4;
		  if(dist(me.pos, wS->otherTeam(nearestOther).pos) < 5.0)
		    priority = 1;
		  else
		    priority = 1; //!! 
		  g.set_AVERAGE();
		  return g;
		}
		// we have both the same tactical job
		// decision based on tactical position
		if(
		   ((tS->ourTeamDefender(myTactNo))&&(tS->ourTeamDefender(ourSecondTac)))
		   ||((tS->ourTeamMidfield(myTactNo))&&(tS->ourTeamMidfield(ourSecondTac)))
		   ||((tS->ourTeamOffender(myTactNo))&&(tS->ourTeamOffender(ourSecondTac)))
		   ){
		  // ball on the left => left tact. player is responsible
		  // ball right => vice versa
		  if(
		     (wS->ball.pos.getY() < me.pos.getY())&&(ourSecondTac < myTactNo)
		     ||
		     (wS->ball.pos.getY() > me.pos.getY())&&(ourSecondTac > myTactNo)
		     ){
		    posGoal = calculateCoverPosFromPlayerNo(wS, nearestOther); 
		    goToPrecision = COVER_DISTANCE/4;
		    if(dist(me.pos, wS->otherTeam(nearestOther).pos) < 5.0)
		      priority = 1;
		    else
		      priority = 1; //!!
		    g.set_AVERAGE();
		    return g;
		  }
		}		  
	      }
	      else{ //  second covers different opponent => I will cover this one
		posGoal = calculateCoverPosFromPlayerNo(wS, nearestOther);
		goToPrecision = COVER_DISTANCE/4;
		if(dist(me.pos, wS->otherTeam(nearestOther).pos) < 5.0)
		  priority = 1;
		else
		  priority = 1;	//!!      
		g.set_AVERAGE();
		return g;	      
	      }
	    }
	    else{ // ourSecond UNKNOWN
	      if(ourFirst == me.number){ // it's me!
		posGoal = calculateCoverPosFromPlayerNo(wS, nearestOther);
		goToPrecision = COVER_DISTANCE/4;
		if(dist(me.pos, wS->otherTeam(nearestOther).pos) < 5.0)
		  priority = 1;
		else
		  priority = 1; //!!
		g.set_AVERAGE();
		return g;	      
	      }
	    }
	  }
	}

      // NO ONE TO COVER
      
      // offside ?
      if(me.pos.getX() > ourOffsideLine){
	POSTRACE("DEF OFF");
	// go back
	posGoal.setX(me.pos.getX()-1);
	// avoid ball
	if(wS->ball.pos.getY() < me.pos.getY())
	  posGoal.setY(me.pos.getY()+1);
	else
	  posGoal.setY(me.pos.getY()-1); 
	priority = 1;
	goToPrecision = 0.5;
	g.set_GOOD();
	return g;
      }
      
      // NO ONE TO COVER, I DO NOT COVER ANYONE
      // => eliminate good pass receivers
      // therefor find the next good one
      if(tS->firstDangerousOpponent != UNKNOWN){
	int bestOtherPassPartner = UNKNOWN;
	Grade test_g;
	Grade help_g;
	test_g.set_IMPOSSIBLE();
	for(int num = 1; num <= 11; num++){
	  if(
	     (wS->otherTeam(num).age < MAXAGE)
	     &&
	     (wS->otherTeam(num).pos.getX() > me.pos.getX())
	     &&
	     (wS->otherGoalieNo != num)
	     ){
	      help_g = passEval->otherEval(wS, tS->firstDangerousOpponent, num);
	      if(
		 (help_g.is_better_than(test_g))
		 ||
		 ((help_g.is_like(test_g))
		  &&(sqrnorm(me.pos - wS->otherTeam(num).pos) 
		     < sqrnorm(me.pos - wS->otherTeam(bestOtherPassPartner).pos)))
		 ){
		test_g = help_g;
		bestOtherPassPartner = num;
	      }
	  }
	}// for
	POSTRACE("Detected good other pass receiver: "<< bestOtherPassPartner);
	// found good one
	if(test_g.is_PERFECT()||test_g.is_GOOD()){
	  posGoal = calculateCoverPosFromPlayerNo(wS, bestOtherPassPartner);
	  goToPrecision = COVER_DISTANCE/4;
	  priority = 3;
	  g.set_AVERAGE();
	  return g;
	}
      } //if(tS->firstDangerousOpponent != UNKNOWN)
      
      // ball in front of me 
      if(wS->ball.pos.getX() > me.pos.getX())
	 if(
	    (wS->ball.pos.getY() > myMinY)
	    &&(wS->ball.pos.getY() < myMaxY)
	    ){
	   POSTRACE("BALL IN MY FRONT");
	   posGoal = wS->ball.pos;
	   priority = 4;
	   g.set_POOR();
	   return g;
	 }
	 else{ // ball in front, but not interesting => go forward
	   posGoal =  calculateFreePosition(wS,
					    me.pos.getX(),
					    min(me.pos.getX()+15, FIELD_LENGTH/2),
					    myMinY,
					    myMaxY);
	   priority = 2;
	   goToPrecision = 0.5;
	   g.set_AVERAGE();
	   return g;
	 }
      
      
      // REALLY NOTHING IMPORTANT TO DO (BE DEFENSIVE)
      // ATTENTION: do not be to defensiv (role dependend ?)
      POSTRACE("GO BACK");
      // if defender OR midfield => return to own area
      if(
	 (tS->ourTeamDefender(myTactNo))
	 ||
	 (tS->ourTeamMidfield(myTactNo))
	 ){
	posGoal.setX((myMaxX + myMinX)/2);
	posGoal.setY((myMaxY + myMinY)/2);
	priority = 3; // defender, who does not cover!
	goToPrecision = 2.0;
	g.set_AVERAGE();
	return g;
      }
      
      // if attacker, search free space (even in defensive situations)
      if(tS->ourTeamOffender(myTactNo)){
	posGoal =  calculateFreePosition(wS,
					 max(max(me.pos.getX()-10, otherTeamMaxX-15), myMinX),
					 min(me.pos.getX()+5, myMaxX),
					 max(me.pos.getY()-20, myMinY),
					 min(me.pos.getY()+20, myMaxY));
	priority = 4;
	goToPrecision = 0.5;
	g.set_AVERAGE();
	return g;
      }

      POSTRACE("FALL BACK ERREICHT");
      posGoal = me.pos;
      posGoal.setX(posGoal.getX()-1);
      priority = 4;
      goToPrecision = 0.5;
      g.set_AVERAGE();
      return g;
    }// Opps
    else{ // NO opps
      POSTRACE("DEF no opps");
      posGoal = wS->ball.pos;
      priority = 3;
      g.set_BAD();
      return g;
    }
  } // their ball
  return g;
}

//PRIVATE

bool PositionModule::amIPassReceiver(const Snapshot  *wS,
				     const TacticalSnapshot *tS){
  int firstResponsibleNumber = tS->firstResponsiblePlayer;
  // if nobody is responsible, I am not passreceiver
  if(firstResponsibleNumber == UNKNOWN) return false;
  
  double distance = dist(me.pos, wS->ourTeam(firstResponsibleNumber).pos);
  // Am I in pass distance or not ?
  return ((distance<25)&&(distance > 10));
}

bool PositionModule::amI2LevelPassReceiver(const Snapshot* wS,
					   const TacticalSnapshot* tS){
  if(wS);
  if(tS);
  
  return false;
}

void PositionModule::calculateMaximalPositions(const Snapshot* wS,
					       const TacticalSnapshot* tS)
{
  double backDefenseLine  = otherOffsideLine, frontDefenseLine  = -FIELD_LENGTH/2+1;
  double backMidfieldLine = FIELD_LENGTH/2-1,  frontMidfieldLine = -FIELD_LENGTH/2+1; 
  double backOffenseLine  = FIELD_LENGTH/2-1,  frontOffenseLine  = ourOffsideLine;

  double xPos;
  
  //calculate some lines
  for(int num = 1; num <= 11; num++){
    if(
       (tS->ourTeamTacticalNo(num) != UNKNOWN)
       &&(wS->ourTeam(num).age <= MAXAGE)
       &&(num != me.number)
       ){
      xPos = wS->ourTeam(num).pos.getX(); // x position of player "num"
      if(tS->ourTeamDefender(tS->ourTeamTacticalNo(num))){//defense 
	frontDefenseLine = max(frontDefenseLine,xPos);
	frontDefenseLine = min(frontDefenseLine, otherTeamMinX+10);
      }
      else{
	if(tS->ourTeamMidfield(tS->ourTeamTacticalNo(num))){//midfield
	  frontMidfieldLine = max(frontMidfieldLine, xPos);
	  backMidfieldLine = min(backMidfieldLine, xPos);
	}
	else{
	  if(tS->ourTeamOffender(tS->ourTeamTacticalNo(num))){//offense
	    backOffenseLine = min(backOffenseLine, xPos);
	  }
	  else{//goalie
	  }
	}
      }
    }
  }
  
  if(tS->ourTeamDefender(myTactNo)){//I am defender 
    myMinX = backDefenseLine+1;  //back
    myMaxX = min(backMidfieldLine-2, 1.0); //front (at least one meter in opponent half)
   }
  else{
    if(tS->ourTeamMidfield(myTactNo)){//I am midfielder
      myMinX = frontDefenseLine+2; //back
      myMaxX = backOffenseLine-2;  //front
    }
    else{
      if(tS->ourTeamOffender(myTactNo)){//I am offender
	myMinX = max(frontMidfieldLine+2,otherTeamMaxX-15); //back
	myMaxX = frontOffenseLine-1;  //front
      }
      else{//goalie
	myMinX = -FIELD_LENGTH/2;
	myMaxX = -FIELD_LENGTH/2+PENALTY_AREA_LENGTH-CATCHABLE_AREA_L/2;
      }
    }
  }
  // of course maxX >= minX
  myMaxX = max(myMinX, myMaxX);
  
  int leftNeighbourNo = tS->ourTeamLeftNeighbour(myTactNo);
  int rightNeighbourNo = tS->ourTeamRightNeighbour(myTactNo);
  POSTRACE("my left" << leftNeighbourNo);
  POSTRACE("my right" << rightNeighbourNo);
  myMaxY = FIELD_WIDTH/2-3;
  myMinY = -FIELD_WIDTH/2+3;
  if(leftNeighbourNo != UNKNOWN){
    myMinY = max(myMinY, wS->ourTeam(leftNeighbourNo).pos.getY()+2);//left
  }
  if(rightNeighbourNo != UNKNOWN){
    myMaxY = min(myMaxY, wS->ourTeam(rightNeighbourNo).pos.getY()-2); //right
  }
  // goalie
  if(!(tS->ourTeamDefender(myTactNo))
     &&!(tS->ourTeamMidfield(myTactNo))
     &&!(tS->ourTeamOffender(myTactNo))){
    myMinY = -PENALTY_AREA_WIDTH/2+CATCHABLE_AREA_L;//left
    myMaxY =  PENALTY_AREA_WIDTH/2-CATCHABLE_AREA_L;//right
  }
  // of course maxY >= minY
  myMaxY = max(myMinY, myMaxY);
  
  POSTRACE("my Position " << me.pos);
  POSTRACE("minX " << myMinX << " maxX " << myMaxX << " minY " << myMinY << " maxY " << myMaxY );
}

void PositionModule::truncatePosition()
{
  // truncate x position
  posGoal.setX(min(posGoal.getX(), myMaxX)); // front
  posGoal.setX(max(posGoal.getX(), myMinX)); // back
  // truncate y position
  posGoal.setY(max(posGoal.getY(), myMinY)); // left
  posGoal.setY(min(posGoal.getY(), myMaxY)); // right
}

bool PositionModule::otherPlayerFree(const Snapshot *wS,int num, double distance){

  // opponent "num" valid ? and our nearest is known
  if(
     (wS->otherTeam(num).age < MAXAGE)
     &&
     (wS->ourTeamClosestTo(wS->otherTeam(num).pos) != UNKNOWN)
     ){
    // in standard situations players near to the ball are NOT free
    if(
       (wS->playMode == PLAY_MODE_KICK_IN_THEM)
       ||
       (wS->playMode == PLAY_MODE_FREE_KICK_THEM)
       ||
       (wS->playMode == PLAY_MODE_CORNER_KICK_THEM)
       ||
       (wS->playMode == PLAY_MODE_GOAL_KICK_THEM)
       ){
      if(dist(wS->otherTeam(num).pos, wS->ball.pos) < FREE_KICK_FORBIDDEN_AREA)
	return false;
    }
    
    // our player near to opponent "num"
    if (dist(wS->otherTeam(num).pos,
	     wS->ourTeam(wS->ourTeamClosestTo(wS->otherTeam(num).pos)).pos)
	< distance){ 
      // if our player is not nearer to another player
      if  (wS->otherTeamClosestTo(wS->ourTeam(wS->ourTeamClosestTo(wS->otherTeam(num).pos)).pos)
	   == num)
	return false;
      else
	return true;
    }
  }

  // in doubt, return true
  return true;
}
bool PositionModule::ourPlayerFree(const Snapshot *wS,int num, double distance){

  // player valid, opponent near ?
  if ((wS->ourTeam(num).age < MAXAGE)
      &&(wS->otherTeamClosestTo(wS->ourTeam(num).pos) != UNKNOWN))
    if (dist(wS->ourTeam(num).pos,
	     wS->otherTeam(wS->otherTeamClosestTo(wS->ourTeam(num).pos)).pos)
	< distance)
      return false;
    else
      return true;
  // in doubt, return true
  return true;  
}

void PositionModule::respectStandardSituations(const Snapshot* wS,
					       const TacticalSnapshot* ){
  switch(wS->playMode){
  case PLAY_MODE_UNKNOWN : 
    break;
  case PLAY_MODE_BEFORE_KICK_OFF :
    posGoal.setX(min(posGoal.getX(),-PLAYER_SIZE));
    break;
  case PLAY_MODE_TIME_OVER :
    posGoal = me.pos;
    break;
  case PLAY_MODE_PLAY_ON : 
    break;
  case PLAY_MODE_KICK_OFF_US :
  case PLAY_MODE_KICK_OFF_THEM :
    posGoal.setX(min(posGoal.getX(),-PLAYER_SIZE));    
    break;
  case PLAY_MODE_KICK_IN_US : 
    break;
  case PLAY_MODE_KICK_IN_THEM : 
    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA))
      posGoal = me.pos;
    break;
  case PLAY_MODE_FREE_KICK_US : 
    break;
  case PLAY_MODE_FREE_KICK_THEM :
    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA))
      if((posGoal - me.pos).getAngle() - (wS->ball.pos - me.pos).getAngle() > 0)//right
	posGoal = (wS->ball.pos - me.pos).perp()*(1/norm(wS->ball.pos - me.pos));
      else // left
	posGoal = (me.pos - wS->ball.pos).perp()*(1/norm(me.pos - wS->ball.pos));
    break;    
  case PLAY_MODE_CORNER_KICK_US : 
    break;
  case PLAY_MODE_CORNER_KICK_THEM :
    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA))
      posGoal = me.pos;
    break;
  case PLAY_MODE_GOAL_KICK_US : 
    break;
  case PLAY_MODE_GOAL_KICK_THEM :
    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA))
      posGoal = me.pos;
    break;
  case PLAY_MODE_GOAL_US :
    posGoal = me.pos;
    break;
  case PLAY_MODE_DROP_BALL :
    break;
  case PLAY_MODE_GOAL_THEM :
    posGoal = me.pos;
    break;    
  case PLAY_MODE_OFFSIDE_US :
    break;
  case PLAY_MODE_OFFSIDE_THEM :
    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA))
      if((posGoal - me.pos).getAngle() - (wS->ball.pos - me.pos).getAngle() > 0)//right
	posGoal = (wS->ball.pos - me.pos).perp()*(1/norm(wS->ball.pos - me.pos));
      else // left
	posGoal = (me.pos - wS->ball.pos).perp()*(1/norm(me.pos - wS->ball.pos));
    break;    
  case PLAY_MODE_HALF_TIME :
    posGoal = me.pos;
    break;        
  case PLAY_MODE_TIME_EXTENDED:
    posGoal = me.pos;    
    break;
  }

  /*
    PLAY_MODE_UNKNOWN = -1,
    PLAY_MODE_BEFORE_KICK_OFF = 0,
    PLAY_MODE_TIME_OVER,
    PLAY_MODE_PLAY_ON,
    PLAY_MODE_KICK_OFF_US,
    PLAY_MODE_KICK_OFF_THEM,
    PLAY_MODE_KICK_IN_US,
    PLAY_MODE_KICK_IN_THEM,
    PLAY_MODE_FREE_KICK_US,
    PLAY_MODE_FREE_KICK_THEM,
    PLAY_MODE_CORNER_KICK_US,
    PLAY_MODE_CORNER_KICK_THEM,
    PLAY_MODE_GOAL_KICK_US,
    PLAY_MODE_GOAL_KICK_THEM,
    PLAY_MODE_GOAL_US,
    PLAY_MODE_GOAL_THEM,
    PLAY_MODE_DROP_BALL,
    PLAY_MODE_OFFSIDE_US,
    PLAY_MODE_OFFSIDE_THEM,
    PLAY_MODE_HALF_TIME,
    PLAY_MODE_TIME_EXTENDED
    */
}

void PositionModule::avoidPlayerCollision(const Snapshot *wS){
  // if otherPlayer or ourPlayer (incl. me!) in my way
  if(dist(posGoal, me.pos) > PLAYER_SIZE){
    if(
       (wS->otherTeamInCorridor(me.pos, me.pos +
				radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle()),
						4*PLAYER_SIZE) > 0)
	||
	(wS->ourTeamInCorridor(me.pos, me.pos +
			       radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle()),
					       4*PLAYER_SIZE) > 1)
       ){
      POSDRAWTRACE(CTEXT("player collision"));
      if((wS->ball.pos - me.pos).getAngle() > 0) // ball right of me
	// ball right => side step right of player
	// to avoid interference with free kick areas
	posGoal = radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle())
	  + radAng2Vector2d(5*PLAYER_SIZE, (posGoal-me.pos).getAngle()).perp();
      else // vice versa
	posGoal = radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle())
	  - radAng2Vector2d(5*PLAYER_SIZE, (posGoal-me.pos).getAngle()).perp();
    }
  }
}

bool PositionModule::playerInMyArea(const WorldPlayer pl){
  return ((pl.pos.getX() < myMaxX)&&(pl.pos.getX() > myMinX)
    && (pl.pos.getY() < myMaxY)&&(pl.pos.getY() > myMinY));
}

bool PositionModule::playerInMySector(const WorldPlayer pl){
  return ((pl.pos.getX() < me.pos.getX()+5)
    && (pl.pos.getY() < myMaxY)&&(pl.pos.getY() > myMinY));
}

double PositionModule::evalPos(const Snapshot *wS, const Vector2d &evalPos){
  double helpValue;
  double ergValue = 0; 
  int ourP = UNKNOWN;
  int otherP = UNKNOWN;
  
  // distance to side line
  ergValue = sqr(min(FIELD_WIDTH/2+evalPos.getY(),
		     FIELD_WIDTH/2-evalGrid[oldX][oldY].getY()));
  // distance to goal line
  helpValue = sqr(min(FIELD_LENGTH/2+evalPos.getX(),
		      FIELD_LENGTH/2-evalPos.getX()));
  ergValue = min(ergValue, helpValue);
  ourP = wS->ourTeamClosestTo(evalPos);
  if(ourP == me.number) ourP = wS->ourTeamClosestTo(evalPos, MAXAGE, 2);
  otherP = wS->otherTeamClosestTo(evalPos);
  if(ourP != UNKNOWN){
    helpValue = sqrnorm(wS->ourTeam(ourP).pos - evalPos);
    ergValue = min(helpValue, ergValue);
  }
  if(otherP != UNKNOWN){
    helpValue = sqrnorm(wS->otherTeam(otherP).pos - evalPos);
    ergValue = min(ergValue, helpValue);
  }
  
  return ergValue;
}

Vector2d  PositionModule::calculateFreePosition(const Snapshot *wS,
						double minX, 
						double maxX, 
						double minY, 
						double maxY){
  /*
    POSTRACE("old pos " << oldPosGoal);
    Vector2d testPos = oldPosGoal;
    Vector2d ergPos = oldPosGoal;
    double testValue;
    double helpValue;
    double ergValue; // schlechter kann eine Pos nicht sein
    int ourP = UNKNOWN;
    int otherP = UNKNOWN;
  */
  
  int testX = oldX, testY = oldY;
  int ergX = oldX, ergY = oldY;
  double ergValue = 0;
  double helpValue;
  double testValue = 0;
  
  int xStart=xPoints-1;
  int xEnd=0;
  int yStart=yPoints-1;
  int yEnd=0;
  while((xStart > 0)&&(evalGrid[xStart][0].getX() > minX )) xStart--;
  while(((xEnd < xPoints-1)&&(evalGrid[xEnd][0].getX() < maxX ))||(xEnd < xStart)) xEnd++;
  while((yStart > 0)&&(evalGrid[0][yStart].getY() > minY)) yStart--;
  while(((yEnd < yPoints-1 )&&(evalGrid[0][yEnd].getY() < maxY))||(yEnd < yStart)) yEnd++;
  POSTRACE("xS " << xStart << " xE " << xEnd << " yS " << yStart << " yE " << yEnd);
  POSTRACE("xS " << evalGrid[xStart][0].getX() << " xE " << evalGrid[xEnd][0].getX() 
	   << " yS " << evalGrid[0][yStart].getY() << " yE " << evalGrid[0][yEnd].getY());

  if(// Did I evaluate last time
     // and is the old position in the new grid ?
     // => evaluate old position again
     // if in this adjacence is a better point (20%)
     // take this one
     (lastEvalTime <= wS->time - 1) // -3?
     &&(oldX >= xStart)&&(oldX <= xEnd)
     &&(oldY >= yStart)&&(oldY <= yEnd)
     ){
    ergValue = evalPos(wS, evalGrid[oldX][oldY]);
    ergX = oldX;
    ergY = oldY;
    for(int xs = -1; xs <= 1; xs++)
      for(int ys = -1; ys <= 1; ys++)
	if(
	   (oldX+xs >= xStart)&&(oldX+xs <= xEnd)
	   &&(oldY+ys >= yStart)&&(oldY+ys <= yEnd)
	   ){
	  helpValue = evalPos(wS, evalGrid[oldX+xs][oldY+ys]);
	  if(helpValue > testValue){
	    testValue = helpValue;
	    testX = oldX+xs;
	    testY = oldY+ys;
	  }
	}
    if(testValue > 1.2*ergValue){
      ergValue = testValue;
      ergX = testX;
      ergY = testY;
    }
  }

  // evaluate everything
  // if there is a substantially better position (50%)
  // take it
  testValue = 0;
  for(int xs = xStart; xs <= xEnd; xs++){
    for(int ys = yStart; ys <= yEnd; ys++){
      helpValue = evalPos(wS, evalGrid[xs][ys]);
      POSDRAWTRACE(CIRCLE(evalGrid[xs][ys], PLAYER_SIZE));
      if(helpValue > testValue){
	testValue = helpValue;
	testX = xs;
	testY = ys;
      }
    }
  }
  
  if(testValue > 1.5*ergValue){
    ergValue = testValue;
    ergX = testX;
    ergY = testY;
  }
  
  // store results
  oldX = ergX;
  oldY = ergY;
  lastEvalTime = wS->time;
  POSDRAWTRACE(CIRCLE(evalGrid[ergX][ergY], 2*PLAYER_SIZE));
  return evalGrid[ergX][ergY];

  // old stuff (before Melbourne)
  /* 
    if(
    (oldPosGoal.getX() > minX)
    &&(oldPosGoal.getX() < maxX)
    &&(oldPosGoal.getY() > minY)
    &&(oldPosGoal.getY() < maxY)
    ){  // bewerten der alten pos 
    // Wenn diese nicht wesenlich schlechter geworden ist, kann sie beibehalten werden.
    // Abstand zur Seitenlinie
    ergValue = sqr(min(FIELD_WIDTH/2+testPos.getY(),FIELD_WIDTH/2-testPos.getY()));
    // Abstand zur Grundlinie
    helpValue = sqr(min(FIELD_LENGTH/2+testPos.getX(),FIELD_LENGTH/2-testPos.getX()));
    ergValue = min(ergValue, helpValue);
    ourP = wS->ourTeamClosestTo(testPos, MAXAGE, 2);
    otherP = wS->otherTeamClosestTo(testPos);
    if(ourP != UNKNOWN){
    helpValue = sqrnorm(wS->ourTeam(ourP).pos - testPos);
      ergValue = min(helpValue, ergValue);
      }
      if(otherP != UNKNOWN){
      helpValue = sqrnorm(wS->otherTeam(otherP).pos - testPos);
      ergValue = min(ergValue, helpValue);
      }
      }
      else ergValue = 0;
      
      int xStart=xPoints-1;
      int xEnd=0;
      int yStart=yPoints-1;
      int yEnd=0;
      while((xStart > 0)&&(evalGrid[xStart][0].getX() > minX )) xStart--;
      while(((xEnd < xPoints-1)&&(evalGrid[xEnd][0].getX() < maxX ))||(xEnd < xStart)) xEnd++;
      while((yStart > 0)&&(evalGrid[0][yStart].getY() > minY)) yStart--;
      while(((yEnd < yPoints-1 )&&(evalGrid[0][yEnd].getY() < maxY))||(yEnd < yStart)) yEnd++;
      POSTRACE("xS " << xStart << " xE " << xEnd << " yS " << yStart << " yE " << yEnd);
      POSTRACE("xS " << evalGrid[xStart][0].getX() << " xE " << evalGrid[xEnd][0].getX() 
      << " yS " << evalGrid[0][yStart].getY() << " yE " << evalGrid[0][yEnd].getY());
      // evaluate other positions
      for(int xs = xStart; xs <= xEnd; xs++){
      for(int ys = yStart; ys <= yEnd; ys++){
      testPos = evalGrid[xs][ys];
      POSDRAWTRACE(CIRCLE(testPos, PLAYER_SIZE));
      // distance to side line
      testValue = sqr(min(FIELD_WIDTH/2+testPos.getY(),FIELD_WIDTH/2-testPos.getY()));
      // distance to goal line
      helpValue = sqr(min(FIELD_LENGTH/2+testPos.getX(),FIELD_LENGTH/2-testPos.getX()));
      testValue = min(testValue, helpValue);
      ourP = wS->ourTeamClosestTo(testPos);
      if(ourP == me.number) ourP = wS->ourTeamClosestTo(testPos, MAXAGE, 2);
      otherP = wS->otherTeamClosestTo(testPos);
      if(ourP != UNKNOWN){
      helpValue = sqrnorm(wS->ourTeam(ourP).pos - testPos);
      testValue = min(testValue, helpValue);
      }
      if(otherP != UNKNOWN){
      helpValue = sqrnorm(wS->otherTeam(otherP).pos - testPos);
      testValue = min(testValue, helpValue);
      }
      // if 10% better
      if(testValue > 1.1*ergValue){
      ergValue = testValue;
      ergPos = testPos;
      }
      } 
      }
      
      return ergPos;
      */
}

Vector2d PositionModule::calculateCoverPosFromPlayerNo(const Snapshot *wS,
						       int plNo){
  	
  return wS->otherTeam(plNo).pos
    +(wS->ball.pos - wS->otherTeam(plNo).pos)
    *(2.0/norm(wS->ball.pos - wS->otherTeam(plNo).pos));
}


bool PositionModule::act(AdvSkills* skills,
			 const Snapshot* wS, TacticalSnapshot* tS)
{

  // are the skills available?
  if(skills == NULL)
    {
      cerr << "POS: act -> no skills available" << endl;
      return false;
    }

  // do we have a valid snapshot?
  if(wS == NULL)
    {
      cerr << "POS: act -> snapshot is NULL" << endl;
      skills->noAction();
      return true;
    }

  // just to undo the unsused-parameter-warnings
  // as sonn as the tactical snapshot will be used in this methode it must be tested
  // if it is not NULL
  if (tS == NULL)
     {
      cerr << "POS: act -> tactSnapshot is NULL" << endl;
      skills->noAction();
      return true;
    }

  if(posBallControl == POS_BC_WE) POSTRACE("Unser Ball.");
  else POSTRACE("Deren Ball.");
  // Compute useable stamina:
  // There are different levels of importance and for each level there is a
  // specific stamina.
  // Always remember: this module is responsible for saving stamina!
  double useableStam = 0; // default
  if(wS->senseBody.stamina > 3400){
    useableStam = 100;
  }
  else
    if(wS->senseBody.stamina > 3100){
      useableStam = posStamina[0][priority];
    }
    else
      if(wS->senseBody.stamina > 2800){
	useableStam = posStamina[1][priority];
      }
      else
	if(wS->senseBody.stamina > 2500){
	  useableStam = posStamina[2][priority];
	}
	else{
	  if(priority == 0) useableStam = STAMINA_INC_MAX;
	}

  
  POSDRAWTRACE(LCIRCLE(posGoal, 2*PLAYER_SIZE));
  POSTRACE("priority " << priority);
  POSTRACE("Przision: " << goToPrecision);
  POSTRACE("Stamina: " << useableStam);
  // keep care of standard situation areas
  respectStandardSituations(wS, tS);
  // avoid player collisions
  avoidPlayerCollision(wS);
  skills->changeView(wS, VIEW_WIDTH_NORMAL);
  // the GOALIE
  if(isGoalie){  
    AdvSkills::TurtleCommand tc 
      = skills->goToPosition(wS, posGoal, goToPrecision,useableStam,
			     true, (wS->ball.pos - me.pos).getAngle());
    // if backwards, double stamina
    if(tc.state == AdvSkills::TGOINGBACKWARDS)
      tc = skills->goToPosition(wS, posGoal, goToPrecision, 2*useableStam,
			     true, (wS->ball.pos - me.pos).getAngle());

    AdvSkills::TurnHeadCommand thc;
    if (tc.state == AdvSkills::TFINISHED){// I am at my desired position => turn to ball
      AdvSkills::TurnCommand tuc 
	= skills->turn(wS,me.bodyDir+me.bodyAngleTo(wS->ball));
      thc = skills->turnHead(wS, (wS->ball.pos - me.pos).getAngle()+me.bodyDir-tuc.bodyDir);
      if(thc.state != AdvSkills::THFINISHED)
	skills->exec(wS, thc);
      if(tuc.state == AdvSkills::RFINISHED)
	skills->noAction();
      else
	skills->exec(wS, tuc);
    }
    else{ // skills not finished
      skills->exec(wS,tc);      
      // turn head to ball (goalie often looks at ball)
      thc = skills->turnHeadToBall(wS,tc);
      if(thc.state != AdvSkills::THFINISHED)
	skills->exec(wS,thc);
    }
    /*    POSTRACE("act called with"<< " posGoal " << posGoal << " my Pos " << me.pos 
	  << " goToPrecision " << goToPrecision 
	  << " Ball " << wS->ball.pos << " Stamina " 
	  << availableStamina[priority]); */
  }
  else{// no goalie
    AdvSkills::TurtleCommand tc 
      = skills->goToPosition(wS, posGoal, goToPrecision, useableStam);
    
    AdvSkills::TurnHeadCommand thc;
    // turn head to ball OR oldest WorldObject 
    if(((internalTime % 4) == 0)
       ||((internalTime % 4) == 1)
       //	 ||((internalTime % 6) == 2)
       ){
      thc = skills->turnHeadToBall(wS,tc);
      if(thc.state != AdvSkills::THFINISHED)
	skills->exec(wS,thc);
    }
    else{
      // detect oldest object
      WorldMovedObject lookGoal = wS->ball;
      for(int i = 1; i <= 11; i++){
	if((wS->ourTeam(i).age != ETERNAL)
	   &&(wS->ourTeam(i).age > lookGoal.age)
	   &&(i != me.number))
	  lookGoal = wS->ourTeam(i);
	if((wS->otherTeam(i).age != ETERNAL)&&(wS->otherTeam(i).age > lookGoal.age))
	  lookGoal = wS->otherTeam(i);
      }
      POSTRACE("oldest Age: " << lookGoal.age << " at: " << lookGoal.pos);
      thc = skills->turnHead(wS, (lookGoal.pos - me.pos).getAngle()+me.bodyDir-tc.bodyDir);
      if(thc.state != AdvSkills::THFINISHED)
	skills->exec(wS,thc);
    }
    
    if (tc.state == AdvSkills::TFINISHED)
      skills->noAction();
    else
      skills->exec(wS,tc);
  }
  
  return true;
}

bool PositionModule::getSayMsg(AudioMsg &audioMsg)
{
  audioMsg.markWrite();
  audioMsg << sendInfo;
  
  // check if error occurred
  if( !audioMsg.getState() )
    {
      cerr << "POS: Stream overflow: " << audioMsg << endl;
      audioMsg.rollbackWrite().clearState();
      
      audioMsg << false;
      
    }

  return true;
}

/////////////////////////////////////////////
//
// PositionModuleInfo::PositionModuleInfo
//
/////////////////////////////////////////////

PositionModuleInfo::PositionModuleInfo(bool _valid)
{
  valid = _valid;
}

/////////////////////////////////////////////
//
// PositionModuleInfo::~PositionModuleInfo
//
/////////////////////////////////////////////

PositionModuleInfo::~PositionModuleInfo()
{
}


/////////////////////////////////////////////
//
// AudioMsg &operator<<
//
/////////////////////////////////////////////

class AudioMsg &operator<< (AudioMsg &aumsg, const PositionModuleInfo &info)
{
  aumsg << info.valid;
  if (info.valid);
  // aumsg << //further data structures to write ;"
  return aumsg;
}


/////////////////////////////////////////////
//
// AudioMsg &operator>>
//
/////////////////////////////////////////////

class AudioMsg &operator>> (AudioMsg &aumsg, PositionModuleInfo &info)
{
  aumsg >> info.valid;
  if (info.valid);
  // aumsg >> // further data structures to read ;"
  return aumsg;
}
